package irrigation;

import java.io.ObjectInputStream.GetField;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.swing.JCheckBox;
import static tools.HelperFunctions.*;
import static irrigation.ScenarioHelper.getFailedDevices;
import static org.junit.Assert.*;

import org.junit.Test;

import simulation.SimTime;
import util.Day;

import device.StoreFailureException;

/**
 * Tests the basic flow of AquaLush's fourth use case.
 */
public class UseCaseFourBasicFlowTest extends UseCaseTest {

	private JCheckBox v09;
	private JCheckBox s3;

	@Test
	public void executeUseCase() {
		prepareUseCase();
		performStepOne();
		testStepTwo();
		performStepFive();
		testStepSix();
		performStepThree();
		testStepFour();
	}

	protected void testStepFour() {
		// 4. AquaLush resets its clock and confirms it to the Maintainer. The
		// current time is used to control automatic irrigation immediately.

		SimTime time = SimTime.instance();
		assertEquals(Day.SATURDAY, time.getDay());
		assertEquals(0, time.getSecond());
		assertEquals(0, time.getMinute());
		assertEquals(19, time.getHour());
		
		clock.tick();

		try {
			assertNotNull(extractField(irrigator, "cycle"));
		} catch (Exception e) {
			e.printStackTrace();
			fail();
		}
	}

	protected void performStepThree() {
		// 3. Maintainer sets the current time.

		// Leave the "Control Irrigation" screen.
		buttons[7].doClick();

		// Change the current mode back to automatic.
		buttons[3].doClick();

		// Open "Set up Automatic Irrigation", ...
		buttons[4].doClick();
		// ... select "Set Irrigation Time", ...
		buttons[3].doClick();
		// ... and change the irrigation time to 1900.
		for (int i = 0; i < 4; i++)
			keys[9].doClick();

		keys[0].doClick();
		keys[8].doClick();
		keys[10].doClick();
		keys[10].doClick();

		// Accept the settings
		buttons[7].doClick();
		buttons[7].doClick();

		clock.tick();

		try {
			assertNull(extractField(irrigator, "cycle"));
		} catch (Exception e) {
			e.printStackTrace();
			fail();
		}

		// Go to "Set the Clock" menu ...
		buttons[2].doClick();
		// ... select Saturday ...
		buttons[3].doClick();
		// ... and set the time to 1900 ...
		for (int i = 0; i < 4; i++)
			keys[9].doClick();

		keys[0].doClick();
		keys[8].doClick();
		keys[10].doClick();
		keys[10].doClick();

		buttons[7].doClick();

	}

	private void testStepSix() {
		// 6. AquaLush adjusts its persistent store to record the repairs.
		// AquaLush begins using the repaired Sensors and Valves for irrigation
		// immediately.

		try {
			assertEquals("true", store.getData("V09.isWorking", "false"));
			assertEquals("true", store.getData("S3.isWorking", "false"));
		} catch (StoreFailureException e) {
			fail();
		}

		s3.doClick();
		v09.doClick();

		// Go back to main menu.
		//buttons[7].doClick(); // done already
		// Go to the "Control Irrigation" screen, ...
		buttons[7].doClick();
		// ... select valve "V10", ...
		for (int i = 0; i < 10; i++)
			buttons[3].doClick();
		// ... open the valve ...
		buttons[2].doClick();
		// ... and propagate the change to the whole zone.
		buttons[4].doClick();

		assertNull(irrigator.getNextFailedDevice());
		assertTrue(simulation.getValve("V09").isOpen());

	}

	private void performStepFive() {
		// 5. Maintainer indicates which of the failed Valves and Sensors are
		// now repaired.

		buttons[2].doClick();
		buttons[2].doClick();
		
		// finished 
		buttons[7].doClick();

	}

	private void testStepTwo() {
		// 2. AquaLush displays the set of failed Valves and Sensors.

		try {
			assertTrue(extractField(uiController, "crntScrn") == extractField(
					uiController, "fixFailuresScrn"));
		} catch (Exception e) {
			e.printStackTrace();
			fail();
		}
	}

	public void performStepOne() {
		// 1. Maintainer begins a repair session.
		buttons[5].doClick();
	}

	public void prepareUseCase() {
		// In the third zone, a valve as well as a sensor
		// have failed.

		SimTime.instance().setHour(18);
		SimTime.instance().setSecond(0);

		s3 = null;
		v09 = null;

		for (JCheckBox cb : checkboxes)
			if (cb.getText().equals("S3"))
				s3 = cb;
			else if (cb.getText().equals("V09"))
				v09 = cb;

		s3.doClick();
		v09.doClick();

		// Change the current mode from automatic to manual
		buttons[3].doClick();
		// Go to the "Control Irrigation" screen, ...
		buttons[7].doClick();
		// ... select valve "V10", ...
		for (int i = 0; i < 10; i++)
			buttons[3].doClick();
		// ... open the valve ...
		buttons[2].doClick();
		// ... and propagate the change to the whole zone.
		buttons[4].doClick();

		// Leave the "Control Irrigation" screen.
		buttons[7].doClick();

		List<String> failedDevices = null;
		try {
			failedDevices = getFailedDevices(irrigator);
		} catch (Exception e) {
			e.printStackTrace();
			fail();
		}

		assertTrue(failedDevices.contains("V09"));
		assertTrue(failedDevices.contains("S3"));
		assertEquals(2, failedDevices.size());

		try {
			assertEquals("false", store.getData("V09.isWorking", "true"));
			assertEquals("false", store.getData("S3.isWorking", "true"));
		} catch (StoreFailureException e) {
			fail();
		}

	}

}
