package irrigation;

import java.util.HashMap;
import java.util.LinkedList;

import org.junit.Test;

import static org.junit.Assert.*;
import static tools.HelperFunctions.*;
import static irrigation.ScenarioHelper.*;
import simulation.SimTime;

/**
 * Tests the basic flow of AquaLush's seventh use case.
 */
public class UseCaseSevenBasicFlowTest extends UseCaseTest {

	public void prepareUseCase() {

		SimTime time = SimTime.instance();

		// Set water allocation to 600 to "speed up" the
		// whole irrigation cycle.

		// Open "Set up Automatic Irrigation", ...
		buttons[4].doClick();
		// ... select "Set Water Allocation" ...
		buttons[1].doClick();
		//
		// ... and change the allocation to 600.
		for (int i = 0; i < 5; i++)
			keys[9].doClick();

		keys[5].doClick();
		keys[10].doClick();
		keys[10].doClick();

		// Accept the settings
		buttons[7].doClick();
		buttons[7].doClick();

		time.setMinute(0);
		time.setSecond(0);
		time.setHour(2);

		clock.tick();
	}

	@Test
	public void executeUsecase() {
		prepareUseCase();
		testWaterAllocation();
		testIrrigation();
	}

	public void testIrrigation() {
		// For each active zone:

		HashMap<String, Integer> workingValves = new HashMap<String, Integer>();
		workingValves.put("Z1", 4);
		workingValves.put("Z2", 4);
		workingValves.put("Z3", 4);

		LinkedList<String> activeZones = new LinkedList<String>();
		activeZones.add("Z1");
		activeZones.add("Z2");
		activeZones.add("Z3");
		
		int waterUsed = 0;

		Zone currentZone;
		IrrigationCycle currentCycle;
		int zoneCount = 0;

		try {
			while ((currentCycle = extractField(irrigator, "cycle")) != null
					&& (currentZone = extractField(currentCycle, "crntZone")) != null) {

				zoneCount++;

				// 5b The zone allocation is not exhausted:
				// 5b1. AquaLush adds
				// the remainder of the zone allocation to the zone allocations
				// of the unirrigated active zones and divides this sum by the
				// number of Valves in the unirrigated active zones to compute a
				// new Valve allocation.
				// 5b2. AquaLush recomputes the zone
				// allocations for each active unirrigated zone by multiplying
				// the new Valve allocation by the number of Valves in the zone.

				int numberOfValves = 0;
				for (String zone : activeZones) {
					numberOfValves += workingValves.get(zone);
				}

				assertEquals(
						(int) (((double) (600 - waterUsed) / numberOfValves) * workingValves
								.get(currentZone.getID())),
						extractField(currentZone, "allocation"));

				// 3. AquaLush opens all the (working) Valves in the zone, and
				// marks the time when it does so.
				for (Valve v : getValves(irrigator, currentZone.getID()))
					assertTrue(v.isOpen());

				if (currentZone.getID().equals("Z2")) {
					for (int i = 0; i < 5; i++)
						clock.tick();

					simulation.getSensor(currentZone.getID().replace("Z", "S"))
							.setLevel(50);

					clock.tick();
				} else {
					int availableAllocation = extractField(currentZone,
							"allocation");
					int usedAllocation = currentZone.getWaterUsed();
					while (currentZone.getWaterUsed() < availableAllocation) {
						clock.tick();

						// 4. Every minute, AquaLush reads the zone Sensor and
						// also computes the
						// zone water usage by multiplying the total flow rate
						// for all Valves in
						// the zone by the time that the Valves have been open.
						assertTrue(usedAllocation < currentZone.getWaterUsed());
						usedAllocation = currentZone.getWaterUsed();
					}
				}

				// 5. If the zones critical moisture level is reached, or the
				// zone allocation is reached, then AquaLush closes all the
				// Valves in the zone.
				for (Valve v : getValves(irrigator, currentZone.getID()))
					assertFalse(v.isOpen());

				activeZones.remove(currentZone.getID());
				waterUsed += currentZone.getWaterUsed();
			}
		} catch (Exception e) {
			e.printStackTrace();
			fail();
		}

		assertEquals(3, zoneCount);

		for (Zone z : zones) {
			assertTrue(irrigator.getWaterUsed() < 624);
			try {
				for (Valve v : getValves(irrigator, z.getID()))
					assertFalse(z.getID() + ":" + v.getID(), v.isOpen());
			} catch (Exception e) {
				e.printStackTrace();
				fail();
			}
		}

	}

	public void testWaterAllocation() {
		// 2. AquaLush counts the (working) Valves in the active zones and
		// divides the water allocation by this count to get a Valve allocation.
		// Each zone is assigned a zone allocation, which is the Valve
		// allocation multiplied by the number of Valves in the zone.

		for (Zone z : zones)
			try {
				assertEquals(200, extractField(z, "allocation"));
			} catch (Exception e) {
				e.printStackTrace();
				fail();
			}

	}

}
