package device;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.fail;
import static tools.HelperFunctions.extractFieldAsField;
import static tools.HelperFunctions.extractStaticField;
import static tools.HelperFunctions.extractStaticFieldAsField;

import java.lang.reflect.Field;
import java.util.Observable;
import java.util.Observer;

import org.junit.Before;
import org.junit.Test;

import util.Day;

public class ClockTest extends TickListenerTest {

	@Test
	public void testSetClockDevice() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests setClockDevice().
		
		MockClockDevice anotherMock = new MockClockDevice();

		// Since it is possible to set the ClockDevice at any time
		// the following behavior is expected
		Clock.setClockDevice(anotherMock);

		assertEquals(anotherMock, extractStaticField(Clock.class, "works"));
		assertEquals(clock, anotherMock.listener);

		try {
			Clock.setClockDevice(null);
			fail();
		} catch (IllegalArgumentException e) {
		}
	}

	protected MockClockDevice mock;
	protected Clock clock;

	@Before
	public void init() throws IllegalArgumentException, IllegalAccessException,
			SecurityException, NoSuchFieldException {
		// Reset the clock.
		Field clockField = extractStaticFieldAsField(Clock.class, "instance");
		clockField.setAccessible(true);
		clockField.set(null, null);

		mock = new MockClockDevice();
		Clock.setClockDevice(mock);
		clock = Clock.instance();
		
		ticks = 0;

		assertEquals(clock, mock.listener);
	}

	@Test
	public void testInstance() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests whether Clock.instance always returns the same object.
		assertEquals(clock, Clock.instance());
		assertNotNull(clock);
	}

	@Test
	public void testInstanceWithoutClockDevice()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// If one tries to access a Clock-device
		// without specifying the ClockDevice first,
		// an Exception other than NullPointerException
		// should be thrown.

		Field clockField = extractFieldAsField(clock, "works");
		clockField.setAccessible(true);
		clockField.set(clock, null);

		try {
			Clock.instance();
			fail();
		} catch (NullPointerException e) {
			fail();
		} catch (Exception e) {

		}

	}

	@Test
	public void testTimeProperty() {
		// Tests a Clock's time property
		clock.setTime(0000);
		assertEquals(0000, clock.getTime());
		assertEquals(clock.getTime(), mock.getTime());

		clock.setTime(2359);
		assertEquals(2359, clock.getTime());
		assertEquals(clock.getTime(), mock.getTime());
	}

	@Test
	public void testTimePropertyIllegalCases() {
		// Tests setTime() with an illegal military time.
		clock.setTime(2359);
		
		try {
			clock.setTime(-1);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			clock.setTime(2400);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			clock.setTime(1775);
			fail();
		} catch (IllegalArgumentException e) {

		}

		// The Clock's initial state should not be changed
		assertEquals(2359, clock.getTime());
		assertEquals(clock.getTime(), mock.getTime());

	}

	@Test
	public void testDayProperty() {
		// Tests the day property
		clock.setDay(Day.MONDAY);
		assertEquals(Day.MONDAY, clock.getDay());
		assertEquals(clock.getDay(), mock.getDay());

		clock.setDay(Day.TUESDAY);
		assertEquals(Day.TUESDAY, clock.getDay());
		assertEquals(clock.getDay(), mock.getDay());
	}

	@Test
	public void testDayPropertyNull() {
		// Tests setDay with the argument null.
		clock.setDay(Day.TUESDAY);
		
		try {
			clock.setDay(null);
			fail();
		} catch (IllegalArgumentException e) {

		}

		assertEquals(Day.TUESDAY, clock.getDay());
		assertEquals(clock.getDay(), mock.getDay());

	}

	private static int ticks;

	@Test
	public void testTick() {
		// Tests tick().
		clock.addObserver(new Observer() {

			@Override
			public void update(Observable o, Object arg) {
				ticks++;
			}
		});

		clock.tick();

		// The observer should be notified that
		// another tick happend.
		assertEquals(1, ticks);

	}

}
