package device;

import static org.junit.Assert.*;

import org.junit.Test;

public abstract class DisplayDeviceTest {

	// These methods must be implemented by a concrete Test
	// of a DisplayDevice.
	protected abstract DisplayDevice getOut();
	protected abstract boolean displayCleared(DisplayDevice dev);
	protected abstract boolean notHighlighted(DisplayDevice dev);
	protected abstract boolean lineCleared(DisplayDevice dev, int line);
	protected abstract boolean isHighlighted(DisplayDevice dev, int line,
			int start, int end);
	protected abstract String lineContent(DisplayDevice dev, int line);

	@Test
	public void testClear() {
		// Tests clear().
		DisplayDevice dev = getOut();

		dev.clear();

		assertTrue(displayCleared(dev));

		dev.writeLine(1, "This is a test");

		dev.clear();

		assertTrue(displayCleared(dev));

	}

	@Test
	public void testClearHighlight() {
		// Tests clearHighlight().
		DisplayDevice dev = getOut();

		dev.clearHighlight();

		assertTrue(notHighlighted(dev));

		dev.writeLine(1, "This is a test");
		dev.highlight(1, 0, 8);
		dev.clearHighlight();

		assertTrue(notHighlighted(dev));

	}

	@Test
	public void testClearLine() {
		// Tests clearLine().
		DisplayDevice dev = getOut();

		dev.writeLine(1, "This is a test");
		dev.clearLine(1);

		assertTrue(lineCleared(dev, 1));
	}

	@Test
	public void testClearLineIllegalArguments() {
		// Tests whether clearLine() checks
		// its preconditions.
		DisplayDevice dev = getOut();
		
		try {
			dev.clearLine(-1);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			dev.clearLine(DisplayDevice.NUM_LINES);
			fail();
		} catch (IllegalArgumentException e) {

		}
	}

	@Test
	public void testHighlight() {
		// Tests highlight().
		DisplayDevice dev = getOut();

		dev.writeLine(0, "This is a test");
		dev.highlight(0, 0, 7);

		assertTrue(isHighlighted(dev, 0, 0, 7));

		dev.clearHighlight();
		dev.highlight(0, 4, 4);

		assertTrue(isHighlighted(dev, 0, 4, 4));

		dev.clearHighlight();
		dev.highlight(0, 4, DisplayDevice.NUM_COLS - 3);

		assertTrue(isHighlighted(dev, 0, 4, DisplayDevice.NUM_COLS - 4));
	}

	@Test
	public void testHighlightIllegalArguments() {
		// Tests highlight() with illegal arguments.
		
		DisplayDevice dev = getOut();

		try {
			dev.highlight(-1, 0, 8);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			dev.highlight(DisplayDevice.NUM_LINES, 0, 8);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			dev.highlight(0, -1, 8);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			dev.highlight(0, DisplayDevice.NUM_COLS, 8);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			dev.highlight(0, 0, -1);
			fail();
		} catch (IllegalArgumentException e) {

		}

	}

	@Test
	public void testWriteLine() {
		// Tests writeLine().
		DisplayDevice dev = getOut();
		String blankLine = "                                        ";
		String testString = "This is a test";

		dev.writeLine(DisplayDevice.NUM_LINES - 2, testString);

		assertEquals(
				(testString + blankLine).substring(0, DisplayDevice.NUM_COLS),
				lineContent(dev, DisplayDevice.NUM_LINES - 2));
	}
		
	@Test
	public void testWriteLineTooLongLine() {
		// Tests writeLine() with a line of text
		// containing more than 40 characters.
		String testString = "This is a text with more than 40 Characters on a line";
		DisplayDevice dev = getOut();
		dev.writeLine(DisplayDevice.NUM_LINES - 2, testString);

		assertEquals(
				testString.substring(0, DisplayDevice.NUM_COLS),
				lineContent(dev, DisplayDevice.NUM_LINES - 2));
	}
	
	@Test
	public void testWriteLineIllegalArguments() {
		// Tests writeLine() with illegal parameters.
		DisplayDevice dev = getOut();

		try {
			dev.writeLine(-1, "Test");
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			dev.writeLine(DisplayDevice.NUM_LINES, "Test");
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			dev.writeLine(1, null);
			fail();
		} catch (IllegalArgumentException e) {

		}

	}

}
