package irrigation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static tools.HelperFunctions.extractField;

import java.util.Iterator;
import java.util.Set;

import org.junit.Test;

import device.DeviceFailureException;

public class AutoCycleTest extends IrrigationCycleTest {

	@Override
	protected IrrigationCycle concreteOut(Set<Zone> zones, int theAllocation) {
		return new AutoCycle(zones, theAllocation);
	}

	protected final AutoCycle getOut(Set<Zone> zones, int theAllocation) {
		return (AutoCycle) concreteOut(zones, theAllocation);
	}

	// The following methods test AutoCycle's tick() method

	@Test
	public void testTickSingleZone() throws DeviceFailureException {
		// Tests whether a cycle irrigates a zone
		// correctly.

		cycle = getOut(zones, 1500);

		cycle.start();

		for (int i = 0; i < 10; i++) {
			cycle.tick();
			assertEquals((i + 1) * 15, firstZone.getWaterUsed());
			assertEquals(firstZone.getWaterUsed(), cycle.getWaterUsed());
		}

	}

	@Test
	public void testTickValveFailureDuringIrrigation()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests if tick() stays in the current zone
		// when a valve fails.

		cycle = getOut(zones, 1500);

		cycle.start();

		for (int i = 0; i < 3; i++) {
			cycle.tick();
		}

		// Let a valve fail
		Valve firstValve = zoneValves.get(firstZone).iterator().next();
		MockValveDevice valveDevice = extractField(firstValve, "device");
		valveDevice.isFailed = true;

		int waterUsed = firstZone.getWaterUsed();

		// Since the first zone is not irrigated sufficiently,
		// it should still be irrigated
		for (int i = 0; i < 3; i++)
			cycle.tick();

		assertEquals(waterUsed + 45, firstZone.getWaterUsed());
	}

	@Test
	public void testTickMoistureLevelReached() throws DeviceFailureException {
		// Tests how a cycle behaves when all zones
		// are sufficiently irrigated

		cycle = getOut(zones);

		// Irrigate the zones a few times

		cycle.start();

		for (int i = 0; i < 5; i++)
			cycle.tick();

		// Assume the moisture level of all zones has changed (perhaps because
		// of a heavy rain ...)
		for (Zone z : zones)
			zoneSensor.get(z).value = 55;

		cycle.tick();

		// Get water used so far
		int oldWaterUsed = 0;
		for (Zone z : zones) {
			oldWaterUsed += z.getWaterUsed();
		}

		// Irrigate some additional minutes
		for (int i = 0; i < 5; i++)
			cycle.tick();

		int newWaterUsed = 0;
		for (Zone z : zones) {
			newWaterUsed += z.getWaterUsed();
		}

		assertEquals(oldWaterUsed, newWaterUsed);

	}

	@Test
	public void testTickAllocationReached() throws DeviceFailureException {

		// Tests whether no more water is irrigated when
		// a cycle's water allocation is used completely

		cycle = getOut(zones, 200);

		cycle.start();
		for (int i = 0; i < 13; i++)
			cycle.tick();

		cycle.tick();

		// Make sure no further water is
		// irrigated when tick is called

		int oldWaterUsed = 0;
		for (Zone z : zones)
			oldWaterUsed += z.getWaterUsed();

		for (int i = 0; i < 5; i++) {
			cycle.tick();
		}

		int newWaterUsed = 0;
		for (Zone z : zones)
			newWaterUsed += z.getWaterUsed();

		assertEquals(oldWaterUsed, newWaterUsed);
	}

	@Test
	public void testTickZoneChange() throws DeviceFailureException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests whether a cycle properly changes from
		// a zone to the next in case the current zone is done

		cycle = getOut(zones, 500);

		cycle.start();
		for (int i = 0; i < 5; i++)
			cycle.tick();

		// In the past minute, the first zone got sufficiently irrigated
		zoneSensor.get(firstZone).value = 55;

		cycle.tick();

		// The first zone is sufficiently irrigated
		int firstZoneWaterUsed = firstZone.getWaterUsed();

		// Since the first zone has been irrigated for 6 minutes
		// there are 410 water units left
		assertEquals(205, extractField(secondZone, "allocation"));
		assertEquals(205, extractField(thirdZone, "allocation"));

		for (int i = 0; i < 4; i++)
			cycle.tick();

		assertEquals(firstZoneWaterUsed, firstZone.getWaterUsed());
		assertEquals(60, secondZone.getWaterUsed());

	}

	@Test
	public void testTickZoneChangeAllocationUsed()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests whether a cycle properly changes from
		// one zone to another in case there is no allocation left

		cycle = getOut(zones, 600);

		cycle.start();
		for (int i = 0; i < 14; i++)
			cycle.tick();

		int firstZoneWaterUsed = firstZone.getWaterUsed();

		assertEquals(195, extractField(secondZone, "allocation"));
		assertEquals(195, extractField(thirdZone, "allocation"));

		for (int i = 0; i < 4; i++)
			cycle.tick();

		assertEquals(firstZoneWaterUsed, firstZone.getWaterUsed());
		assertEquals(60, secondZone.getWaterUsed());

	}

	@Test
	public void testTickSensorDeviceFailure() throws DeviceFailureException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests whether a cycle detects failures of SensorDevices

		cycle = getOut(zones, 500);

		cycle.start();

		for (int i = 0; i < 5; i++)
			cycle.tick();

		try {
			for (int i = 0; i < 10; i++) {
				// After a given time, cycle should proceed
				// to the second zone for irrigation
				if (i == 5)
					zoneSensor.get(firstZone).value = 55;
				if (i == 8)
					secondZone.addSensor("s0", new MockSensorDevice(5), "l0");
				cycle.tick();
			}
			fail();
		} catch (DeviceFailureException e) {
			assertEquals(1, e.getDevices().size());
			assertTrue(e.getDevices().contains("s0"));
		}

		assertEquals(45, secondZone.getWaterUsed());
		assertEquals(290, extractField(thirdZone, "allocation"));

		// When a sensor fails, the next zone should be
		// irrigated.

		for (int i = 0; i < 3; i++)
			cycle.tick();

		assertEquals(45, secondZone.getWaterUsed());
		assertEquals(45, thirdZone.getWaterUsed());

	}

	@Test
	public void testTickValveDeviceFailureLastZoneException()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests whether a cycle detects failures of ValveDevices
		// in the zone just finished.

		cycle = getOut(zones, 500);

		cycle.start();

		for (int i = 0; i < 3; i++)
			cycle.tick();

		// Setup two failed valves
		Iterator<Valve> firstZoneIterator = zoneValves.get(firstZone)
				.iterator();
		Valve firstValve = firstZoneIterator.next();
		Valve secondValve = firstZoneIterator.next();
		((MockValveDevice) extractField(firstValve, "device")).isFailed = true;
		((MockValveDevice) extractField(secondValve, "device")).isFailed = true;

		try {
			for (int i = 0; i < 15; i++) {
				// After some irrigation, the first zone is done
				if (i == 5)
					zoneSensor.get(firstZone).value = 55;
				cycle.tick();
			}
			fail();
		} catch (DeviceFailureException e) {
			assertEquals(2, e.getDevices().size());
			assertTrue(e.getDevices().contains(firstValve.getID()));
			assertTrue(e.getDevices().contains(secondValve.getID()));
		}
	}

	@Test
	public void testTickValveDeviceFailureLastZone()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests whether a cycle detects failures of ValveDevices
		// in the zone just finished and starts irrigating the next zone.

		cycle = getOut(zones, 500);

		cycle.start();

		for (int i = 0; i < 3; i++)
			cycle.tick();

		// Setup two failed valves
		Iterator<Valve> firstZoneIterator = zoneValves.get(firstZone)
				.iterator();
		Valve firstValve = firstZoneIterator.next();
		Valve secondValve = firstZoneIterator.next();
		((MockValveDevice) extractField(firstValve, "device")).isFailed = true;
		((MockValveDevice) extractField(secondValve, "device")).isFailed = true;

		try {
			for (int i = 0; i < 15; i++) {
				// After some irrigation, the first zone is done
				if (i == 5)
					zoneSensor.get(firstZone).value = 55;
				cycle.tick();
			}
			fail();
		} catch (DeviceFailureException e) {
		}

		// The next zone should be irrigated
		// normally.

		assertEquals(182, extractField(secondZone, "allocation"));
		assertEquals(182, extractField(thirdZone, "allocation"));

		for (int i = 0; i < 5; i++)
			cycle.tick();

		assertEquals(75, secondZone.getWaterUsed());
	}

	@Test
	public void testTickValveDeviceFailureNextZoneException()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests whether a cycle detects failures of ValveDevices
		// in the next zone.

		cycle = getOut(zones, 500);

		cycle.start();

		for (int i = 0; i < 3; i++)
			cycle.tick();

		// Set up two failed valves
		Iterator<Valve> secondZoneIterator = zoneValves.get(secondZone)
				.iterator();
		Valve firstValve = secondZoneIterator.next();
		Valve secondValve = secondZoneIterator.next();
		((MockValveDevice) extractField(firstValve, "device")).isFailed = true;
		((MockValveDevice) extractField(secondValve, "device")).isFailed = true;

		try {
			for (int i = 0; i < 15; i++) {
				// After some irrigation, the first zone is done
				if (i == 5)
					zoneSensor.get(firstZone).value = 55;
				cycle.tick();
			}
			fail();
		} catch (DeviceFailureException e) {
			assertEquals(2, e.getDevices().size());
			assertTrue(e.getDevices().contains(firstValve.getID()));
			assertTrue(e.getDevices().contains(secondValve.getID()));
		}
	}

	@Test
	public void testTickValveDeviceFailureNextZone()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests whether a cycle detects failures of ValveDevices
		// in the next zone.

		cycle = getOut(zones, 500);

		cycle.start();

		for (int i = 0; i < 3; i++)
			cycle.tick();

		// Setup two failed valves
		Iterator<Valve> secondZoneIterator = zoneValves.get(secondZone)
				.iterator();
		Valve firstValve = secondZoneIterator.next();
		Valve secondValve = secondZoneIterator.next();
		((MockValveDevice) extractField(firstValve, "device")).isFailed = true;
		((MockValveDevice) extractField(secondValve, "device")).isFailed = true;

		try {
			for (int i = 0; i < 15; i++) {
				// After some irrigation, the first zone is done
				if (i == 5)
					zoneSensor.get(firstZone).value = 55;
				cycle.tick();
			}
			fail();
		} catch (DeviceFailureException e) {
		}

		// Allocations should be adjusted
		// Since the next zone has only one working valve, it only
		// gets 25% of the remaining allocation
		assertEquals(91, extractField(secondZone, "allocation"));
		assertEquals(273, extractField(thirdZone, "allocation"));

		// The next zone should be irrigated
		// normally.

		for (int i = 0; i < 5; i++)
			cycle.tick();

		assertEquals(25, secondZone.getWaterUsed());
	}

	@Test
	public void testTickValveDeviceFailureLastAndNextZone()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests whether a cycle detects failures of ValveDevices
		// in the current and in the next zone.

		cycle = getOut(zones, 500);

		cycle.start();

		for (int i = 0; i < 3; i++)
			cycle.tick();

		// Setup two failed valves in the current and
		// the next zone
		Iterator<Valve> secondZoneIterator = zoneValves.get(secondZone)
				.iterator();
		Valve firstValve = secondZoneIterator.next();
		Valve secondValve = secondZoneIterator.next();
		((MockValveDevice) extractField(firstValve, "device")).isFailed = true;
		((MockValveDevice) extractField(secondValve, "device")).isFailed = true;

		secondZoneIterator = zoneValves.get(firstZone).iterator();
		firstValve = secondZoneIterator.next();
		secondValve = secondZoneIterator.next();
		((MockValveDevice) extractField(firstValve, "device")).isFailed = true;
		((MockValveDevice) extractField(secondValve, "device")).isFailed = true;

		try {
			for (int i = 0; i < 15; i++) {
				// After some irrigation, the first zone is done
				if (i == 5)
					zoneSensor.get(firstZone).value = 55;
				cycle.tick();
			}
			fail();
		} catch (DeviceFailureException e) {
		}

		// Allocations should be adjusted
		// Since the next zone has only one working valve, it only
		// gets 25% of the remaining allocation
		assertEquals(91, extractField(secondZone, "allocation"));
		assertEquals(273, extractField(thirdZone, "allocation"));

		for (int i = 0; i < 5; i++)
			cycle.tick();

		assertEquals(25, secondZone.getWaterUsed());
	}

	@Test
	public void testTickValveDeviceFailureLastAndNextZoneException()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests whether tick() detects failures of ValveDevices
		// in the current and in the next zone.

		cycle = getOut(zones, 500);

		cycle.start();

		for (int i = 0; i < 3; i++)
			cycle.tick();

		// Setup two failed valves in the current and
		// the next zone
		Iterator<Valve> secondZoneIterator = zoneValves.get(secondZone)
				.iterator();
		Valve firstValve = secondZoneIterator.next();
		Valve secondValve = secondZoneIterator.next();
		((MockValveDevice) extractField(firstValve, "device")).isFailed = true;
		((MockValveDevice) extractField(secondValve, "device")).isFailed = true;

		secondZoneIterator = zoneValves.get(firstZone).iterator();
		firstValve = secondZoneIterator.next();
		secondValve = secondZoneIterator.next();
		((MockValveDevice) extractField(firstValve, "device")).isFailed = true;
		((MockValveDevice) extractField(secondValve, "device")).isFailed = true;

		try {
			for (int i = 0; i < 15; i++) {
				// After some irrigation, the first zone is done
				if (i == 5)
					zoneSensor.get(firstZone).value = 55;
				cycle.tick();
			}
			fail();
		} catch (DeviceFailureException e) {
			assertEquals(4, e.getDevices().size());
			assertTrue(e.getDevices().contains(firstValve.getID()));
			assertTrue(e.getDevices().contains(secondValve.getID()));
		}
	}

	// The following tests test the start() method of class AutoCycle

	@Test
	public void testStart() throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {
		// tests the ideal case where no errors happen

		cycle = getOut(zones, 1500);

		cycle.start();

		for (Zone z : zones)
			assertEquals(500, extractField(z, "allocation"));

		for (Valve v : zoneValves.get(firstZone))
			assertTrue(v.isOpen());

	}

	@Override
	public void testStartAfterEnd() throws DeviceFailureException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		super.testStartAfterEnd();

		assertEquals(45, firstZone.getWaterUsed());
		assertEquals(0, secondZone.getWaterUsed());
		assertEquals(0, thirdZone.getWaterUsed());

	}

	@Test
	public void testStartNoIrrigationNeeded() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {
		// the second zone does not need irrigation,
		// so it should be ignored.

		cycle = getOut(zones, 1500);

		zoneSensor.get(secondZone).value = 75;

		cycle.start();

		for (int i = 0; i < 10; i++) {
			if (i == 5)
				zoneSensor.get(firstZone).value = 55;
			cycle.tick();
		}

		assertTrue(firstZone.getWaterUsed() > 0);
		assertEquals(0, secondZone.getWaterUsed());
		assertTrue(thirdZone.getWaterUsed() > 0);

	}

	@Test
	public void testStartSensorDeviceFailure() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {

		// Tests whether start initializes a cycle properly
		// if a sensor device fails.

		cycle = getOut(zones, 1500);

		MockSensorDevice failedSensor = new MockSensorDevice(6);

		firstZone.addSensor("s0", failedSensor, firstZone.getLocation());

		try {
			cycle.start();
			fail();
		} catch (DeviceFailureException e) {
			assertEquals(1, e.getDevices().size());
			assertTrue(e.getDevices().contains("s0"));
		}

		assertEquals(750, extractField(secondZone, "allocation"));
		assertEquals(0, extractField(firstZone, "allocation"));
		assertEquals(750, extractField(thirdZone, "allocation"));

		for (Valve v : zoneValves.get(secondZone))
			assertTrue(v.isOpen());

		// Although start has thrown an exception,
		// it should still be possible to start the cycle
		for (int i = 0; i < 5; i++)
			cycle.tick();

		assertEquals(75, cycle.getWaterUsed());
		assertEquals(0, firstZone.getWaterUsed());
		assertEquals(75, secondZone.getWaterUsed());

	}

	@Test
	public void testStartValveDeviceFailure() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {

		// Tests start in the case a valve device fails

		cycle = getOut(zones, 1600);

		Valve failedValve;
		((MockValveDevice) extractField(failedValve = zoneValves.get(firstZone)
				.iterator().next(), "device")).isFailed = true;

		// There is a failed valve in the first zone
		// so an exception is expected.
		try {
			cycle.start();
			fail();
		} catch (DeviceFailureException e) {
			assertEquals(1, e.getDevices().size());
			assertTrue(e.getDevices().contains(failedValve.getID()));
		}

		for (Valve v : zoneValves.get(firstZone))
			assertTrue(v.isFailed() || v.isOpen());

		// Since the first zone has only two working valves,
		// its allocation should be adjusted
		assertEquals(400, extractField(firstZone, "allocation"));
		assertEquals(600, extractField(secondZone, "allocation"));
		assertEquals(600, extractField(thirdZone, "allocation"));

		// Although start has thrown an exception,
		// it should still be possible to start the cycle
		for (int i = 0; i < 5; i++)
			cycle.tick();

		assertEquals(50, firstZone.getWaterUsed());
	}

	@Test
	public void testStartValveDeviceFailureSetup()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {

		// Tests start in the case a valve device fails
		// in one of the later zones

		cycle = getOut(zones, 1500);

		MockValveDevice failedValve = new MockValveDevice();
		failedValve.isFailed = true;
		secondZone.addValve("v41", failedValve, "t4", 5, "loc");

		// There is no failed valve in the first zone, so no
		// exception is expected since no attempt was made
		// to open one of the failed valves.
		cycle.start();

		// There are 10 valves which currently seem to work
		// so each one gets 150 water units
		for (Zone z : zones)
			assertEquals(150 * z.getNumWorkingValves(),
					extractField(z, "allocation"));
	}

	// The following tests test AutoCycle's isDone() method

	@Test
	public void testIsDoneAllocationReached() throws DeviceFailureException {
		// Make sure that an AutoCycle behaves properly when
		// there's no water left

		cycle = getOut(zones, 200);

		cycle.start();
		for (int i = 0; i < 13; i++)
			cycle.tick();

		assertFalse(cycle.isDone());

		cycle.tick();

		assertTrue(cycle.isDone());
	}

	@Test
	public void testIsDoneMoistureLevelReached() throws DeviceFailureException {
		// Tests what happens when every zone has reached
		// its critical moisture level

		cycle = getOut(zones, 400);

		cycle.start();

		cycle.tick();
		cycle.tick();

		assertFalse(cycle.isDone());

		for (Zone z : zones)
			zoneSensor.get(z).value = 50;

		// The following ticks are necessary because
		// the cycle has to realize it is done
		cycle.tick();
		cycle.tick();
		cycle.tick();

		assertTrue(cycle.isDone());
	}

	@Test
	public void testIsDoneMoistureLevelAndAllocationReached()
			throws DeviceFailureException {
		// Make sure that an AutoCycle stops
		// irrigation when the moisture level
		// is reached for all zones and there's
		// not water allocation left.

		cycle = getOut(zones, 135);

		cycle.start();

		for (Zone z : zones) {
			for (int i = 0; i < 3; i++)
				cycle.tick();
			zoneSensor.get(z).value = 60;
		}

		cycle.tick();

		assertTrue(cycle.isDone());

	}

	@Test
	public void testIsDoneNoIrrigationNeeded() throws DeviceFailureException {
		// Tests isDone() in a situation where
		// no zone needs irrigation

		cycle = getOut(zones, 500);

		for (Zone z : zones)
			zoneSensor.get(z).value = 55;

		cycle.start();

		assertTrue(cycle.isDone());

	}

	@Test
	public void testIsDoneAfterEnd() throws DeviceFailureException {
		// Tests isDone() after a call to
		// end

		cycle = getOut(zones, 500);

		cycle.start();

		assertFalse(cycle.isDone());

		cycle.end();

		assertTrue(cycle.isDone());

		cycle.start();
	}

	@Test
	public void testIsDoneZoneChange() throws DeviceFailureException {
		// Tests isDone() when one of the zones
		// is done
		cycle = getOut(zones, 500);

		cycle.start();

		for (int i = 0; i < 3; i++)
			cycle.tick();

		zoneSensor.get(firstZone).value = 55;

		cycle.tick();

		assertFalse(cycle.isDone());
	}

	@Test
	public void testIsDoneNotStarted() {
		cycle = getOut(zones, 100);

		assertTrue(cycle.isDone());
	}

	// The following code tests AutoCycle's end() method

	@Override
	@Test
	public void testEndWithoutStart() throws DeviceFailureException {
		super.testEndWithoutStart();

		assertTrue(cycle.isDone());
	}

	@Override
	@Test
	public void testEndAllZonesDone() throws DeviceFailureException {
		super.testEndAllZonesDone();

		assertTrue(cycle.isDone());
	}

	@Test
	public void testEndSomeZonesDone() throws DeviceFailureException {
		// Tests how end() works in the middle of an irrigation cycle

		cycle = getOut(zones, 500);

		cycle.start();

		for (int i = 0; i < 10; i++) {
			cycle.tick();
			if (i == 5)
				zoneSensor.get(firstZone).value = 55;
		}

		// Should work
		cycle.end();

		for (int i = 0; i < 5; i++)
			cycle.tick();

		assertEquals(45, secondZone.getWaterUsed());

		// All Valves should be closed
		for (Zone z : zones)
			for (Valve v : zoneValves.get(z))
				assertFalse(v.isOpen());

		assertTrue(cycle.isDone());
	}

	@Override
	@Test
	public void testEndSomeZonesDoneSensorFailure()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		super.testEndSomeZonesDoneSensorFailure();

		assertEquals(45, secondZone.getWaterUsed());
		assertTrue(cycle.isDone());
	}

	// The following code tests AutoCycle's constructor

	@Test
	public void testConstructorAllocationSimpleUse()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		cycle = getOut(zones, 1000);

		assertEquals(1000, extractField(cycle, "allocation"));
	}

	@Test
	public void testConstructorAllocationHugeAllocation()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		cycle = getOut(zones, Integer.MAX_VALUE);

		assertEquals(Integer.MAX_VALUE, extractField(cycle, "allocation"));
	}

	@Test
	public void testConstructorAllocationNegativeValues() {
		// A negative allocation should not be allowed (since
		// it doesn't make sense)

		try {
			cycle = getOut(zones, -1);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			cycle = getOut(zones, Integer.MIN_VALUE);
			fail();
		} catch (IllegalArgumentException e) {

		}
	}

	// The following code tests setAllocation

	@Test
	public void testSetAllocationNegativeArguments()
			throws DeviceFailureException {
		// Tests how setAllocation behaves if it receives a
		// negative argument

		cycle = getOut(zones, 600);

		try {
			((AutoCycle) cycle).setAllocation(-1);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			((AutoCycle) cycle).setAllocation(Integer.MIN_VALUE);
			fail();
		} catch (IllegalArgumentException e) {

		}

	}

	@Test
	public void testSetAllocationnNotRunningCycle()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests how setAllocation behaves if it is called
		// on a cycle that isn't running

		cycle = getOut(zones, 600);

		((AutoCycle) cycle).setAllocation(800);

		assertEquals(600, extractField(cycle, "allocation"));

	}

	@Test
	public void testSetAllocationnEndedCycle() throws DeviceFailureException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests how setAllocation behaves if it is called
		// on a cycle that is no longer running

		cycle = getOut(zones, 600);

		cycle.start();
		cycle.tick();
		cycle.end();

		((AutoCycle) cycle).setAllocation(800);

		assertEquals(600, extractField(cycle, "allocation"));

	}

	@Test
	public void testSetAllocationnRunningCycle() throws DeviceFailureException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests how setAllocation behaves if it is called
		// on a running cycle

		cycle = getOut(zones, 600);

		cycle.start();

		for (int i = 0; i < 3; i++)
			cycle.tick();

		((AutoCycle) cycle).setAllocation(800);

		assertEquals(800, extractField(cycle, "allocation"));

		for (Zone z : zones)
			assertEquals(251, extractField(z, "allocation"));

		for (int i = 0; i < 3; i++)
			cycle.tick();

		assertEquals(90, cycle.getWaterUsed());

	}

	@Test
	public void testSetAllocationnRunningCycleNegativeArgument()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests how setAllocation behaves if its preconditions
		// are violated.

		cycle = getOut(zones, 600);

		cycle.start();

		for (int i = 0; i < 3; i++)
			cycle.tick();

		try {
			((AutoCycle) cycle).setAllocation(-800);
			fail();
		} catch (IllegalArgumentException e) {

		}

		assertEquals(600, extractField(cycle, "allocation"));

		for (Zone z : zones)
			assertEquals(200, extractField(z, "allocation"));

		for (int i = 0; i < 3; i++)
			cycle.tick();

		assertEquals(90, cycle.getWaterUsed());

	}

	@Test
	public void testSetAllocationnRunningCycleTooSmallArgument()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests how setAllocation behaves if the argument
		// is so small that irrigation has to be stopped immediately

		cycle = getOut(zones, 600);

		cycle.start();

		for (int i = 0; i < 30; i++)
			cycle.tick();

		((AutoCycle) cycle).setAllocation(0);

		assertTrue(cycle.isDone());

		for (int i = 0; i < 3; i++)
			cycle.tick();

		assertEquals(450, cycle.getWaterUsed());

	}

	@Test
	public void testSetAllocationnFinishedCycle()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests how setAllocation behaves if it is called
		// on a cycle that is finished

		cycle = getOut(zones, 600);

		cycle.start();

		for (int i = 0; i < 40; i++)
			cycle.tick();

		((AutoCycle) cycle).setAllocation(800);

		assertTrue(cycle.isDone());
		assertEquals(600, extractField(cycle, "allocation"));

	}

	@Test
	public void testSetAllocationnRunningCycleFailingValves()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests how setAllocation behaves if there is a
		// failed valve in the current zone

		cycle = getOut(zones, 600);

		cycle.start();

		for (int i = 0; i < 3; i++)
			cycle.tick();

		Valve failedValve;
		((MockValveDevice) extractField(failedValve = zoneValves.get(firstZone)
				.iterator().next(), "device")).isFailed = true;

		try {
			((AutoCycle) cycle).setAllocation(800);
			fail();
		} catch (DeviceFailureException e) {
			assertEquals(1, e.getDevices().size());
			assertTrue(e.getDevices().contains(failedValve.getID()));
		}

		assertEquals(800, extractField(cycle, "allocation"));

		assertEquals(188, extractField(firstZone, "allocation"));
		assertEquals(283, extractField(secondZone, "allocation"));
		assertEquals(283, extractField(thirdZone, "allocation"));

		for (int i = 0; i < 3; i++)
			cycle.tick();

		assertEquals(90, cycle.getWaterUsed());

	}

	@Test
	public void testSetAllocationnRunningCycleFailingValvesOnOpen()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests how setAllocation behaves if a valve fails in the
		// next zone

		cycle = getOut(zones, 600);

		cycle.start();

		for (int i = 0; i < 3; i++)
			cycle.tick();

		Valve failedValve;
		((MockValveDevice) extractField(failedValve = zoneValves
				.get(secondZone).iterator().next(), "device")).isFailedOnOpen = true;

		try {
			((AutoCycle) cycle).setAllocation(800);
			fail();
		} catch (DeviceFailureException e) {
			assertEquals(1, e.getDevices().size());
			assertTrue(e.getDevices().contains(failedValve.getID()));
		}

		assertEquals(800, extractField(cycle, "allocation"));

		assertEquals(188, extractField(secondZone, "allocation"));
		assertEquals(283, extractField(firstZone, "allocation"));
		assertEquals(283, extractField(thirdZone, "allocation"));

		for (int i = 0; i < 3; i++)
			cycle.tick();

		assertEquals(90, cycle.getWaterUsed());

	}

	@Test
	public void testEndSomeZonesDoneValveFailure()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests how end() works in the middle of an irrigation cycle
		// if one of the current zone's valves fails

		cycle = getOut(zones);

		cycle.start();

		Valve failedValve = null;

		for (int i = 0; i < 10; i++) {
			cycle.tick();
			if (i == 5)
				zoneSensor.get(firstZone).value = 55;
			if (i == 9) {
				((MockValveDevice) extractField(
						failedValve = zoneValves.get(secondZone).iterator()
								.next(), "device")).isFailed = true;
				((MockValveDevice) extractField(zoneValves.get(thirdZone)
						.iterator().next(), "device")).isFailed = true;
			}
		}

		// Should work
		try {
			cycle.end();
			fail();
		} catch (DeviceFailureException e) {
			assertEquals(1, e.getDevices().size());
			assertTrue(e.getDevices().contains(failedValve.getID()));
		}

		for (int i = 0; i < 5; i++)
			cycle.tick();

		// All Valves should be closed
		for (Zone z : zones)
			for (Valve v : zoneValves.get(z))
				assertFalse(v.isOpen());

		assertTrue(cycle.isDone());
		assertEquals(45, secondZone.getWaterUsed());

	}

}
