package irrigation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static tools.HelperFunctions.extractField;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.junit.Before;
import org.junit.Test;

import device.DeviceFailureException;

public abstract class IrrigationCycleTest {

	protected HashSet<Zone> zones;
	protected HashMap<Zone, HashSet<Valve>> zoneValves;
	protected HashMap<Zone, MockSensorDevice> zoneSensor;
	protected Zone firstZone;
	protected Zone secondZone;
	protected Zone thirdZone;
	protected IrrigationCycle cycle;

	// Returns a copy of the current object under test
	protected final IrrigationCycle getOut(Set<Zone> zones) {
		return concreteOut(zones, Integer.MAX_VALUE);
	}

	protected abstract IrrigationCycle concreteOut(Set<Zone> zones,
			int theAllocation);

	@SuppressWarnings("unchecked")
	@Before
	public void init() throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Creates a set of zones, each containing
		// some valves and a sensor.
		zoneValves = new HashMap<Zone, HashSet<Valve>>();
		zones = new HashSet<Zone>();
		zoneSensor = new HashMap<Zone, MockSensorDevice>();
		for (int i = 0; i < 3; i++) {
			Zone z = new Zone("i" + i, "l" + i);
			z.setCriticalMoistureLevel(50);
			z.setAllocation(0);
			zones.add(z);

			for (int j = 0; j < 3; j++) {
				z.addValve("v" + j, new MockValveDevice(), "t" + i, 5,
						z.getLocation());
			}
			zoneValves.put(z, (HashSet<Valve>) extractField(z, "valves"));

			MockSensorDevice sensorDevice = new MockSensorDevice(0);
			sensorDevice.value = 25;
			zoneSensor.put(z, sensorDevice);
			z.addSensor("s0", sensorDevice, z.getLocation());
		}

		Iterator<Zone> iterator = zones.iterator();
		firstZone = iterator.next();
		secondZone = iterator.next();
		thirdZone = iterator.next();
	}

	@Test
	public void testIrrigationCycleConstructor() {
		// Tests the creation of an IrrigationCycle
		for (Zone z : zones)
			z.tick();

		cycle = getOut(zones);

		// A newly created cycle should not have used any
		// water so far.
		assertEquals(0, cycle.waterUsed);
		// Is each zone contained in the zones
		// the cycle is going to irrigate?
		assertEquals(zones, cycle.zones);
		for (Zone z : zones) {
			// The zones should all be known to the
			// cycle and they should not have any
			// water used so far
			assertTrue(cycle.zones.contains(z));
			assertEquals(0, z.getWaterUsed());
		}
	}

	@Test
	public void testIrrigationCycleConstructorNull() {
		// Should throw some sort of exception
		// since no proper set of zones was supplied
		try {
			cycle = getOut(null);
			fail();
		} catch (IllegalArgumentException e) {

		}
	}

	@Test
	public void testIrrigationCycleConstructorEmptySet() {
		// Should throw some sort of exception
		// since a cycle for an empty set of zones makes no sense
		try {
			cycle = getOut(new HashSet<Zone>());
			fail();
		} catch (IllegalArgumentException e) {

		}
	}

	@Test
	public void testTick() throws DeviceFailureException {
		// An irrigation cycle continues as long as
		// there is a zone which requires irrigation.

		cycle = getOut(zones);

		int oldWaterUsed = 0;

		for (Zone z : zones) {
			oldWaterUsed += z.getWaterUsed();
		}
		
		// Since tick() should have no effect, no water
		// should be used
		assertEquals(0, oldWaterUsed);

		cycle.start(); // Some subclasses of IrrigationCycle may
						// require this

		for (int i = 0; i < 15; i++)
			cycle.tick();
		cycle.end();

		int newWaterUsed = 0;
		for (Zone z : zones) {
			newWaterUsed += z.getWaterUsed();
		}

		assertTrue(newWaterUsed > oldWaterUsed);
		assertEquals(newWaterUsed, cycle.getWaterUsed());

	}

	@Test
	public void testEnd() throws DeviceFailureException,
			IllegalArgumentException, IllegalAccessException,
			SecurityException, NoSuchFieldException {
		// Independent of concrete implementation in subclasses,
		// after a call to end() no more water should be
		// irrigated by a cycle

		cycle = getOut(zones);

		cycle.start(); // Some subclasses of IrrigationCycle may
						// require this

		for (int i = 0; i < 16; i++)
			cycle.tick();
		cycle.end();

		int waterUsed = 0;
		for (Zone z : zones)
			waterUsed += z.getWaterUsed();

		cycle.tick();
		cycle.tick();

		int newWaterUsed = 0;
		for (Zone z : zones)
			newWaterUsed += z.getWaterUsed();

		assertEquals(waterUsed, newWaterUsed);

	}

	@Test
	public void testEndSensorFailure() throws DeviceFailureException {
		// end() should not be influenced by a failed sensor

		cycle = getOut(zones);

		cycle.start(); // Some subclasses of IrrigationCycle may
						// require this

		secondZone.addSensor("s0", new MockSensorDevice(5), "l0");

		for (int i = 0; i < 5; i++)
			cycle.tick();

		// No exception expected
		cycle.end();

	}

	@Test
	public void testEndDeviceFailure() throws DeviceFailureException {
		// If there is a failure in one of
		// the ValveDevice's, each implementation
		// of end() must report this failure

		cycle = getOut(zones);

		// Insert a new Valve into one of the zones
		MockValveDevice failedDevice = new MockValveDevice();
		firstZone.addValve("v4", failedDevice, "t4", 5, "loc");

		cycle.start(); // Some subclasses of IrrigationCycle may
						// require this

		for (int i = 0; i < 16; i++)
			cycle.tick();
		// Now, let the device fail
		failedDevice.isFailed = true;
		for (int i = 0; i < 4; i++)
			cycle.tick();

		try {
			cycle.end();
			fail();
		} catch (DeviceFailureException e) {
			assertEquals(1, e.getDevices().size());
			assertTrue(e.getDevices().contains("v4"));
		}

	}

	@Test
	public void testStartAfterEnd() throws DeviceFailureException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests whether start() also
		// works after a call to end()

		cycle = getOut(zones);

		// Start the cycle the first time
		cycle.start();
		cycle.tick();
		cycle.tick();

		// End the first cycle
		cycle.end();

		int waterUsed = 0;
		for (Zone z : zones)
			waterUsed += z.getWaterUsed();

		// Start the cycle again
		cycle.start();

		cycle.tick();

		int newWaterUsed = 0;
		for (Zone z : zones)
			newWaterUsed += z.getWaterUsed();

		assertTrue(waterUsed < newWaterUsed);
	}

	@Test
	public void testEndWithoutStart() throws DeviceFailureException {
		// Tests how end() works if start() was never called

		cycle = getOut(zones);

		// Should work
		cycle.end();

		// All Valves should be closed
		for (Zone z : zones)
			for (Valve v : zoneValves.get(z))
				assertFalse(v.isOpen());
	}

	@Test
	public void testEndAllZonesDone() throws DeviceFailureException {
		// Tests how end() works if all zones are done

		cycle = getOut(zones);

		cycle.start();

		for (int i = 0; i < 15; i++) {
			cycle.tick();
			if (i == 5)
				zoneSensor.get(firstZone).value = 55;
			if (i == 10)
				zoneSensor.get(secondZone).value = 55;
		}

		zoneSensor.get(thirdZone).value = 55;

		cycle.tick();

		// Should work
		cycle.end();

		// All Valves should be closed
		for (Zone z : zones)
			for (Valve v : zoneValves.get(z))
				assertFalse(v.isOpen());

	}

	@Test
	public void testEndSomeZonesDoneSensorFailure()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests how end() works in the middle of an irrigation cycle
		// when a sensor suddenly fails.

		cycle = getOut(zones);

		cycle.start();

		for (int i = 0; i < 10; i++) {
			cycle.tick();
			if (i == 5)
				zoneSensor.get(firstZone).value = 55;
			if (i == 9)
				secondZone.addSensor("s0", new MockSensorDevice(6),
						secondZone.getLocation());
		}

		// Should work
		cycle.end();

		for (int i = 0; i < 5; i++)
			cycle.tick();

		// All Valves should be closed
		for (Zone z : zones)
			for (Valve v : zoneValves.get(z))
				assertFalse(v.isOpen());

	}

}
