package irrigation;

import static irrigation.ScenarioHelper.getFailedDevices;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertSame;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static tools.HelperFunctions.assertObserved;
import static tools.HelperFunctions.extractField;
import static tools.HelperFunctions.extractFieldAsField;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.junit.Before;
import org.junit.Test;

import simulation.SimTime;
import startup.StartupException;
import util.Day;
import device.Clock;
import device.DeviceFailureException;
import device.MockStorageDevice;
import device.StoreFailureException;
import device.sim.MockSimClockDevice;

public class IrrigatorTest {

	private MockStorageDevice store;
	private Irrigator i;
	private Clock clock;
	private HashSet<Zone> zones;
	private ArrayList<String> failedDevices;

	@Before
	public void init() throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, StartupException {
		// Set up an irrigator ...

		// Reset the clock.
		Clock.setClockDevice(new MockSimClockDevice());
		Clock.instance().deleteObservers();

		store = new MockStorageDevice();
		i = new Irrigator(store);
		for (int i = 0; i < 3; i++) {
			String zoneID = "Z" + i;
			String location = "locZ" + i;
			// ... by adding a set of zones, all containing ...
			this.i.addZone(zoneID, location);
			// ... sensors ...
			this.i.addSensor(zoneID, "S" + i, new MockSensorDevice(0), location);
			for (int j = 0; j < 3; j++)
				// ... and valves.
				this.i.addValve(zoneID, "V" + Integer.toString(i) + j,
						new MockValveDevice(), "T" + j, 100, location);

		}

		// Extract interesting objects.
		clock = extractField(i, "clock");
		zones = extractField(i, "zones");
		failedDevices = extractField(i, "failedDevices");

		assertObserved(Clock.instance(), i);
	}

	@SuppressWarnings("unchecked")
	@Test
	public void testIrrigator() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests whether a newly created irrigator is set up properly.
		i = new Irrigator(store);

		assertEquals(Mode.AUTOMATIC, i.getMode());
		assertEquals(store, extractField(i, "store"));
		assertEquals(0, i.getAllocation());
		assertEquals(0, i.getIrrigationTime());
		assertTrue(((HashSet<Day>) i.getIrrigationDays()).isEmpty());
		assertEquals(Clock.instance(), clock);
		assertTrue(((HashSet<Zone>) extractField(i, "zones")).isEmpty());
		assertNull(extractField(i, "cycle"));
		assertTrue(((ArrayList<String>) extractField(i, "failedDevices"))
				.isEmpty());
		assertFalse(i.isStoreFailure());
	}

	@Test
	public void testIrrigatorNull() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// According to the contract of Irrigator's constructor,
		// it should be possible to create an irrigator
		// without specifying an existing storage device.
		i = new Irrigator(null);
		assertNull(extractField(i, "store"));
	}

	@Test
	public void testRestoreState() throws StartupException,
			StoreFailureException {
		// Tests restoreState().

		HashSet<Day> days = new HashSet<Day>();
		days.add(Day.FRIDAY);
		days.add(Day.SUNDAY);

		// Write the old configuration to the store.
		store.setData("mode", "manual");
		store.setData("allocation", "100");
		store.setData("irrigationTime", "2200");
		store.setData("irrigationDays", "FSu");

		i = new Irrigator(store);
		i.restoreState();

		assertEquals(Mode.MANUAL, i.getMode());
		assertEquals(100, i.getAllocation());
		assertEquals(2200, i.getIrrigationTime());
		assertTrue(this.i.getIrrigationDays().contains(Day.FRIDAY));
		assertTrue(this.i.getIrrigationDays().contains(Day.SUNDAY));
	}

	@Test
	public void testRestoreStateEmptyStore() throws StartupException {
		// Tests restoreState(). The store is empty.

		i = new Irrigator(new MockStorageDevice());
		i.restoreState();

		// Irrigation parameters should be set to the
		// defaults.
		assertEquals(Mode.AUTOMATIC, i.getMode());
		assertEquals(10000, i.getAllocation());
		assertEquals(200, i.getIrrigationTime());
		for (int i = 0; i < 7; i++)
			assertTrue(this.i.getIrrigationDays().contains(Day.valueOf(i)));
	}

	@Test
	public void testRestoreStateNullStore() throws StartupException {
		// Tests restoreState(). The store is set to null.

		// No exception expected (based on the pre-/postconditions
		// of Irrigator's constructor.)
		i = new Irrigator(null);
		i.restoreState();
	}

	@Test
	public void testRestoreStateModeIllegal() throws StartupException,
			StoreFailureException {
		// Tests restoreState() when different invalid entries exist
		// in the store.

		for (int i = 0; i < 4; i++) {

			// Irrigation parameters are written to the store ...
			MockStorageDevice mockStore = new MockStorageDevice();
			mockStore.setData("mode", "automatic");
			mockStore.setData("allocation", "1000");
			mockStore.setData("irrigationTime", "2100");
			mockStore.setData("irrigationDays", "MSaSo");

			this.i = new Irrigator(mockStore);

			// ... and then all parameters are changed in turn to
			// an illegal format ...
			switch (i) {
			case 0:
				mockStore.setData("mode", "auto");
				break;
			case 1:
				mockStore.setData("allocation", "ninety");
				break;
			case 2:
				mockStore.setData("irrigationTime", "nine");
				break;
			case 3:
				mockStore.setData("irrigationDays", "Sunnday");
				break;
			}

			// ... so a StartupException is expected when
			// restoreState() is invoked.
			try {
				this.i.restoreState();
				fail();
			} catch (StartupException e) {

			}
		}
	}

	@Test
	public void testAddZone() throws StartupException {
		// Tests addZone().
		i.addZone("Z00", "locZ00");

		assertEquals(4, zones.size());
		Zone zone = null;
		for (Object z : zones) {
			if (((Zone) z).getID().equals("Z00")) {
				zone = (Zone) z;
				break;
			}
		}
		assertEquals("Z00", zone.getID());
		assertEquals("locZ00", zone.getLocation());
		assertEquals(50, zone.getCriticalMoistureLevel());
	}

	@Test
	public void testAddZoneAddedEarlier() throws StartupException,
			StoreFailureException {
		// Tests addZone(). The store already contains
		// an entry specifying the zone's critical level ...
		store.setData("Z00.criticalMoistureLevel", "67");

		i.addZone("Z00", "locZ00");

		assertEquals(4, zones.size());
		Zone zone = null;
		for (Object z : zones) {
			if (((Zone) z).getID().equals("Z00")) {
				zone = (Zone) z;
				break;
			}
		}
		assertEquals("Z00", zone.getID());
		assertEquals("locZ00", zone.getLocation());
		// ... to the newly added zone's level should be
		// set to that level.
		assertEquals(67, zone.getCriticalMoistureLevel());
	}

	@Test
	public void testAddZoneStorageFailure() throws StartupException,
			StoreFailureException {
		// Tests addZone() if the store fails.
		store.setData("Z00.criticalMoistureLevel", "67");
		store.setIsFailed(true);

		try {
			i.addZone("Z00", "locZ00");
			fail();
		} catch (StartupException e) {

		}

		assertEquals(4, zones.size());
		Zone zone = null;
		for (Object z : zones) {
			if (((Zone) z).getID().equals("Z00")) {
				zone = (Zone) z;
				break;
			}
		}
		assertEquals("Z00", zone.getID());
		assertEquals("locZ00", zone.getLocation());
		assertEquals(0, zone.getCriticalMoistureLevel());
	}

	@Test
	public void testAddZoneIllegalDataInStorage() throws StartupException,
			StoreFailureException {
		// Tests addZone(). The store contains invalid data, ...
		store.setData("Z00.criticalMoistureLevel", "A66");

		try {
			i.addZone("Z00", "locZ00");
			fail();
		} catch (StartupException e) {
			// ... so a startup exception is expected.
		}

		assertEquals(4, zones.size());
		Zone zone = null;
		for (Object z : zones) {
			if (((Zone) z).getID().equals("Z00")) {
				zone = (Zone) z;
				break;
			}
		}
		assertEquals("Z00", zone.getID());
		assertEquals("locZ00", zone.getLocation());
		assertEquals(0, zone.getCriticalMoistureLevel());
	}

	@Test
	public void testAddZoneIDNull() throws StartupException,
			StoreFailureException {
		// Tests whether addZone() checks its
		// precondition.

		try {
			i.addZone(null, "locZ00");
			fail();
		} catch (IllegalArgumentException e) {
			// Exception expected
		}

		// Tests whether the existing zones were changed.
		assertEquals(3, zones.size());
		for (Zone z : zones) {
			assertFalse(z.getID().equals(null));
			assertFalse(z.getLocation().equals("locZ00"));
		}
	}

	@Test
	public void testAddZoneExistingID() throws StartupException,
			StoreFailureException {
		// Tests whether addZone() adds two zones
		// both having the same id.
		i.addZone("Z00", "locZ00");

		i.addZone("Z00", "locZ00");

		ArrayList<Zone> zoneList = new ArrayList<Zone>();
		assertEquals(5, zones.size());
		for (Zone z : zones) {
			if (z.getID().equals("Z00"))
				zoneList.add(z);
		}

		// Both zones should be added.
		assertEquals(2, zoneList.size());
		for (Zone z : zoneList) {
			assertEquals("locZ00", z.getLocation());
		}

	}

	@Test
	public void testAddSensor() throws StartupException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests addSensor().
		MockSensorDevice mockDevice;
		i.addZone("Z00", "locZ00");
		i.addSensor("Z00", "S00", mockDevice = new MockSensorDevice(0),
				"locZ00");

		Zone zone = null;
		for (Object z : zones) {
			if (((Zone) z).getID().equals("Z00")) {
				zone = (Zone) z;
				break;
			}
		}
		Sensor sensor = extractField(zone, "sensor");
		assertEquals("S00", sensor.getID());
		assertEquals("locZ00", sensor.getLocation());
		assertEquals(mockDevice, extractField(sensor, "device"));
		assertFalse(sensor.isFailed());
	}

	@Test
	public void testAddSensorStoreEntry() throws StartupException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, StoreFailureException {
		// Tests addSensor(). There is an entry in the store
		// describing the added sensor's state.
		MockSensorDevice mockDevice;
		store.setData("S00.isWorking", "false");
		i.addZone("Z00", "locZ00");
		i.addSensor("Z00", "S00", mockDevice = new MockSensorDevice(0),
				"locZ00");

		Zone zone = null;
		for (Object z : zones) {
			if (((Zone) z).getID().equals("Z00")) {
				zone = (Zone) z;
				break;
			}
		}
		Sensor sensor = extractField(zone, "sensor");
		assertEquals("S00", sensor.getID());
		assertEquals("locZ00", sensor.getLocation());
		assertEquals(mockDevice, extractField(sensor, "device"));
		// Must correspond to the stored state.
		assertTrue(sensor.isFailed());
	}

	@Test
	public void testAddSensorStoreFailure() throws StartupException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, StoreFailureException {
		// Tests addSensor(). The store fails.
		MockSensorDevice mockDevice = null;
		store.setData("S00.isWorking", "false");
		i.addZone("Z00", "locZ00");

		store.setIsFailed(true);
		try {
			i.addSensor("Z00", "S00", mockDevice = new MockSensorDevice(0),
					"locZ00");
			fail();
		} catch (StartupException e) {
			// Exception expected due to storage failure.
		}

		Zone zone = null;
		for (Object z : zones) {
			if (((Zone) z).getID().equals("Z00")) {
				zone = (Zone) z;
				break;
			}
		}
		Sensor sensor = extractField(zone, "sensor");
		assertEquals("S00", sensor.getID());
		assertEquals("locZ00", sensor.getLocation());
		assertEquals(mockDevice, extractField(sensor, "device"));
		assertFalse(sensor.isFailed());
	}

	@Test
	public void testAddSensorIDNull() throws StartupException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, StoreFailureException {
		// Tests addSensor(). The new sensor's ID is null.

		store.setData(".isWorking", "false");
		i.addZone("Z00", "locZ00");

		try {
			i.addSensor("Z00", null, new MockSensorDevice(0), "locZ00");
			fail();
		} catch (IllegalArgumentException e) {
			// Expected due to the sensor's ID.
		}

		Zone zone = null;
		for (Object z : zones) {
			if (((Zone) z).getID().equals("Z00")) {
				zone = (Zone) z;
				break;
			}
		}
		// No sensor has been added to the zone.
		Sensor sensor = extractField(zone, "sensor");
		assertNull(sensor);
	}

	@Test
	public void testAddSensorIllegalZone() throws StartupException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, StoreFailureException {
		// Tests addSensor(). No zone having the specified id
		// exists.

		store.setData("S00.isWorking", "false");
		i.addZone("Z00", "locZ00");
		i.addZone("Z01", "locZ01");
		HashMap<Zone, Sensor> sensorMap = new HashMap<Zone, Sensor>();

		// Store the current sensors.
		for (Zone z : zones)
			sensorMap.put(z, (Sensor) extractField(z, "sensor"));

		try {
			i.addSensor("Z000", "S00", new MockSensorDevice(0), "locZ00");
			fail();
		} catch (IllegalArgumentException e) {
			// Expected because of bad zone id.
		}

		// None of the existing zone's should have their
		// sensor changed.
		for (Zone z : zones)
			assertEquals(sensorMap.get(z), extractField(z, "sensor"));
	}

	@Test
	public void testAddSensorZoneNull() throws StartupException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, StoreFailureException {
		// Tests addSensor(). Tries to add the sensor to zone null.

		store.setData("S00.isWorking", "false");
		i.addZone("Z00", "locZ00");
		i.addZone("Z01", "locZ01");
		HashMap<Zone, Sensor> sensorMap = new HashMap<Zone, Sensor>();

		// Store the current sensors.
		for (Zone z : zones)
			sensorMap.put(z, (Sensor) extractField(z, "sensor"));

		try {
			i.addSensor(null, "S00", new MockSensorDevice(0), "locZ00");
			fail();
		} catch (IllegalArgumentException e) {

		}

		// None of the existing zone's should have its
		// sensor changed.
		for (Zone z : zones)
			assertEquals(sensorMap.get(z), extractField(z, "sensor"));
	}

	@Test
	public void testAddValve() throws StartupException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests addValve().
		MockValveDevice mockDevice;
		i.addZone("Z00", "locZ00");
		i.addValve("Z00", "V00", mockDevice = new MockValveDevice(), "tV00",
				100, "locZ00");

		// Extract the valves of zone "Z00"
		Zone zone = null;
		for (Object z : zones) {
			if (((Zone) z).getID().equals("Z00")) {
				zone = (Zone) z;
				break;
			}
		}
		HashSet<Valve> valves = extractField(zone, "valves");

		assertEquals(1, valves.size());
		Valve valve = (Valve) valves.toArray()[0];
		assertEquals("locZ00", valve.getLocation());
		assertEquals(mockDevice, extractField(valve, "device"));
		assertFalse(valve.isFailed());
		assertEquals(100, extractField(valve, "flowRate"));
	}

	@Test
	public void testAddValveStoreEntry() throws StartupException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, StoreFailureException {
		// Tests addValve(). There is an entry in the store specifying
		// the added valve's failure status ...
		MockValveDevice mockDevice;
		store.setData("V00.isWorking", "false");
		i.addZone("Z00", "locZ00");
		i.addValve("Z00", "V00", mockDevice = new MockValveDevice(), "tV00",
				100, "locZ00");

		// Extract the valves of zone "Z00"
		Zone zone = null;
		for (Object z : zones) {
			if (((Zone) z).getID().equals("Z00")) {
				zone = (Zone) z;
				break;
			}
		}
		HashSet<Valve> valves = extractField(zone, "valves");

		assertEquals(1, valves.size());
		Valve valve = (Valve) valves.toArray()[0];
		assertEquals("locZ00", valve.getLocation());
		assertEquals(mockDevice, extractField(valve, "device"));
		// ... so the added valve must correspond to this entry.
		assertTrue(valve.isFailed());
		assertEquals(100, extractField(valve, "flowRate"));
	}

	@Test
	public void testAddValveStoreFailure() throws StartupException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, StoreFailureException {
		// Tests addValve(). The store fails.
		MockValveDevice mockDevice = null;
		store.setData("V00.isWorking", "false");
		i.addZone("Z00", "locZ00");

		store.setIsFailed(true);
		try {
			i.addValve("Z00", "V00", mockDevice = new MockValveDevice(),
					"tV00", 100, "locZ00");
			fail();
		} catch (StartupException e) {
			// Expected because of the failed store.
		}

		// Extract the valves of zone "Z00"
		Zone zone = null;
		for (Object z : zones) {
			if (((Zone) z).getID().equals("Z00")) {
				zone = (Zone) z;
				break;
			}
		}
		HashSet<Valve> valves = extractField(zone, "valves");

		assertEquals(1, valves.size());
		Valve valve = (Valve) valves.toArray()[0];
		assertEquals("locZ00", valve.getLocation());
		assertEquals(mockDevice, extractField(valve, "device"));
		assertFalse(valve.isFailed());
		assertEquals(100, extractField(valve, "flowRate"));
	}

	@Test
	public void testAddValveIDNull() throws StartupException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, StoreFailureException {
		// Tests addValve(). The added valve should have the id null.

		store.setData(".isWorking", "false");
		i.addZone("Z00", "locZ00");

		try {
			i.addValve("Z00", null, new MockValveDevice(), "tV00", 100,
					"locZ00");
			fail();
		} catch (IllegalArgumentException e) {

		}

		// Extract the valves of zone "Z00"
		Zone zone = null;
		for (Object z : zones) {
			if (((Zone) z).getID().equals("Z00")) {
				zone = (Zone) z;
				break;
			}
		}
		HashSet<Valve> valves = extractField(zone, "valves");
		assertEquals(0, valves.size());
	}

	@Test
	public void testAddValveIllegalZone() throws StartupException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, StoreFailureException {
		// Tries to add a valve to a non-existing zone.

		store.setData("S00.isWorking", "false");
		i.addZone("Z00", "locZ00");
		i.addZone("Z01", "locZ01");
		HashMap<Zone, HashSet<Valve>> valveSets = new HashMap<Zone, HashSet<Valve>>();

		// For each zone, a copy of the HashMap holding
		// the zone's valves is created.
		for (Zone z : zones) {
			HashSet<Valve> valves = extractField(z, "valves");
			HashSet<Valve> copyOfValves = new HashSet<Valve>();
			for (Valve v : valves)
				copyOfValves.add(v);
			valveSets.put(z, copyOfValves);
		}

		try {
			i.addValve("Z000", "V00", new MockValveDevice(), "tV00", 100,
					"locZ00");
		} catch (IllegalArgumentException e) {

		}

		// Each zone's valves must be present in the
		// previously copied HashMap - no additional valve
		// should have been added to one of the zones.
		for (Zone z : zones) {
			HashSet<Valve> valves = extractField(z, "valves");
			HashSet<Valve> copyOfValves = valveSets.get(z);
			for (Valve v : valves)
				assertTrue(copyOfValves.contains(v));
		}
	}

	@Test
	public void testAddValveZoneNull() throws StartupException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, StoreFailureException {
		// Tries to add a valve to the zone null.

		store.setData("S00.isWorking", "false");
		i.addZone("Z00", "locZ00");
		i.addZone("Z01", "locZ01");
		HashMap<Zone, HashSet<Valve>> valveSets = new HashMap<Zone, HashSet<Valve>>();

		// For each zone, a copy of the HashMap holding
		// the zone's valves is created.
		for (Zone z : zones) {
			HashSet<Valve> valves = extractField(z, "valves");
			HashSet<Valve> copyOfValves = new HashSet<Valve>();
			for (Valve v : valves)
				copyOfValves.add(v);
			valveSets.put(z, copyOfValves);
		}

		try {
			i.addValve(null, "V00", new MockValveDevice(), "tV00", 100,
					"locZ00");
		} catch (IllegalArgumentException e) {

		}

		// Each zone's valves must be present in the
		// previously copied HashMap - no additional valve
		// should have been added to one of the zones.
		for (Zone z : zones) {
			HashSet<Valve> valves = extractField(z, "valves");
			HashSet<Valve> copyOfValves = valveSets.get(z);
			for (Valve v : valves)
				assertTrue(copyOfValves.contains(v));
		}
	}

	@Test
	public void testIsFailedDevice() {
		// Initially, no device has failed
		assertFalse(i.isFailedDevice());
	}

	@Test
	public void testIsFailedDeviceFailureExists() {
		// Add a failed device to see whether
		// the result of isFailedDevice() changes
		failedDevices.add("Test");
		assertTrue(i.isFailedDevice());
	}

	@Test
	public void testGetNextFailedDevice() {
		// Since there is no failed device initially,
		// null is returned by getNextFailedDevice()
		assertNull(i.getNextFailedDevice());
	}

	@Test
	public void testGetNextFailedDeviceFailureExists() {
		// If there exists a failed device
		// it should be reported. Afterwards, the list
		// of failed devices should be empty.
		failedDevices.add("Test");
		assertEquals("Test", i.getNextFailedDevice());
		assertNull(i.getNextFailedDevice());
	}

	@Test
	public void testSetIsStoreFailure() {
		// Tests setIsStoreFailure().
		i.setIsStoreFailure(true);
		assertTrue(i.isStoreFailure());

		i.setIsStoreFailure(false);
		assertFalse(i.isStoreFailure());
	}

	@Test
	public void testIrrigationDaysProperty() throws StoreFailureException {
		// Tests getIrrigationDays() and setIrrigationDays().
		HashSet<Day> irrigationDays = new HashSet<Day>();
		irrigationDays.add(Day.MONDAY);
		irrigationDays.add(Day.SUNDAY);
		irrigationDays.add(Day.FRIDAY);

		i.setIrrigationDays(irrigationDays);

		// The irrigation days should be the same days
		// that are in irrigationDays
		Set<Day> days = i.getIrrigationDays();
		assertEquals(3, days.size());
		for (Day d : days)
			assertTrue(irrigationDays.contains(d));

		// Additionally, the store should contain an entry
		// specifying the irrigation days.
		String storeEntry = store.getData("irrigationDays", "");
		assertTrue("SuMF".equals(storeEntry) || "SuFM".equals(storeEntry)
				|| "FSuM".equals(storeEntry) || "FMSu".equals(storeEntry)
				|| "MFSu".equals(storeEntry) || "MSuF".equals(storeEntry));
	}

	@Test
	public void testIrrigationDaysPropertyDifferentDays()
			throws StoreFailureException {
		// Tests getIrrigationDays() and setIrrigationDays() using
		// a different set of days.
		HashSet<Day> irrigationDays = new HashSet<Day>();
		irrigationDays.add(Day.THURSDAY);
		irrigationDays.add(Day.SATURDAY);

		i.setIrrigationDays(irrigationDays);

		// The irrigation days should be the same days
		// that are in irrigationDays
		Set<Day> days = i.getIrrigationDays();
		assertEquals(2, days.size());
		for (Day d : days)
			assertTrue(irrigationDays.contains(d));

		// Additionally, the store should contain an entry
		// specifying the irrigation days.
		String storeEntry = store.getData("irrigationDays", "");
		assertTrue("ThSa".equals(storeEntry) || "SaTh".equals(storeEntry));
	}

	@Test
	public void testIrrigationDaysPropertyEmptySet()
			throws StoreFailureException {
		// Tests setIrrigationDays() using an empty set.
		HashSet<Day> irrigationDays = new HashSet<Day>();

		i.setIrrigationDays(irrigationDays);

		Set<Day> days = i.getIrrigationDays();
		assertEquals(0, days.size());

		assertEquals("", store.getData("irrigationDays", ""));
	}

	@Test
	public void testIrrigationDaysPropertyNull() throws StoreFailureException {
		// Tests setIrrigationDays() using null.
		i.setIrrigationDays(null);

		Set<Day> days = i.getIrrigationDays();
		assertEquals(0, days.size());

		assertEquals("", store.getData("irrigationDays", ""));
	}

	@Test
	public void testGetZoneReports() {
		// Tests getZoneReports().
		List<ZoneReport> reports = i.getZoneReports();
		
		// For each zone, there has to be a ZoneReport.
		assertEquals(3, i.getZoneReports().size());
		for (Zone z : zones) {
			ZoneReport report = z.getZoneReport();
			ZoneReport matchingReport = null;
			for (ZoneReport r : reports)
				if (r.id == report.id)
					matchingReport = r;
			assertNotNull(matchingReport);
		}
	}

	@Test
	public void testGetZoneReportsNoExistingZones() {
		// Tests getZoneReports() when called on a
		// newly created irrigator.
		i = new Irrigator(store);
		assertEquals(0, i.getZoneReports().size());
	}

	@Test
	public void testGetValveReports() {
		// Tests getValveReports().
		List<ValveReport> reports = i.getValveReports();
		
		assertEquals(9, reports.size());
		// For each zone ...
		for (Zone z : zones) {
			// ... and each valve contained in that zone ...
			for (ValveReport report : z.getValveReports()) {
				ValveReport matchingReport = null;
				// ... there has to be a ValveReport.
				for (ValveReport r : reports)
					if (r.id == report.id && r.zoneID == report.zoneID)
						matchingReport = r;
				assertNotNull(matchingReport);
			}
		}
	}

	@Test
	public void testGetValveReportsNoExistingZones() {
		// Tests getValveReport() on a newly created
		// irrigator.
		i = new Irrigator(store);
		assertEquals(0, i.getValveReports().size());
	}

	@Test
	public void testGetFailureReport() {
		// Tests getFailureReport() on a newly
		// created irrigator
		FailureReport report = i.getFailureReport();
		assertTrue(report.failedSensors.isEmpty());
		assertTrue(report.failedValves.isEmpty());
	}

	@Test
	public void testGetFailureReportFailedDevices()
			throws DeviceFailureException {
		// Tests getFailureReport(). Some devices are actually failing.
		Object[] zoneArray = zones.toArray();
		HashMap<String, String> valveToZoneMap = new HashMap<String, String>();
		
		// Set up the failing devices:
		for (int i = 0; i < zoneArray.length; i++) {
			Zone currentZone = (Zone) zoneArray[i];
			// One half of the zones gets a failing valve, ...
			if (i % 2 == 0) {
				MockValveDevice mockValve = new MockValveDevice();
				mockValve.isFailed = true;
				currentZone.addValve("V4" + i, mockValve, "T4", 100,
						currentZone.getLocation());
				valveToZoneMap.put("V4" + i, currentZone.getID());
			} else {
				// ... the other half a failing sensor.
				MockSensorDevice mockSensor = new MockSensorDevice(6);
				currentZone.addSensor("S0", mockSensor,
						currentZone.getLocation());
			}
		}

		// Then, use all valves and sensors.
		for (Zone z : zones)
			i.openAllValves(z.getID());
		try {
			i.getMeasuredMoistureLevel(((Zone) zoneArray[1]).getID());
		} catch (DeviceFailureException e) {

		}

		// Since there are 3 zones in total, there is
		// one failing sensor ...
		FailureReport report = i.getFailureReport();
		assertEquals(1, report.failedSensors.size());
		SensorReport sensorReport = (SensorReport) report.failedSensors
				.toArray()[0];
		assertEquals("S0", sensorReport.id);
		
		// ... and two failing valves.
		assertEquals(2, report.failedValves.size());
		Object[] failedValvesArray = report.failedValves.toArray();
		for (int i = 0; i < failedValvesArray.length; i++) {
			ValveReport currentReport = (ValveReport) failedValvesArray[i];
			assertEquals("V4" + i * 2, currentReport.id);
			assertEquals(valveToZoneMap.get("V4" + i * 2), currentReport.zoneID);
		}
	}

	@Test
	public void testGetFailedDeviceReport() {
		HashMap<String, String> locationMap = new HashMap<String, String>();
		Object[] zoneArray = zones.toArray();
		
		// Set up the failing devices.
		for (int i = 0; i < zoneArray.length; i++) {
			Zone currentZone = (Zone) zoneArray[i];
			if (i % 2 == 0) {
				String location = null;
				MockValveDevice mockValve = new MockValveDevice();
				mockValve.isFailed = true;
				currentZone.addValve("V4" + i, mockValve, "T4", 100,
						location = currentZone.getLocation());
				locationMap.put("V4" + i, location);
			} else {
				MockSensorDevice mockSensor = new MockSensorDevice(6);
				currentZone.addSensor("S0", mockSensor,
						currentZone.getLocation());
				locationMap.put("S0", currentZone.getLocation());
			}
		}

		for (Zone z : zones)
			i.openAllValves(z.getID());
		try {
			i.getMeasuredMoistureLevel(((Zone) zoneArray[1]).getID());
		} catch (DeviceFailureException e) {

		}

		// For each failed device, getFailedDeviceReport should
		// return an appropriate report.
		FailedDeviceReport report = i.getFailedDeviceReport("V40");
		assertEquals("V40", report.deviceID);
		assertEquals(locationMap.get("V40"), report.location);
		report = i.getFailedDeviceReport("S0");
		assertEquals("S0", report.deviceID);
		assertEquals(locationMap.get("S0"), report.location);
		report = i.getFailedDeviceReport("V42");
		assertEquals("V42", report.deviceID);
		assertEquals(locationMap.get("V42"), report.location);
	}

	@Test
	public void testGetCriticalMoistureLevel() {
		// Tests getCriticalMoistureLevel().
		HashMap<String, Integer> levelMap = new HashMap<String, Integer>();
		int i = 0;
		
		// Change each zone's moisture level.
		for (Zone z : zones) {
			int level = 50 + i * 5;
			i++;
			z.setCriticalMoistureLevel(level);
			levelMap.put(z.getID(), level);
		}

		for (Zone z : zones) {
			assertEquals((Integer) levelMap.get(z.getID()),
					(Integer) this.i.getCriticalMoistureLevel(z.getID()));
		}
	}

	@Test
	public void testGetCriticalMoistureLevelIllegalZone() {
		// Tests getCriticalMoistureLevel. No zone with the given
		// name exists.
		try {
			i.getCriticalMoistureLevel("Z000");
			fail();
		} catch (IllegalArgumentException e) {

		}
	}

	@Test
	public void testGetCriticalMoistureLevelNullZone() {
		// Tests getCriticalMoistureLevel. The argument is null.
		try {
			i.getCriticalMoistureLevel(null);
			fail();
		} catch (IllegalArgumentException e) {

		}
	}

	@Test
	public void testGetMeasuredMoistureLevel() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {
		// Tests getMeasuredMoistureLevel().
		Zone zone1 = null;
		Zone zone2 = null;
		for (Zone z : zones) {
			if (z.getID().equals("Z0"))
				zone1 = z;
			if (z.getID().equals("Z1"))
				zone2 = z;
		}

		// Set the values reported by the underlying
		// mock devices.
		((MockSensorDevice) extractField(
				((Sensor) extractField(zone1, "sensor")), "device")).value = 55;
		((MockSensorDevice) extractField(
				((Sensor) extractField(zone2, "sensor")), "device")).value = 65;

		assertEquals(55, i.getMeasuredMoistureLevel("Z0"));
		assertEquals(65, i.getMeasuredMoistureLevel("Z1"));
	}

	@Test
	public void testGetMeasuredMoistureLevelSensorFailure()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {
		// Tests getMeasuredMoistureLevel(). The Sensor in the queried Zone is
		// failing.
		Zone zone1 = null;
		for (Zone z : zones) {
			if (z.getID().equals("Z0"))
				zone1 = z;
		}

		// Change the status of the underlying sensor device
		Sensor sensor;
		Field sensorDevice = extractFieldAsField(
				sensor = ((Sensor) extractField(zone1, "sensor")), "device");
		sensorDevice.setAccessible(true);
		sensorDevice.set(sensor, new MockSensorDevice(6));

		try {
			i.getMeasuredMoistureLevel("Z0");
			fail();
		} catch (DeviceFailureException e) {
			assertEquals(1, e.getDevices().size());
			assertTrue(e.getDevices().contains("S0"));
		}

	}

	@Test
	public void testGetMeasuredMoistureLevelIllegalZone()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {
		// Tries to get the measured moisture level from
		// a zone not existing.
		try {
			i.getMeasuredMoistureLevel("Z000");
			fail();
		} catch (IllegalArgumentException e) {
		}
	}

	@Test
	public void testGetMeasuredMoistureLevelNullZone()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {
		// Calls testGetMoistureLevel(). The argument is null.
		try {
			i.getMeasuredMoistureLevel(null);
			fail();
		} catch (IllegalArgumentException e) {
		}

	}

	@Test
	public void testGetWaterUsed() {
		// A newly created zone has not used
		// any water.
		assertEquals(0, i.getWaterUsed());
	}

	@Test
	public void testGetWaterUsedExistingCycle()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {
		// Tests getWaterUsed() if a cycle is present.
		i.startManualCycle();

		IrrigationCycle cycle = extractField(i, "cycle");

		// Make the cycle use some water.
		for (Zone z : zones)
			i.openAllValves(z.getID());

		for (int i = 0; i < 3; i++)
			cycle.tick();

		assertEquals(2700, i.getWaterUsed());

		cycle.tick();

		assertEquals(3600, i.getWaterUsed());
	}

	@Test
	public void testSetModeIdenticalMode() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests what happens when the mode is set to
		// its current value.
		IrrigationCycle cycle = extractField(i, "cycle");

		i.setMode(Mode.AUTOMATIC);

		// Since the mode was already set to Automatic,
		// no changes are expected
		assertEquals(Mode.AUTOMATIC, i.getMode());
		assertEquals(cycle, extractField(i, "cycle"));
	}

	@Test
	public void testSetModeAutomaticNullCycle()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests what happens when the mode is
		// changed to manual and there is no running cycle
		i.setMode(Mode.MANUAL);

		// Since the mode was changed, the following
		// values are expected.
		assertEquals(Mode.MANUAL, i.getMode());
		assertNull(extractField(i, "cycle"));
	}

	@SuppressWarnings("unchecked")
	@Test
	public void testSetModeAutomaticExistingCycle()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, StoreFailureException {
		// Tests what happens when the mode is changed
		// to manual and there is a running cycle
		AutoCycle oldCycle;
		Field cycleField = extractFieldAsField(i, "cycle");
		
		// Pretend that there is a cycle running.
		cycleField.setAccessible(true);
		cycleField.set(i,
				oldCycle = new AutoCycle((Set<Zone>) extractField(i, "zones"),
						1000));

		i.setMode(Mode.MANUAL);

		assertTrue(oldCycle.isDone());
		assertEquals(Mode.MANUAL, i.getMode());
		assertNull(extractField(i, "cycle"));
		assertEquals("Manual", store.getData("mode", "automatic"));
		assertNull(i.getNextFailedDevice());
	}

	@Test
	public void testSetModeAutomaticExistingCycleError()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, StoreFailureException {
		// Tests what happens when the mode is changed
		// to manual and there is a running cycle which causes an error
		@SuppressWarnings("unchecked")
		MockAutoCycle oldCycle = new MockAutoCycle((Set<Zone>) extractField(i,
				"zones"), 1000);
		oldCycle.setFails(true);
		oldCycle.superDone = true;
		// Install the mock cycle.
		Field cycleField = extractFieldAsField(i, "cycle");
		cycleField.setAccessible(true);
		cycleField.set(i, oldCycle);

		i.setMode(Mode.MANUAL);

		// No abnormal behavior is expected.
		assertTrue(oldCycle.isDone());
		assertEquals(Mode.MANUAL, i.getMode());
		assertNull(extractField(i, "cycle"));
		assertEquals("Manual", store.getData("mode", "automatic"));
		List<String> failedDevices = null;
		try {
			failedDevices = getFailedDevices(i);
		} catch (Exception e) {
			e.printStackTrace();
			fail();
		}
		assertEquals("MockAutoCycle", failedDevices.get(0));
	}

	@Test
	public void testSetModeIdenticalManualMode()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, StoreFailureException {
		// Tests how setMode behaves if it is set
		// repeatedly.
		i.setMode(Mode.MANUAL);

		i.setMode(Mode.MANUAL);

		assertEquals(Mode.MANUAL, i.getMode());
		assertNull(extractField(i, "cycle"));
		assertEquals("Manual", store.getData("mode", "automatic"));
		assertNull(i.getNextFailedDevice());
	}

	@Test
	public void testSetModeChangeToAutomatic() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			StoreFailureException {
		// Tests setMode if the mode is set back
		// to Automatic
		i.setMode(Mode.MANUAL);
		i.setMode(Mode.AUTOMATIC);

		assertEquals(Mode.AUTOMATIC, i.getMode());
		assertNull(extractField(i, "cycle"));
		assertEquals("Automatic", store.getData("mode", "Manual"));
		assertNull(i.getNextFailedDevice());
	}

	@SuppressWarnings("unchecked")
	@Test
	public void testSetModeChangeToAutomaticCycleRunning()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, StoreFailureException {
		// Tests setMode if the mode is set back
		// to Automatic and a manual cycle is active
		i.setMode(Mode.MANUAL);

		MockManualCycle mockCycle;

		// Start an auto cycle.
		i.startManualCycle();
		Field cycleField = extractFieldAsField(i, "cycle");
		cycleField.setAccessible(true);
		// Replace the current cycle with a mock cycle.
		cycleField.set(i, mockCycle = new MockManualCycle(
				(Set<Zone>) extractField(i, "zones")));
		mockCycle.setFails(true);

		i.setMode(Mode.AUTOMATIC);

		assertEquals(Mode.AUTOMATIC, i.getMode());
		assertEquals(mockCycle, extractField(i, "cycle"));
		assertEquals("Automatic", store.getData("mode", "Manual"));
		assertNull(i.getNextFailedDevice());
	}

	@Test
	public void testSetModeArgumentNull() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			StoreFailureException {
		// According to setMode's precondition,
		// this should work fine.
		i.setMode(null);
	}

	@Test
	public void testIrrigationTimeProperty() throws StoreFailureException {
		// Tests the irrigator's irrigation time property.
		i.setIrrigationTime(2100);
		assertEquals(2100, i.getIrrigationTime());

		assertEquals("2100", store.getData("irrigationTime", "0000"));
	}

	@Test
	public void testIrrigationTimePropertyDifferentTime()
			throws StoreFailureException {
		// Tests the irrigator's irrigation time property
		// using a different time.
		i.setIrrigationTime(800);
		assertEquals(800, i.getIrrigationTime());

		assertEquals("800", store.getData("irrigationTime", "0000"));
	}

	@Test
	public void testIrrigationTimePropertyIllegalTime()
			throws StoreFailureException {
		// Tests the irrigator's irrigation time property
		// using an invalid irrigation time.
		i.setIrrigationTime(800);

		String oldTime = store.getData("irrigationTime", "0000");

		try {
			i.setIrrigationTime(2188);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			i.setIrrigationTime(2548);
			fail();
		} catch (IllegalArgumentException e) {

		}

		assertEquals(oldTime, store.getData("irrigationTime", "0000"));
	}

	@Test
	public void testSetAllocation() throws StoreFailureException {
		// Tests setAllocation().
		i.setAllocation(1);

		assertEquals(1, i.getAllocation());

		assertEquals("1", store.getData("allocation", "1000"));

	}

	@Test
	public void testSetAllocationRunningCycle() throws StoreFailureException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {
		// Tests setAllocation() while an AutoCycle is running.
		MockAutoCycle cycle;

		Field cycleField = extractFieldAsField(i, "cycle");
		cycleField.setAccessible(true);
		cycleField.set(i, cycle = new MockAutoCycle(zones, 1500));

		i.setAllocation(Integer.MAX_VALUE);

		assertEquals(Integer.MAX_VALUE, i.getAllocation());
		assertEquals(Integer.MAX_VALUE, cycle.allocation);

		assertEquals(Integer.toString(Integer.MAX_VALUE),
				store.getData("allocation", "1000"));

	}

	@Test
	public void testSetAllocationCycleFailure() throws StoreFailureException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {
		// Tests setAllocation() while an AutoCycle is running which fails.
		MockAutoCycle cycle;

		Field cycleField = extractFieldAsField(i, "cycle");
		cycleField.setAccessible(true);
		cycleField.set(i, cycle = new MockAutoCycle(zones, 1500));

		cycle.setFails(true);

		i.setAllocation(Integer.MAX_VALUE);

		List<String> failedDevices = null;
		try {
			failedDevices = getFailedDevices(i);
		} catch (Exception e) {
			e.printStackTrace();
			fail();
		}
		assertEquals("MockAutoCycle", failedDevices.get(0));
		assertEquals(Integer.MAX_VALUE, cycle.allocation);

		assertEquals(Integer.toString(Integer.MAX_VALUE),
				store.getData("allocation", "1000"));

	}

	@Test
	public void testSetAllocationIllegalArgument()
			throws StoreFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {
		// Tests whether setAllocation() checks its precondition.
		i.setAllocation(1000);

		try {
			i.setAllocation(0);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			i.setAllocation(-Integer.MAX_VALUE);
			fail();
		} catch (IllegalArgumentException e) {

		}

		assertEquals(1000, i.getAllocation());

		assertEquals("1000", store.getData("allocation", "0"));

	}

	@Test
	public void testSetCriticalMoistureLevel() throws StoreFailureException {
		// Tests setCriticalMoistureLevel().
		i.setCriticalMoistureLevel("Z1", 0);

		Zone zone = null;
		for (Zone z : zones)
			if (z.getID().equals("Z1"))
				zone = z;

		assertEquals(0, zone.getCriticalMoistureLevel());
		assertEquals("0", store.getData("Z1.criticalMoistureLevel", "1000"));

		i.setCriticalMoistureLevel("Z1", 100);

		zone = null;
		for (Zone z : zones)
			if (z.getID().equals("Z1"))
				zone = z;

		assertEquals(100, zone.getCriticalMoistureLevel());
		assertEquals("100", store.getData("Z1.criticalMoistureLevel", "1000"));
	}

	@Test
	public void testSetCriticalMoistureLevelZoneNull()
			throws StoreFailureException {
		// Tests whether setCriticalMoistureLevel() checks
		// its preconditions regarding the zone's name
		try {
			i.setCriticalMoistureLevel(null, 75);
			fail();
		} catch (IllegalArgumentException e) {

		}

		assertTrue(store.getHashMap().isEmpty());
		for (Zone z : zones)
			assertEquals(50, z.getCriticalMoistureLevel());
	}

	@Test
	public void testSetCriticalMoistureLevelZoneInvalid()
			throws StoreFailureException {
		// Tests whether setCriticalMoistureLevel() checks
		// its preconditions regarding the zone's existance correctly
		try {
			i.setCriticalMoistureLevel("Z000", 75);
			fail();
		} catch (IllegalArgumentException e) {

		}

		assertTrue(store.getHashMap().isEmpty());
		for (Zone z : zones)
			assertEquals(50, z.getCriticalMoistureLevel());
	}

	@Test
	public void testSetCriticalMoistureLevelLevelInvalid()
			throws StoreFailureException {
		// Tests whether setCriticalMoistureLevel() checks
		// its preconditions regarding the new critical level
		// correctly
		try {
			i.setCriticalMoistureLevel("Z0", Integer.MIN_VALUE);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			i.setCriticalMoistureLevel("Z0", -1);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			i.setCriticalMoistureLevel("Z0", 101);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			i.setCriticalMoistureLevel("Z0", Integer.MAX_VALUE);
			fail();
		} catch (IllegalArgumentException e) {

		}

		assertTrue(store.getHashMap().isEmpty());
		for (Zone z : zones)
			assertEquals(50, z.getCriticalMoistureLevel());
	}

	@Test
	public void testStartManualCycle() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Starts a new manual cycle
		i.startManualCycle();

		@SuppressWarnings("unused")
		ManualCycle cycle;
		assertNotNull(cycle = extractField(i, "cycle"));
	}

	@Test
	public void testStartManualCycleExistingCycle()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tries to start a new manual cycle when
		// there's already an existing cycle.
		i.startManualCycle();

		ManualCycle cycle;
		assertNotNull(cycle = extractField(i, "cycle"));

		i.startManualCycle();
		assertSame(cycle, extractField(i, "cycle"));
	}

	@Test
	public void testStopManualCycle() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests stopManualCycle().
		i.startManualCycle();

		MockManualCycle mockCycle;

		// Replace the current cycle with a
		// mock cycle.
		Field cycleField = extractFieldAsField(i, "cycle");
		cycleField.setAccessible(true);
		cycleField.set(i, mockCycle = new MockManualCycle(zones));

		i.stopManualCycle();

		// The cycle should now be stopped.
		assertNull(extractField(i, "cycle"));
		assertTrue(mockCycle.isEnded());

	}

	@Test
	public void testStopManualCycleCycleFails()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests stopManualCycle(). The stopped cycle is failing.
		i.startManualCycle();

		MockManualCycle mockCycle;

		// Replace the current cycle with a
		// mock cycle.
		Field cycleField = extractFieldAsField(i, "cycle");
		cycleField.setAccessible(true);
		cycleField.set(i, mockCycle = new MockManualCycle(zones));
		mockCycle.setFails(true);

		i.stopManualCycle();

		assertTrue(mockCycle.isEnded());
		assertNull(extractField(i, "cycle"));
		List<String> failedDevices = null;
		try {
			failedDevices = getFailedDevices(i);
		} catch (Exception e) {
			e.printStackTrace();
			fail();
		}
		assertEquals("MockManualCycle", failedDevices.get(0));

	}

	@Test
	public void testStopManualCycleNoRunningCycle() {
		// Tests stopManualCycle when there is no
		// manual cycle running.
		i.stopManualCycle();
	}

	@Test
	public void testStopManualCycleAutoCycle() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests stopManualCycle when there is an AutoCycle
		// running.
		i.startManualCycle();

		MockAutoCycle mockCycle;

		// Replace the currently running cycle
		// with a new AutoCycle.
		Field cycleField = extractFieldAsField(i, "cycle");
		cycleField.setAccessible(true);
		cycleField.set(i, mockCycle = new MockAutoCycle(zones, 15000));
		mockCycle.setFails(true);

		try {
			i.stopManualCycle();
			fail();
		} catch (Exception e) {
			// Since there is no manual cycle which could
			// be stopped, an exception is expected to
			// inform the caller about this issue.
		}

	}

	@SuppressWarnings("unchecked")
	@Test
	public void testToggleValve() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {
		// Tests toggleValve().
		
		// Extract 2 valves from different zones.
		Zone zone0 = null;
		Zone zone1 = null;
		for (Zone z : zones)
			if (z.getID().equals("Z0"))
				zone0 = z;
			else if (z.getID().equals("Z1"))
				zone1 = z;

		Valve valveZone0 = ((HashSet<Valve>) extractField(zone0, "valves"))
				.iterator().next();
		Valve valveZone1 = ((HashSet<Valve>) extractField(zone1, "valves"))
				.iterator().next();

		valveZone0.open();

		i.toggleValve("Z0", valveZone0.getID());
		assertFalse(valveZone0.isOpen());

		i.toggleValve("Z1", valveZone1.getID());
		assertTrue(valveZone1.isOpen());

	}

	@SuppressWarnings("unchecked")
	@Test
	public void testToggleValveFailingDevice() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {
		// Tests toggleValve() on a failing valve.
		Zone zone0 = null;
		for (Zone z : zones)
			if (z.getID().equals("Z0"))
				zone0 = z;

		// Extract a valve and change is failure state.
		Valve valveZone0 = ((HashSet<Valve>) extractField(zone0, "valves"))
				.iterator().next();
		MockValveDevice device = extractField(valveZone0, "device");
		device.isFailed = true;

		i.toggleValve("Z0", valveZone0.getID());

		assertFalse(valveZone0.isOpen());
		List<String> failedDevices = null;
		try {
			failedDevices = getFailedDevices(i);
		} catch (Exception e) {
			e.printStackTrace();
			fail();
		}
		assertEquals(valveZone0.getID(), failedDevices.get(0));

	}

	@Test
	public void testToggleValveZoneArgumentIllegalCases()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {
		// Tests whether toggleValve() checks its
		// preconditions regarding the first argument.
		try {
			i.toggleValve(null, "V1");
			fail();
		} catch (IllegalArgumentException e) {

		}

		// Zone Z000 does not exist.
		try {
			i.toggleValve("Z000", "V1");
			fail();
		} catch (IllegalArgumentException e) {

		}

	}

	@Test
	public void testToggleValveValveArgumentIllegalCases()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {
		// Tests whether toggleValve() checks its
		// preconditions regarding the second argument.
		try {
			i.toggleValve("Z1", null);
			fail();
		} catch (IllegalArgumentException e) {

		}

		// Valve V100 does not exist.
		try {
			i.toggleValve("Z0", "V100");
			fail();
		} catch (IllegalArgumentException e) {

		}

	}

	@Test
	public void testOpenAllValves() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests openAllValves().
		i.openAllValves("Z0");

		// Extract the valves from zone "Z0"
		Zone zone0 = null;
		for (Zone z : zones) {
			if (z.getID().equals("Z0"))
				zone0 = z;
		}

		HashSet<Valve> valves = extractField(zone0, "valves");

		for (Valve v : valves)
			assertTrue(v.isOpen());

	}

	@Test
	public void testOpenAllValvesFailingValve()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests openAllValves(). One of the valves is failing.
		
		// Extract the valves and let one of them fail.
		Zone zone0 = null;
		for (Zone z : zones) {
			if (z.getID().equals("Z0"))
				zone0 = z;
		}

		HashSet<Valve> valves = extractField(zone0, "valves");

		Valve failingValve = valves.iterator().next();
		MockValveDevice device = extractField(failingValve, "device");
		device.isFailed = true;

		i.openAllValves("Z0");

		for (Valve v : valves)
			if (v == failingValve)
				assertFalse(v.isOpen());
			else
				assertTrue(v.isOpen());
		
		List<String> failedDevices = null;
		try {
			failedDevices = getFailedDevices(i);
		} catch (Exception e) {
			e.printStackTrace();
			fail();
		}
		assertEquals(failingValve.getID(), failedDevices.get(0));

	}

	@Test
	public void testOpenAllValvesIllegalZone() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests openAllValves() with illegal Arguments.
		try {
			i.openAllValves("Z000");
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			i.openAllValves(null);
			fail();
		} catch (IllegalArgumentException e) {

		}

		// All valves should still be closed.
		for (Zone z : zones) {
			HashSet<Valve> valves = extractField(z, "valves");
			for (Valve v : valves)
				assertFalse(v.isOpen());
		}
	}

	@Test
	public void testCloseAllValves() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests closeAllValves().
		i.openAllValves("Z0");

		i.closeAllValves("Z0");

		// Extract Z0's valves.
		Zone zone0 = null;
		for (Zone z : zones) {
			if (z.getID().equals("Z0"))
				zone0 = z;
		}

		HashSet<Valve> valves = extractField(zone0, "valves");

		for (Valve v : valves)
			assertFalse(v.isOpen());

	}

	@Test
	public void testCloseAllValvesFailingValve()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests closeAllValves(). One of the valves is failing.
		i.openAllValves("Z1");

		// Extract all valves of zone Z1.
		Zone zone1 = null;
		for (Zone z : zones) {
			if (z.getID().equals("Z1"))
				zone1 = z;
		}

		HashSet<Valve> valves = extractField(zone1, "valves");

		// Change one valve's failure state.
		Valve failingValve = valves.iterator().next();
		MockValveDevice device = extractField(failingValve, "device");
		device.isFailed = true;

		i.closeAllValves("Z1");

		List<String> failedDevices = null;
		try {
			failedDevices = getFailedDevices(i);
		} catch (Exception e) {
			e.printStackTrace();
			fail();
		}
		assertEquals(failingValve.getID(), failedDevices.get(0));

		for (Valve v : valves)
			if (v == failingValve)
				// The valve failed while closing, so it is
				// still open ...
				assertTrue(v.isOpen());
			else
				assertFalse(v.isOpen());

	}

	@Test
	public void testCloseAllValvesIllegalZone()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests closeAllValves() with illegal Arguments.
		for (Zone z : zones)
			i.openAllValves(z.getID());

		try {
			i.closeAllValves("Z000");
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			i.closeAllValves(null);
			fail();
		} catch (IllegalArgumentException e) {

		}

		// All valves in all zones should still be open.
		for (Zone z : zones) {
			HashSet<Valve> valves = extractField(z, "valves");
			for (Valve v : valves)
				assertTrue(v.isOpen());
		}
	}

	@SuppressWarnings("unchecked")
	@Test
	public void testRepairDevice() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			DeviceFailureException, StoreFailureException {
		// Tests testRepairDevice().
		
		// Extract a valve and a sensor from two different zones
		// and set their failure state.
		Zone zone0 = null;
		Zone zone1 = null;

		for (Zone z : zones) {
			if (z.getID().equals("Z1"))
				zone0 = z;
			if (z.getID().equals("Z0"))
				zone1 = z;
		}

		HashSet<Valve> valves = extractField(zone0, "valves");
		Valve failingValve = valves.iterator().next();
		MockValveDevice valveDevice = extractField(failingValve, "device");
		valveDevice.isFailed = true;

		zone1.addSensor("S0", new MockSensorDevice(6), zone1.getLocation());
		Sensor failingSensor = extractField(zone1, "sensor");

		// Let the devices fail.
		i.openAllValves(zone0.getID());
		try {
			i.getMeasuredMoistureLevel(zone1.getID());
			fail();
		} catch (DeviceFailureException e) {

		}

		// Now they get repaired again.
		i.repairDevice(failingValve.getID());
		assertFalse(failingValve.isFailed());
		assertEquals("true",
				store.getData(failingValve.getID() + ".isWorking", "false"));
		assertFalse(((List<String>) extractField(i, "failedDevices"))
				.contains(failingValve.getID()));

		i.repairDevice("S0");
		assertFalse(failingSensor.isFailed());
		assertEquals("true",
				store.getData(failingSensor.getID() + ".isWorking", "false"));
		assertFalse(((List<String>) extractField(i, "failedDevices"))
				.contains(failingSensor.getID()));

	}

	@SuppressWarnings("unchecked")
	@Test
	public void testRepairDeviceStoreFailure() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			DeviceFailureException, StoreFailureException {
		// Tests testRepairDevice(). The store does not
		// work properly.
		store.setIsFailed(true);

		// Let a sensor fail.
		Zone zone1 = null;

		for (Zone z : zones) {
			if (z.getID().equals("Z0"))
				zone1 = z;
		}

		zone1.addSensor("S0", new MockSensorDevice(6), zone1.getLocation());
		Sensor failingSensor = extractField(zone1, "sensor");

		try {
			i.getMeasuredMoistureLevel(zone1.getID());
			fail();
		} catch (DeviceFailureException e) {

		}

		i.repairDevice("S0");
		assertFalse(failingSensor.isFailed());
		assertTrue(i.isStoreFailure());
		assertFalse(((List<String>) extractField(i, "failedDevices"))
				.contains(failingSensor.getID()));

	}

	@SuppressWarnings("unchecked")
	@Test
	public void testRepairDeviceWorkingDevice()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, StoreFailureException {
		// Tests testRepairDevice(). No device fails, so no
		// effect is expected.

		i.repairDevice("V22");

		assertEquals("true", store.getData("V22.isWorking", "false"));
		assertFalse(((List<String>) extractField(i, "failedDevices"))
				.contains("V22"));

		// All valves and sensors should still be working.
		for (Zone z : zones) {
			HashSet<Valve> valves = extractField(z, "valves");
			Sensor sensor = extractField(z, "sensor");
			assertFalse(sensor.isFailed());
			for (Valve v : valves) {
				assertFalse(v.isFailed());
			}
		}

	}

	@SuppressWarnings("unchecked")
	@Test
	public void testRepairDeviceIllegalDevice()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, StoreFailureException {
		// Tests whether repairDevice() checks its preconditions.
		Zone zone1 = null;

		// Let a sensor fail.
		for (Zone z : zones) {
			if (z.getID().equals("Z0"))
				zone1 = z;
		}

		zone1.addSensor("S0", new MockSensorDevice(6), zone1.getLocation());
		Sensor failingSensor = extractField(zone1, "sensor");

		try {
			i.getMeasuredMoistureLevel(zone1.getID());
			fail();
		} catch (DeviceFailureException e) {

		}

		try {
			i.repairDevice("V222");
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			i.repairDevice(null);
			fail();
		} catch (IllegalArgumentException e) {

		}

		// The sensor should still be failing.
		assertTrue(failingSensor.isFailed());
		assertEquals("false",
				store.getData(failingSensor.getID() + ".isWorking", "false"));
		assertTrue(((List<String>) extractField(i, "failedDevices"))
				.contains(failingSensor.getID()));

	}

	@Test
	public void testUpdateTimeNotReached() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests update(). The current time is not an irrigation time.
		SimTime time = SimTime.instance();
		time.setDay(Day.MONDAY);
		time.setHour(12);
		time.setMinute(30);
		time.setSecond(0);

		HashSet<Day> days = new HashSet<Day>();
		days.add(Day.SUNDAY);
		days.add(Day.FRIDAY);
		days.add(Day.MONDAY);

		i.setIrrigationTime(1300);
		i.setIrrigationDays(days);
		i.setMode(Mode.AUTOMATIC);

		i.update(null, null);
		
		// No auto cycle should be running.
		assertNull(extractField(i, "cycle"));
	}

	@Test
	public void testUpdateDayNotReached() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests update(). The current day is not an irrigation day.
		SimTime time = SimTime.instance();
		time.setDay(Day.THURSDAY);
		time.setHour(12);
		time.setMinute(0);
		time.setSecond(0);

		HashSet<Day> days = new HashSet<Day>();
		days.add(Day.SUNDAY);
		days.add(Day.FRIDAY);
		days.add(Day.MONDAY);

		i.setIrrigationTime(1200);
		i.setIrrigationDays(days);
		i.setMode(Mode.AUTOMATIC);

		i.update(null, null);
		
		// No irrigation cycle should be running.
		assertNull(extractField(i, "cycle"));
	}

	@Test
	public void testUpdateModeManual() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests update(). The current mode is set to manual, so
		// no AutoCycle should start.
		SimTime time = SimTime.instance();
		time.setDay(Day.MONDAY);
		time.setHour(12);
		time.setMinute(0);
		time.setSecond(0);

		HashSet<Day> days = new HashSet<Day>();
		days.add(Day.SUNDAY);
		days.add(Day.FRIDAY);
		days.add(Day.MONDAY);

		i.setIrrigationTime(1200);
		i.setIrrigationDays(days);
		i.setMode(Mode.MANUAL);

		i.update(null, null);
		
		// No irrigation cycle should be running.
		assertNull(extractField(i, "cycle"));
	}

	@Test
	public void testUpdate() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests update(). Irrigation time and day are reached,
		// so a new AutoCycle should be started.
		
		SimTime time = SimTime.instance();
		time.setDay(Day.MONDAY);
		time.setHour(12);
		time.setMinute(0);
		time.setSecond(0);

		HashSet<Day> days = new HashSet<Day>();
		days.add(Day.SUNDAY);
		days.add(Day.FRIDAY);
		days.add(Day.MONDAY);

		i.setIrrigationTime(1200);
		i.setIrrigationDays(days);
		i.setMode(Mode.AUTOMATIC);

		i.setAllocation(15000);

		i.update(null, null);
		AutoCycle cycle = extractField(i, "cycle");
		assertNotNull(cycle);
		assertFalse(cycle.isDone());
		assertEquals(null, i.getNextFailedDevice());
	}

	@Test
	public void testUpdateStartFailure() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests update(). An AutoCycle is started, but there's
		// also a failure in one of the devices.
		SimTime time = SimTime.instance();
		time.setDay(Day.MONDAY);
		time.setHour(12);
		time.setMinute(0);
		time.setSecond(0);

		HashSet<Day> days = new HashSet<Day>();
		days.add(Day.SUNDAY);
		days.add(Day.FRIDAY);
		days.add(Day.MONDAY);

		i.setIrrigationTime(1200);
		i.setIrrigationDays(days);
		i.setMode(Mode.AUTOMATIC);

		i.setAllocation(15000);

		// Add a failing valve.
		Zone failingZone = zones.iterator().next();
		MockValveDevice mockValve;
		failingZone.addValve("valveFailure", mockValve = new MockValveDevice(),
				"type", 100, failingZone.getLocation());
		mockValve.isFailed = true;

		i.update(null, null);
		AutoCycle cycle = extractField(i, "cycle");
		assertNotNull(cycle);
		assertFalse(cycle.isDone());
		List<String> failedDevices = null;
		try {
			failedDevices = getFailedDevices(i);
		} catch (Exception e) {
			e.printStackTrace();
			fail();
		}
		assertEquals("valveFailure", failedDevices.get(0));
	}

	@Test
	public void testUpdateCallsTickAutoCycle() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests update(). Checks whether an existing AutoCycle's tick method
		// gets called.
		SimTime time = SimTime.instance();
		time.setDay(Day.MONDAY);
		time.setHour(12);
		time.setMinute(0);
		time.setSecond(0);

		HashSet<Day> days = new HashSet<Day>();
		days.add(Day.SUNDAY);
		days.add(Day.FRIDAY);
		days.add(Day.MONDAY);

		i.setIrrigationTime(1200);
		i.setIrrigationDays(days);
		i.setMode(Mode.AUTOMATIC);

		i.setAllocation(15000);

		i.update(null, null);
		Field cycleField = extractFieldAsField(i, "cycle");
		MockAutoCycle mockCycle = new MockAutoCycle(zones, 15000);
		cycleField.setAccessible(true);
		cycleField.set(i, mockCycle);

		// The mock cycle's count should be incremented.
		i.update(null, null);
		assertEquals(1, mockCycle.ticks);
		assertEquals(null, i.getNextFailedDevice());

	}

	@Test
	public void testUpdateCallsTickAutoCycleFailure()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests update(). Checks whether a failure in
		// a running AutoCycle is reported.
		SimTime time = SimTime.instance();
		time.setDay(Day.MONDAY);
		time.setHour(12);
		time.setMinute(0);
		time.setSecond(0);

		HashSet<Day> days = new HashSet<Day>();
		days.add(Day.SUNDAY);
		days.add(Day.FRIDAY);
		days.add(Day.MONDAY);

		i.setIrrigationTime(1200);
		i.setIrrigationDays(days);
		i.setMode(Mode.AUTOMATIC);

		i.setAllocation(15000);

		i.update(null, null);
		
		// Replace the current cycle with a mock cycle.
		Field cycleField = extractFieldAsField(i, "cycle");
		MockAutoCycle mockCycle = new MockAutoCycle(zones, 15000);
		mockCycle.setFails(true);
		cycleField.setAccessible(true);
		cycleField.set(i, mockCycle);

		i.update(null, null);
		assertEquals(1, mockCycle.ticks);
		List<String> failedDevices = null;
		try {
			failedDevices = getFailedDevices(i);
		} catch (Exception e) {
			e.printStackTrace();
			fail();
		}
		assertEquals("MockAutoCycle", failedDevices.get(0));

	}

	@Test
	public void testUpdateCallsTickAutoCycleDone()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests update(). Checks whether a finished AutoCycle is
		// terminated properly.
		SimTime time = SimTime.instance();
		time.setDay(Day.SUNDAY);
		time.setHour(18);
		time.setMinute(45);
		time.setSecond(0);

		HashSet<Day> days = new HashSet<Day>();
		days.add(Day.SUNDAY);
		days.add(Day.FRIDAY);
		days.add(Day.MONDAY);

		i.setIrrigationTime(1845);
		i.setIrrigationDays(days);
		i.setMode(Mode.AUTOMATIC);

		i.setAllocation(15000);

		i.update(null, null);
		
		// Replace the current cycle with a mock reporting
		// it has finished.
		Field cycleField = extractFieldAsField(i, "cycle");
		MockAutoCycle mockCycle = new MockAutoCycle(zones, 15000);
		mockCycle.done = true;
		cycleField.setAccessible(true);
		cycleField.set(i, mockCycle);

		i.update(null, null);
		assertEquals(1, mockCycle.ticks);
		// Now, the cycle should be stopped.
		assertEquals(null, i.getNextFailedDevice());
		assertNull(extractField(i, "cycle"));

	}

	@Test
	public void testUpdateCallsTickManualCycle()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests update(). Checks whether an existing ManualCycle's tick method
		// gets called.
		i.setMode(Mode.MANUAL);

		i.startManualCycle();

		// Replace the current cycle with a mock cycle.
		Field cycleField = extractFieldAsField(i, "cycle");
		MockManualCycle mockCycle = new MockManualCycle(zones);
		cycleField.setAccessible(true);
		cycleField.set(i, mockCycle);

		i.update(null, null);
		assertEquals(1, mockCycle.ticks);
		assertEquals(null, i.getNextFailedDevice());

	}

}
