package irrigation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static tools.HelperFunctions.extractField;

import java.util.Set;

import org.junit.Test;

import device.DeviceFailureException;

public class ManualCycleTest extends IrrigationCycleTest {

	@Override
	protected IrrigationCycle concreteOut(Set<Zone> zones, int theAllocation) {
		return new ManualCycle(zones);
	}

	// The following methods test ManualCycle's tick() method

	@Test
	public void testTickMoistureLevelReached() throws DeviceFailureException {
		// Tests whether a cycle irrigates zones despite
		// they reached their moisture level

		cycle = getOut(zones);

		cycle.start();

		// Set the reported moisture level of the second sensor.
		zoneSensor.get(secondZone).value = 55;

		// After each tick(), all zones should still be irrigating.
		for (int i = 0; i < 10; i++) {
			cycle.tick();
			assertEquals((i + 1) * 15, firstZone.getWaterUsed());
			assertEquals(firstZone.getWaterUsed(), thirdZone.getWaterUsed());
			assertEquals(firstZone.getWaterUsed(), secondZone.getWaterUsed());
		}

		assertEquals(cycle.getWaterUsed(), firstZone.getWaterUsed()
				+ secondZone.getWaterUsed() + thirdZone.getWaterUsed());

	}

	@Test
	public void testTickValveFailureDuringIrrigation()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests if the cycle continues to irrigate
		// when a valve fails.

		cycle = getOut(zones);

		cycle.start();

		for (int i = 0; i < 3; i++) {
			cycle.tick();
		}

		// Let a valve in the first zone fail
		Valve firstValve = zoneValves.get(firstZone).iterator().next();
		MockValveDevice valveDevice = extractField(firstValve, "device");
		valveDevice.isFailed = true;

		int waterUsed = firstZone.getWaterUsed();

		// Since the first zone is not done,
		// it should still be irrigated
		for (int i = 0; i < 3; i++)
			cycle.tick();

		assertEquals(waterUsed + 45, firstZone.getWaterUsed());
	}

	@Test
	public void testTickZoneMoistureLevelReached()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests whether a cycle continues to irrigate a zone
		// if the zone's critical moisture level is reached

		cycle = getOut(zones);

		cycle.start();
		for (int i = 0; i < 5; i++)
			cycle.tick();

		// In the past minute, the first zone got sufficiently irrigated
		zoneSensor.get(firstZone).value = 55;

		for (int i = 0; i < 5; i++)
			cycle.tick();

		assertEquals(150, firstZone.getWaterUsed());
		assertEquals(150, secondZone.getWaterUsed());
		assertEquals(150, thirdZone.getWaterUsed());

	}

	@Test
	public void testTickSensorDeviceFailure() throws DeviceFailureException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests whether a manual cycle continues to irrigate
		// its zones in case a sensor failure happens

		cycle = getOut(zones);

		cycle.start(); 

		for (int i = 0; i < 5; i++)
			cycle.tick();

		for (int i = 0; i < 10; i++) {
			if (i == 6) {
				// Let a sensor fail.
				secondZone.addSensor("s0", new MockSensorDevice(5), "l0");
			}
			cycle.tick();
		}

		for (@SuppressWarnings("unused") Zone z : zones)
			assertEquals(45, secondZone.getWaterUsed());

	}

	// The following tests test the start() method of class ManualCycle

	@Test
	public void testStartSensorDeviceFailure() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {

		// Tests whether start initializes a cycle properly
		// if a sensor device fails.

		cycle = getOut(zones);

		MockSensorDevice failedSensor = new MockSensorDevice(6);

		firstZone.addSensor("s0", failedSensor, firstZone.getLocation());

		cycle.start();

		// Although there is a failed sensor,
		// the cycle should work as usual.
		for (int i = 0; i < 5; i++)
			cycle.tick();

		assertEquals(225, cycle.getWaterUsed());
		assertEquals(75, firstZone.getWaterUsed());
		assertEquals(75, secondZone.getWaterUsed());
		assertEquals(75, thirdZone.getWaterUsed());

	}

	@Test
	public void testStartValveDeviceFailure() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {

		// Tests start in the case a valve device fails

		cycle = getOut(zones);

		Valve failedValve;
		((MockValveDevice) extractField(failedValve = zoneValves.get(firstZone)
				.iterator().next(), "device")).isFailed = true;

		// There is a failed valve in the first zone
		// so an exception is expected since start()
		// has to do some sort of preparation for the
		// irrigation cycle which involves most
		// probably the valves.
		try {
			cycle.start();
			fail();
		} catch (DeviceFailureException e) {
			assertEquals(1, e.getDevices().size());
			assertTrue(e.getDevices().contains(failedValve.getID()));
		}

		for (int i = 0; i < 5; i++)
			cycle.tick();

		assertEquals(75, thirdZone.getWaterUsed());
		assertEquals(75, secondZone.getWaterUsed());
		assertEquals(50, firstZone.getWaterUsed());
	}

	// The following tests test ManualCycle's isDone() method

	@Test
	public void testIsDoneMoistureLevelReached() throws DeviceFailureException {
		// Tests isDone() in the case where every zone has reached
		// its critical moisture level

		cycle = getOut(zones);

		cycle.start();

		cycle.tick();
		cycle.tick();

		assertFalse(cycle.isDone());

		for (Zone z : zones)
			zoneSensor.get(z).value = 50;

		cycle.tick();
		cycle.tick();
		cycle.tick();

		assertFalse(cycle.isDone());
	}

	@Test
	public void testIsDoneNoIrrigationNeeded() throws DeviceFailureException {
		// Tests isDone() in a situation where
		// no zone needs irrigation

		cycle = getOut(zones);

		for (Zone z : zones)
			zoneSensor.get(z).value = 55;

		cycle.start();

		assertFalse(cycle.isDone());

	}

	@Test
	public void testIsDoneAfterEnd() throws DeviceFailureException {
		// Tests isDone() after a call to
		// end

		cycle = getOut(zones);

		cycle.start();

		assertFalse(cycle.isDone());

		cycle.end();

		assertFalse(cycle.isDone());
	}

	@Test
	public void testIsDoneZoneChange() throws DeviceFailureException {
		// Tests isDone() when one of the zones
		// is done
		cycle = getOut(zones);

		cycle.start();

		for (int i = 0; i < 3; i++)
			cycle.tick();

		zoneSensor.get(firstZone).value = 55;

		cycle.tick();

		assertFalse(cycle.isDone());
	}

	@Test
	public void testIsDoneNotStarted() {
		cycle = getOut(zones);

		assertFalse(cycle.isDone());
	}

	// The following code tests ManualCycle's end() method

	@Override
	@Test
	public void testEndWithoutStart() throws DeviceFailureException {
		super.testEndWithoutStart();

		assertFalse(cycle.isDone());
	}

	@Override
	@Test
	public void testEndAllZonesDone() throws DeviceFailureException {
		super.testEndAllZonesDone();

		assertFalse(cycle.isDone());
	}

	@Test
	public void testEndSomeZonesDone() throws DeviceFailureException {
		// Tests how end() works in the middle of an irrigation cycle

		cycle = getOut(zones);

		cycle.start();

		for (int i = 0; i < 10; i++) {
			cycle.tick();
			if (i == 5)
				zoneSensor.get(firstZone).value = 55;
		}

		// Should work
		cycle.end();

		for (int i = 0; i < 5; i++)
			cycle.tick();

		assertEquals(45, secondZone.getWaterUsed());

		// All Valves should be closed
		for (Zone z : zones)
			for (Valve v : zoneValves.get(z))
				assertFalse(v.isOpen());

		assertFalse(cycle.isDone());
	}

	@Override
	@Test
	public void testEndSomeZonesDoneSensorFailure()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		super.testEndSomeZonesDoneSensorFailure();

		assertFalse(cycle.isDone());
	}

	@Test
	public void testEndSomeZonesDoneValveFailure()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests how end() works in the middle of an irrigation cycle
		// if one of the current zone's valves fails

		cycle = getOut(zones);

		cycle.start();

		Valve failedValve1 = null;
		Valve failedValve2 = null;

		for (int i = 0; i < 10; i++) {
			cycle.tick();
			if (i == 5)
				zoneSensor.get(firstZone).value = 55;
			if (i == 9) {
				// Let some valves fail.
				((MockValveDevice) extractField(
						failedValve1 = zoneValves.get(secondZone).iterator()
								.next(), "device")).isFailed = true;
				((MockValveDevice) extractField(
						failedValve2 = zoneValves.get(thirdZone).iterator()
								.next(), "device")).isFailed = true;
			}
		}

		// Should work
		try {
			cycle.end();
			fail();
		} catch (DeviceFailureException e) {
			assertEquals(2, e.getDevices().size());
			assertTrue(e.getDevices().contains(failedValve1.getID()));
			assertTrue(e.getDevices().contains(failedValve2.getID()));
		}

		for (int i = 0; i < 5; i++)
			cycle.tick();

		// All Valves should be closed
		for (Zone z : zones)
			for (Valve v : zoneValves.get(z))
				assertFalse(v.isOpen());

		assertFalse(cycle.isDone());

	}
}
