package irrigation;

import static org.junit.Assert.*;
import static tools.HelperFunctions.*;

import org.junit.Test;

import device.DeviceFailureException;

public class SensorTest {

	@Test
	public void testSensorConstructor() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests Sensor's constructor.
		MockSensorDevice mock = new MockSensorDevice(0);

		Sensor sensor = new Sensor("id", mock, "loc");

		assertEquals("id", sensor.getID());
		assertEquals("loc", sensor.getLocation());
		assertFalse(sensor.isFailed());
		assertTrue(sensor.isRecorded());
		assertEquals(mock, extractField(sensor, "device"));
	}

	@Test
	public void testSensorConstructorNullValues()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests Sensor's constructor with different arguments.
		Sensor sensor = new Sensor(null, null, null);

		assertEquals(null, sensor.getID());
		assertEquals(null, sensor.getLocation());
		assertFalse(sensor.isFailed());
		assertTrue(sensor.isRecorded());
		assertNull(extractField(sensor, "device"));
	}

	@Test
	public void testGetSensorReport() throws DeviceFailureException {

		// Tests the creation of a sensor report for a working sensor
		
		MockSensorDevice mock = new MockSensorDevice(0);
		mock.value = 10;
		Sensor sensor = new Sensor("id", mock, "location");
		SensorReport report = sensor.getSensorReport("zone");

		assertEquals(sensor.getID(), report.id);
		assertEquals(sensor.getLocation(), report.location);
		assertEquals("zone", report.zoneID);
		assertTrue(report.isRecorded);
		assertFalse(report.isFailed);
		assertEquals(10, report.level);
	}

	@Test
	public void testGetSensorReportFailedSensor() throws DeviceFailureException {
		// Test the creation of a sensor report for a failed sensor
		
		MockSensorDevice mock = new MockSensorDevice(0);
		mock.value = 22;
		
		Sensor sensor = new Sensor("id2", mock, "loc");
		sensor.setIsRecorded(false);
		sensor.setIsFailed(true);

		SensorReport report = sensor.getSensorReport("zone2");

		assertEquals(sensor.getID(), report.id);
		assertEquals(sensor.getLocation(), report.location);
		assertEquals("zone2", report.zoneID);
		assertFalse(report.isRecorded);
		assertTrue(report.isFailed);
		assertEquals(sensor.read(), report.level);
	}

	@Test
	public void testGetSensorReportFailedSensorDevice() throws DeviceFailureException {
		
		// Test the creation of a SensorReport in case
		// the underlying sensor device fails.
		
		MockSensorDevice mock = new MockSensorDevice(3);
		Sensor sensor = new Sensor("id3", mock, "loca");
		sensor.setIsFailed(false);
		SensorReport report = sensor.getSensorReport("zone3");

		assertEquals(sensor.getID(), report.id);
		assertEquals(sensor.getLocation(), report.location);
		assertEquals("zone3", report.zoneID);
		assertTrue(report.isRecorded);
		assertTrue(report.isFailed);
		assertEquals(1, report.level);

	}

	@Test
	public void testRead() throws DeviceFailureException {
		// Tests read().
		MockSensorDevice mock = new MockSensorDevice(0);

		Sensor sensor = new Sensor("id", mock, "location");

		mock.value = 6;
		assertEquals(6, sensor.read());

		mock.value = 4;
		assertEquals(4, sensor.read());
	}

	@Test
	public void testReadFailedSensorDevice() {
		// Tests what happens when an underlying Sensor Device
		// fails more than 2 times
		
		MockSensorDevice mock = new MockSensorDevice(3);
		Sensor sensor = new Sensor("id", mock, "location");

		try {
			sensor.read();
			fail();
		} catch (DeviceFailureException e) {
			// Expected
			assertEquals(1, e.getDevices().size());
			assertTrue(e.getDevices().contains("id"));
		}
	}

	@Test
	public void testReadSometimesFailedSensorDevice() throws DeviceFailureException {
		// Tests what happens if the attached Sensor Device
		// failes less than 3 times
		
		MockSensorDevice mock = new MockSensorDevice(2);
		mock.value = 4;
		Sensor sensor = new Sensor("id", mock, null);

		assertEquals(4, sensor.read());

		mock = new MockSensorDevice(1);
		mock.value = 5;
		sensor = new Sensor("id", mock, null);

		assertEquals(5, sensor.read());

	}

	@Test
	public void testSetIsFailed() {
		Sensor sensor = new Sensor("id", null, "loc");

		sensor.setIsFailed(true);
		assertTrue(sensor.isFailed());

		sensor.setIsFailed(false);
		assertFalse(sensor.isFailed());
	}

	@Test
	public void testSetIsRecorded() {
		Sensor sensor = new Sensor("id", null, "loc");

		sensor.setIsRecorded(false);
		assertFalse(sensor.isRecorded());

		sensor.setIsRecorded(true);
		assertTrue(sensor.isRecorded());
	}

}
