package irrigation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static tools.HelperFunctions.extractField;

import org.junit.Test;

import device.DeviceFailureException;

public class ValveTest {

	@Test
	public void testValveConstructor() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests Valve's constructor.

		MockValveDevice mock = new MockValveDevice();
		Valve valve = new Valve("id", mock, "type", Integer.MAX_VALUE, "loc");

		assertEquals("id", valve.getID());
		assertEquals(mock, extractField(valve, "device"));
		assertEquals("type", extractField(valve, "type"));
		assertEquals("loc", valve.getLocation());
		assertEquals(Integer.MAX_VALUE, extractField(valve, "flowRate"));
		assertFalse(valve.isFailed());
		assertFalse(valve.isOpen());
		assertTrue(valve.isRecorded());
		assertEquals(0, extractField(valve, "minutesOpen"));
		assertEquals(0, valve.getWaterUsed());
	}

	@Test
	public void testValveConstructorNullArguments()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests Valve's constructor with alternative arguments.
		Valve valve = new Valve(null, null, null, Integer.MIN_VALUE, null);

		assertNull(valve.getID());
		assertNull(extractField(valve, "device"));
		assertNull(extractField(valve, "type"));
		assertNull(valve.getLocation());
		assertEquals(Integer.MIN_VALUE, extractField(valve, "flowRate"));
		assertFalse(valve.isFailed());
		assertFalse(valve.isOpen());
		assertTrue(valve.isRecorded());
		assertEquals(0, extractField(valve, "minutesOpen"));
		assertEquals(0, valve.getWaterUsed());

	}

	@Test
	public void testGetWaterUsed() throws DeviceFailureException {
		// Tests getWaterUsed().
		Valve valve = new Valve("id", new MockValveDevice(), "type", 3, "loc");

		valve.open();

		assertEquals(0, valve.getWaterUsed());

		// Now let the valve irrigate some water.
		valve.tick();
		valve.tick();
		valve.tick();

		assertEquals(9, valve.getWaterUsed());

		valve.reset();

		assertEquals(0, valve.getWaterUsed());
	}

	@Test
	public void testSetIsFailed() {
		// Tests setIsFailed();
		Valve valve = new Valve("id", null, "type", 0, "loc");

		valve.setIsFailed(true);
		assertTrue(valve.isFailed());

		valve.setIsFailed(false);
		assertFalse(valve.isFailed());

	}

	@Test
	public void testSetIsRecorded() {
		// Tests setIsRecorded().
		Valve valve = new Valve(null, null, null, 0, null);

		valve.setIsRecorded(false);
		assertFalse(valve.isRecorded());

		valve.setIsRecorded(true);
		assertTrue(valve.isRecorded());

	}

	@Test
	public void testOpen() throws DeviceFailureException {
		// Tests open()
		MockValveDevice mock = new MockValveDevice();
		Valve valve = new Valve("id", mock, "type", 2, "location");

		valve.open();

		// Both the valve and the valve device should be open.
		assertTrue(mock.isOpen);
		assertTrue(valve.isOpen());
	}

	@Test
	public void testClose() throws DeviceFailureException {
		// Tests close();
		MockValveDevice mock = new MockValveDevice();
		Valve valve = new Valve("id", mock, "type", 2, "location");
		valve.open();

		valve.close();

		// Both the valve and the valve device should be closed.
		assertFalse(mock.isOpen);
		assertFalse(valve.isOpen());
	}

	@Test
	public void testOpenFailedValve() {
		// Test what happens when a failed
		// valve is opened

		MockValveDevice mock = new MockValveDevice();
		Valve valve = new Valve("id", mock, "type", 2, "location");

		mock.isFailed = true;

		try {
			valve.open();
			fail();
		} catch (DeviceFailureException e) {
			// Expected
			assertEquals(1, e.getDevices().size());
			assertTrue(e.getDevices().contains("id"));
		}
		// The valve failed when it should open,
		// so it is still closed ...
		assertFalse(valve.isOpen());
		assertTrue(valve.isFailed());
	}

	@Test
	public void testCloseFailedValve() throws DeviceFailureException {
		// Test what happens when a failed
		// valve is closed

		MockValveDevice mock = new MockValveDevice();
		Valve valve = new Valve("id", mock, "type", 2, "location");
		valve.open();

		mock.isFailed = true;

		try {
			valve.close();
			fail();
		} catch (DeviceFailureException e) {
			// Expected
			assertEquals(1, e.getDevices().size());
			assertTrue(e.getDevices().contains("id"));
		}
		// The valve failed while closing, so it should still
		// be open (at least to some extend) ...
		assertTrue(valve.isOpen());
		assertTrue(valve.isFailed());
	}

	@Test
	public void testOpenCloseValveWithNullDevices()
			throws DeviceFailureException {

		// Tests how a valve opens and closes when no ValveDevice
		// is attached

		Valve valve = new Valve(null, null, null, 2, null); // Correct according
															// to
															// contract

		valve.open();

		assertTrue(valve.isOpen());

		valve.close();

		assertFalse(valve.isOpen());

	}

	@Test
	public void testGetValveReportOpenFailedValve()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {

		// Tests the creation of a valve report for a failed, open valve

		Valve valve = new Valve("i", new MockValveDevice(), "t", 3, "loc");

		valve.open();
		valve.setIsFailed(true);

		ValveReport report = valve.getValveReport("z1");

		assertEquals(valve.isOpen(), report.isOpen);
		assertEquals(valve.getID(), report.id);
		assertEquals(valve.getLocation(), report.location);
		assertEquals(extractField(valve, "flowRate"), report.flowRate);
		assertEquals("z1", report.zoneID);
		assertEquals(0, report.minutesOpen);
		assertEquals(valve.isFailed(), report.isFailed);
		assertEquals(valve.isRecorded(), report.isRecorded);
	}

	@Test
	public void testGetValveReportClosedUsedValve()
			throws DeviceFailureException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests the creation of a valve report for a used, currently closed
		// valve

		Valve valve = new Valve("i2", new MockValveDevice(), "T", 5, "loc2");

		// Let the valve use some water.
		valve.open();
		for (int i = 0; i < 5; i++)
			valve.tick();
		valve.close();

		ValveReport report = valve.getValveReport("z2");

		assertEquals(valve.isOpen(), report.isOpen);
		assertEquals(valve.getID(), report.id);
		assertEquals(valve.getLocation(), report.location);
		assertEquals(extractField(valve, "flowRate"), report.flowRate);
		assertEquals("z2", report.zoneID);
		assertEquals(5, report.minutesOpen);
		assertEquals(valve.isFailed(), report.isFailed);
		assertEquals(valve.isRecorded(), report.isRecorded);

	}

	@Test
	public void testReset() throws DeviceFailureException {
		Valve valve = new Valve("id", new MockValveDevice(), "type", 1,
				"location");
		// Tests reset().
		valve.open();

		// Use the valve a few times
		valve.tick();
		valve.tick();
		valve.tick();

		assertEquals(3, valve.getWaterUsed());

		valve.reset();

		assertEquals(0, valve.getWaterUsed());
	}

	@Test
	public void testTick() throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {
		// Tests tick(). 
		
		Valve valve = new Valve(null, new MockValveDevice(), null, 2, null);

		valve.open();
		valve.tick();
		assertEquals(2, valve.getWaterUsed());

		valve.tick();
		assertEquals(4, valve.getWaterUsed());

		// Is it possible to use tick after a
		// valve has been reset?
		valve.reset();
		valve.tick();

		assertEquals(2, valve.getWaterUsed());
	}

	@Test
	public void testTickLargeAmountOfWaterUsed()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		
		// Testing tick by using a very large amount of used water 
		Valve valve = new Valve(null, new MockValveDevice(), null, Integer.MAX_VALUE - 100000, null);

		valve.tick();
		valve.tick();

		// A negative amount of used water
		// makes no sense ...
		assertTrue(valve.getWaterUsed() >= 0);

	}

}
