package irrigation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static tools.HelperFunctions.extractField;
import static tools.HelperFunctions.extractFieldAsField;

import java.lang.reflect.Field;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

import org.junit.Before;
import org.junit.Test;

import device.DeviceFailureException;

public class ZoneTest {

	private Zone zone;
	private MockValveDevice v1;
	private MockValveDevice v2;
	private MockValveDevice v3;
	private HashSet<MockValveDevice> mocks;
	private Valve[] valves;
	private MockSensorDevice mockSensor;
	private Sensor sensor;
	private HashSet<Valve> valveSet;

	@Before
	public void init() throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Set up a zone ...
		zone = new Zone("id", "loc");

		v1 = new MockValveDevice();
		v2 = new MockValveDevice();
		v3 = new MockValveDevice();

		// ... with its sensor ...
		mockSensor = new MockSensorDevice(0);
		zone.addSensor("s1", mockSensor, "loc");
		sensor = extractField(zone, "sensor");

		// ... and 3 valves.
		zone.addValve("i1", v1, "t1", 2, "loc");
		zone.addValve("i2", v2, "t2", 3, "loc");
		zone.addValve("i3", v3, "t3", 4, "loc");

		mocks = new HashSet<MockValveDevice>();

		mocks.add(v1);
		mocks.add(v2);
		mocks.add(v3);

		// Prepare some helpful objects.
		valveSet = extractField(zone, "valves");
		Object[] vObjects = valveSet.toArray();
		valves = new Valve[vObjects.length];
		for (int i = 0; i < vObjects.length; i++) {
			Valve v = (Valve) vObjects[i];
			if (v.getID().equals("i1"))
				valves[0] = v;
			if (v.getID().equals("i2"))
				valves[1] = v;
			if (v.getID().equals("i3"))
				valves[2] = v;
		}
	}

	@Test
	public void testConstructor() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Test Zone's constructor with valid arguments.
		Zone zone = new Zone("id", "loc");

		assertEquals("id", zone.getID());
		assertEquals("loc", zone.getLocation());
		assertEquals(0, zone.getCriticalMoistureLevel());
		assertEquals(0, zone.getWaterUsed());
		assertEquals(0, extractField(zone, "allocation"));
		valveSet = extractField(zone, "valves");
		assertNotNull(valveSet);
		assertEquals(0, valveSet.size());
		assertNull(extractField(zone, "sensor"));
	}

	@Test
	public void testConstructorNull() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests Zone's constructor with null arguments.
		zone = new Zone(null, null);

		assertNull(zone.getID());
		assertNull(zone.getLocation());
		assertEquals(0, zone.getCriticalMoistureLevel());
		assertEquals(0, zone.getWaterUsed());
		assertEquals(0, extractField(zone, "allocation"));
		valveSet = extractField(zone, "valves");
		assertNotNull(valveSet);
		assertEquals(0, valveSet.size());
		assertNull(extractField(zone, "sensor"));

	}

	@Test
	public void testGetNumWorkingValvesNoFailures()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException,
			DeviceFailureException {

		// Tests getNumWorkingValves(). The zone has been created just before, so
		// all valves should work
		assertEquals(3, zone.getNumWorkingValves());
	}

	@Test
	public void testGetNumWorkingValvesSingleFailure() {
		// Tests getNumWorkingValves(). One of three valves has failed
		valves[1].setIsFailed(true);

		assertEquals(2, zone.getNumWorkingValves());
	}

	@Test
	public void testGetNumWorkingValvesMultipleFailures() {
		// Tests getNumWorkingValves(). All valves have failed.
		for (Valve v : valves)
			v.setIsFailed(true);

		assertEquals(0, zone.getNumWorkingValves());
	}

	@Test
	public void testGetNumWorkingValvesNewZone() {
		// Checks whether getNumWorkingValves is working
		// when called on a newly created zone.
		zone = new Zone("id", "loc");

		assertEquals(0, zone.getNumWorkingValves());
	}

	@Test
	public void testGetNumWorkingValvesValveDeviceFailures()
			throws DeviceFailureException {
		// Tests getNumWorkingValves(). The underlying valve devices
		// fail.
		
		for (Iterator<MockValveDevice> i = mocks.iterator(); i.hasNext();) {
			i.next().isFailed = true;
		}

		// Although the underlying ValveDevices failed
		// getNumWorkingShould still return 3 since no failures
		// where detected by opening or closing them
		assertEquals(3, zone.getNumWorkingValves());

		try {
			zone.openAllValves();
			fail();
		} catch (DeviceFailureException e) {
		}

		// Now, all valves should be recognized as failed
		assertEquals(0, zone.getNumWorkingValves());
	}

	@Test
	public void testGetZoneReport() {
		// Tests getZoneReport().
		
		ZoneReport report = zone.getZoneReport();

		// The report should contain the zone's ID, location
		// and critical moisture level
		assertEquals(zone.getID(), report.id);
		assertEquals(zone.getLocation(), report.location);
		assertEquals(zone.getCriticalMoistureLevel(), report.criticalLevel);
	}

	@Test
	public void testGetValveReportsCommonAttributes() {
		// Tests getZoneReports().
		
		Collection<ValveReport> reports = zone.getValveReports();

		// Check attributes that are equal for all reports
		for (ValveReport v : reports) {
			assertEquals(zone.getID(), v.zoneID);
			assertEquals(zone.getLocation(), v.location);
		}
	}

	private boolean areValveReportsEqual(ValveReport r1, ValveReport r2)
			throws IllegalArgumentException, IllegalAccessException {
		// Compares to ValveReports field by field.
		Field[] fields = r1.getClass().getDeclaredFields();
		for (Field f : fields) {
			f.setAccessible(true);
			if (!f.get(r1).equals(f.get(r2)))
				return false;
		}

		return true;
	}

	@Test
	public void testGetValveReportsForEachValve()
			throws IllegalArgumentException, IllegalAccessException {
		// Tests whether getValveReports provides
		// correct reports.
		
		valves[0].setIsFailed(true);
		valves[2].setIsRecorded(false);

		Collection<ValveReport> reports = zone.getValveReports();

		// Each valve (3 in total) in the zone should be described by a report
		assertEquals(3, reports.size());

		// reportMap documents for each Valve whether its
		// report was found in the collection or not
		HashMap<String, Boolean> reportMap = new HashMap<String, Boolean>();

		for (Valve v : valves) {
			reportMap.put(v.getID(), false);
		}

		for (ValveReport r : reports)
			for (Valve v : valves)
				if (r.id.equals(v.getID())) {
					reportMap.put(v.getID(), true);
					// Since report r describes a valve
					// with the same id as v, their reports
					// should also be the same
					assertTrue(areValveReportsEqual(r,
							v.getValveReport(zone.getID())));
					break;
				}

		// Finally, all valves should have their own report
		for (Valve v : valves) {
			assertTrue(reportMap.get(v.getID()));
		}

	}

	@Test
	public void testGetFailedValveReportsAllValvesWorking() {
		// Tests getFailedValveReports().
		Collection<ValveReport> reports = zone.getFailedValveReports();

		// All valves are working
		assertTrue(reports.isEmpty());
	}

	@Test
	public void testGetFailedValveReportsSomeValvesFail()
			throws IllegalArgumentException, IllegalAccessException {
		// Tests getFailedValveReports(). Some valves are failing.
		
		valves[1].setIsFailed(true);
		valves[2].setIsFailed(true);

		Collection<ValveReport> reports = zone.getFailedValveReports();

		// There should be two reports about failed valves
		assertEquals(2, reports.size());

		// reportMap documents for each Valve whether its
		// report was found in the collection or not
		HashMap<String, Boolean> reportMap = new HashMap<String, Boolean>();

		reportMap.put(valves[1].getID(), false);
		reportMap.put(valves[2].getID(), false);

		Valve[] failedValves = new Valve[] { valves[1], valves[2] };

		for (ValveReport r : reports)
			for (Valve v : failedValves)
				if (r.id.equals(v.getID())) {
					reportMap.put(v.getID(), true);
					// Since report r describes a valve
					// with the same id as v, their reports
					// should also be the same
					assertTrue(areValveReportsEqual(r,
							v.getValveReport(zone.getID())));
					break;
				}

		// Finally, all valves should have their own report
		for (Valve v : failedValves) {
			assertTrue(reportMap.get(v.getID()));
		}

	}

	@Test
	public void testGetFailedSensorReportsNoSensorFailes()
			throws DeviceFailureException {
		// Tests getFailedSensorReport().
		
		zone.addSensor("s2", new MockSensorDevice(0), zone.getLocation());

		Collection<SensorReport> reports = zone.getFailedSensorReports();

		// The sensor did not fail at all, so no
		// report should be generated.
		assertEquals(0, reports.size());
	}

	@Test
	public void testGetFailedSensorReportsSomeSensorsFail()
			throws DeviceFailureException {
		// Tests getFailedSensorReport(). The sensor is failing.
		
		sensor.setIsFailed(true);

		// The underlying SensorDevice does not fail, so no
		// DeviceFailureException is thrown
		Collection<SensorReport> reports = zone.getFailedSensorReports();

		// We have one sensor, so one report is expected
		assertEquals(1, reports.size());

		SensorReport failureReport = reports.iterator().next();

		// Check whether the report is accurate
		assertEquals(sensor.getID(), failureReport.id);
		assertEquals(sensor.getLocation(), failureReport.location);
		assertEquals(zone.getID(), failureReport.zoneID);
		assertEquals(sensor.isRecorded(), failureReport.isRecorded);
		assertEquals(true, failureReport.isRecorded);
		assertEquals(sensor.read(), failureReport.level);

	}

	@Test
	public void testGetFailedSensorReportsSensorDeviceFailes() {
		// Tests getFailedSensorReport(). The underlying
		// SensorDevice is failing.
		mockSensor = new MockSensorDevice(10);

		zone.addSensor("s1", mockSensor, zone.getLocation());

		// Make sure sensor.isFailed() == true
		try {
			zone.getMeasuredMoistureLevel();
			fail(); // Exception expected since the underlying
					// SensorDevice fails
		} catch (DeviceFailureException e) {
		}

		try {
			zone.getFailedSensorReports();
			fail(); // Exception expected since the underlying
					// SensorDevice fails
		} catch (DeviceFailureException e) {
			assertEquals(1, e.getDevices().size());
			assertTrue(e.getDevices().contains("s1"));
		}
	}

	@Test
	public void testGetFailedDeviceReportNoFailingDevices() {
		try {
			zone.getFailedDeviceReport("s1");
			fail(); 
		} catch (IllegalArgumentException e) {
			// Exception expected since s1 is not failing
			// Check whether e delivers the proper message
			assertEquals("Device s1 has not failed.", e.getMessage());
		}

		for (Valve v : valves)
			try {
				zone.getFailedDeviceReport(v.getID());
				fail(); 
			} catch (IllegalArgumentException e) {
				// Exception expected since the valves are
				// all working fine.
				// Check whether e delivers the proper message
				assertEquals("Device " + v.getID() + " has not failed.",
						e.getMessage());
			}
	}

	@Test
	public void testGetFailedDeviceReportSensorFailing() {
		// Tests getFailedDeviceReport(). There is a failed sensor.
		
		sensor.setIsFailed(true);

		FailedDeviceReport report = zone.getFailedDeviceReport("s1");

		// Check whether the report is appropriate
		assertEquals("s1", report.deviceID);
		assertEquals(sensor.isRecorded(), report.isRecorded);
		assertEquals(zone.getLocation(), report.location);
		assertEquals(zone.getID(), report.zoneID);
	}

	@Test
	public void testGetFailedDeviceReportValveFailing() {
		// Tests getFailedDeviceReport(). There is a failed valve.
		
		Valve failedValve = valves[2];
		failedValve.setIsFailed(true);

		FailedDeviceReport report = zone.getFailedDeviceReport(failedValve
				.getID());

		// Check whether the report is appropriate
		assertEquals(failedValve.getID(), report.deviceID);
		assertEquals(failedValve.isRecorded(), report.isRecorded);
		assertEquals(zone.getLocation(), report.location);
		assertEquals(zone.getID(), report.zoneID);
	}

	@Test
	public void testSetIsFailed() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {

		// Tests setIsFailed().
		
		valves[2].setIsFailed(true);
		sensor.setIsFailed(true);

		// Test whether the correct devices' failure
		// state is changed

		zone.setIsFailed(valves[0].getID(), true);
		assertTrue(valves[0].isFailed());
		assertFalse(valves[1].isFailed());
		assertTrue(valves[2].isFailed());
		assertTrue(sensor.isFailed());

		zone.setIsFailed(valves[2].getID(), false);
		assertTrue(valves[0].isFailed());
		assertFalse(valves[1].isFailed());
		assertFalse(valves[2].isFailed());
		assertTrue(sensor.isFailed());

		zone.setIsFailed(sensor.getID(), false);
		assertTrue(valves[0].isFailed());
		assertFalse(valves[1].isFailed());
		assertFalse(valves[2].isFailed());
		assertFalse(sensor.isFailed());
	}

	@Test
	public void testSetIsFailedNull() {
		// Tests how setIsFailed deals with null
		
		zone.setIsFailed(null, true);

		// No changes expected
		assertFalse(valves[0].isFailed());
		assertFalse(valves[1].isFailed());
		assertFalse(valves[2].isFailed());
		assertFalse(sensor.isFailed());
	}

	@Test
	public void testSetIsFailedWrongName() {
		// Test how setIsFailed deals with a
		// not existing device name.

		zone.setIsFailed("dummy", true);

		// No changes expected
		assertFalse(valves[0].isFailed());
		assertFalse(valves[1].isFailed());
		assertFalse(valves[2].isFailed());
		assertFalse(sensor.isFailed());
	}

	@Test
	public void testSetIsFailedNewlyCreatedZone() {
		// Tests setIsFailed on a newly created, empty zone
		// No exceptions expected

		zone = new Zone("id", "location");

		zone.setIsFailed("id", true);
	}

	@SuppressWarnings("unchecked")
	@Test
	public void testSetIsFailedNewlyCreatedZoneSingleValve()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {

		// Tests setIsFailed on a newly created Zone with
		// only a single valve attached.
		// No exceptions expected.

		zone = new Zone("id", "location");
		zone.addValve("id", v1, "type", 1, "loc");

		zone.setIsFailed("id", true);

		// Check if the only valve in the zone
		// is now failing.
		assertTrue(((Valve) (((HashSet<Valve>) extractField(zone, "valves"))
				.toArray()[0])).isFailed());
	}

	@Test
	public void testSetIsFailedNewlyCreatedZoneSensor()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {

		// Tests setIsFailed on a newly created zone with
		// only a sensor attached.
		// No exceptions expected.

		zone = new Zone("id", "location");
		zone.addSensor("id", mockSensor, "loc");

		zone.setIsFailed("id", true);
		assertTrue(((Sensor) extractField(zone, "sensor")).isFailed());

	}

	@Test
	public void testSetIsRecorded() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {

		// Tests setIsRecorded().
		
		valves[1].setIsRecorded(false);
		sensor.setIsRecorded(false);

		// Check whether the correct devices's recorded
		// state is changed.

		zone.setIsRecorded(valves[0].getID(), false);
		assertFalse(valves[0].isRecorded());
		assertFalse(valves[1].isRecorded());
		assertTrue(valves[2].isRecorded());
		assertFalse(sensor.isRecorded());

		zone.setIsRecorded(valves[1].getID(), true);
		assertFalse(valves[0].isRecorded());
		assertTrue(valves[1].isRecorded());
		assertTrue(valves[2].isRecorded());
		assertFalse(sensor.isRecorded());

		zone.setIsRecorded(sensor.getID(), true);
		assertFalse(valves[0].isRecorded());
		assertTrue(valves[1].isRecorded());
		assertTrue(valves[2].isRecorded());
		assertTrue(sensor.isRecorded());
	}

	@Test
	public void testSetIsRecordedNull() {

		// Test setIsRecorded() with
		// device id == "null"

		zone.setIsRecorded(null, true);

		assertTrue(valves[0].isRecorded());
		assertTrue(valves[1].isRecorded());
		assertTrue(valves[2].isRecorded());
		assertTrue(sensor.isRecorded());
	}

	@Test
	public void testSetIsRecordedWrongName() {
		// Test setIsRecorded with a wrong
		// device name

		zone.setIsRecorded("dummy", true);

		assertTrue(valves[0].isRecorded());
		assertTrue(valves[1].isRecorded());
		assertTrue(valves[2].isRecorded());
		assertTrue(sensor.isRecorded());
	}

	@Test
	public void testSetIsRecordedEmptyZone() {

		// Test setIsRecorded on a newly created, empty zone
		zone = new Zone("id", "location");

		zone.setIsRecorded("s1", true);
	}

	@SuppressWarnings("unchecked")
	@Test
	public void testSetIsRecordedSingleValve() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Test setIsRecorded on a zone with only
		// a single valve attached

		zone = new Zone("id", "location");
		zone.addValve("id", v1, "type", 1, "loc");

		zone.setIsRecorded("id", true);

		assertTrue(((Valve) (((HashSet<Valve>) extractField(zone, "valves"))
				.toArray()[0])).isRecorded());
	}

	@Test
	public void testSetIsRecordedSensorOnly() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Test setIsRecorded on a zone with only
		// a sensor attached

		zone = new Zone("id", "location");
		zone.addSensor("id", mockSensor, "loc");

		zone.setIsFailed("id", true);
		assertTrue(((Sensor) extractField(zone, "sensor")).isRecorded());
	}

	@Test
	public void testGetMeasuredMoistureLevel() throws DeviceFailureException {
		// Tests whether getMeasuredMoistureLevel reports
		// the moisture level of the attached SensorDevice
		assertEquals(mockSensor.value, zone.getMeasuredMoistureLevel());

		mockSensor.value = mockSensor.value + 2;

		assertEquals(mockSensor.value, zone.getMeasuredMoistureLevel());
	}

	@Test
	public void testGetMeasuredMoistureLevelFailedSensors() {
		// Tests how getMeasuredMoistureLevel behaves
		// if the SensorDevice fails

		mockSensor = new MockSensorDevice(3);

		zone.addSensor("s02", mockSensor, "loc2");

		try {
			zone.getMeasuredMoistureLevel();
			fail();
		} catch (DeviceFailureException e) {
			assertEquals(1, e.getDevices().size());
			assertTrue(e.getDevices().contains("s02"));
		}
	}

	@Test
	public void testGetMeasuredMoistureLevelNewZone() {
		// Assume a newly created zone and test
		// getMeasuredMoistureLevel on it

		zone = new Zone("id", "location");

		try {
			zone.getMeasuredMoistureLevel();
			fail(); // some sort of exception (no NullPointerException) is
					// expected
					// since no sensor is specified (but needed)
		} catch (NullPointerException e) {
			throw e;
		} catch (Exception e) {

		}

	}

	@Test
	public void testCriticalMoistureLevelProperty() {

		// Test the criticalMoistureLevel property

		zone.setCriticalMoistureLevel(0);
		assertEquals(0, zone.getCriticalMoistureLevel());

		zone.setCriticalMoistureLevel(100);
		assertEquals(100, zone.getCriticalMoistureLevel());
	}

	@Test
	public void testCriticalMoistureLevelPropertyIllegalArguments() {

		// Tests setCriticalMoistureLevel with illegal arguments

		try {
			zone.setCriticalMoistureLevel(-1);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			zone.setCriticalMoistureLevel(Integer.MIN_VALUE);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			zone.setCriticalMoistureLevel(Integer.MAX_VALUE);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			zone.setCriticalMoistureLevel(101);
			fail();
		} catch (IllegalArgumentException e) {

		}

	}

	@Test
	public void testAllocationProperty() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests the allocation property

		zone.setAllocation(0);
		assertEquals(0, extractField(zone, "allocation"));

		zone.setAllocation(Integer.MAX_VALUE);
		assertEquals(Integer.MAX_VALUE, extractField(zone, "allocation"));
	}

	@Test
	public void testAllocationPropertyIllegalArguments() {
		// Tests setAllocation with illegal parameters

		try {
			zone.setAllocation(-1);
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			zone.setAllocation(Integer.MIN_VALUE);
			fail();
		} catch (IllegalArgumentException e) {

		}
	}

	@Test
	public void testAddSensor() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests addSensor().
		
		zone.addSensor("id", mockSensor, "loc");

		// Check if the added sensor corresponds to
		// the the supplied parameters
		Sensor sensor = extractField(zone, "sensor");
		assertEquals("id", sensor.getID());
		assertEquals("loc", sensor.getLocation());
		assertEquals(mockSensor, extractField(sensor, "device"));
		assertFalse(sensor.isFailed());
		assertTrue(sensor.isRecorded());

		mockSensor = new MockSensorDevice(3);

		// Check if the added sensor corresponds to
		// the the supplied parameters
		zone.addSensor("id2", mockSensor, null);
		sensor = extractField(zone, "sensor");
		assertEquals("id2", sensor.getID());
		assertEquals(null, sensor.getLocation());
		assertEquals(mockSensor, extractField(sensor, "device"));
		assertFalse(sensor.isFailed());
		assertTrue(sensor.isRecorded());
	}

	@Test
	public void testAddSensorIllegalParameters() {

		// Test addSensor() with illegal parameters

		try {
			zone.addSensor(null, mockSensor, "loc");
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			zone.addSensor("id", null, "loc");
			fail();
		} catch (IllegalArgumentException e) {

		}
	}

	@SuppressWarnings("unchecked")
	@Test
	public void testAddValve() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests addValve().
		
		MockValveDevice v4 = new MockValveDevice();
		zone.addValve("i4", v4, "t4", 1, "loc4");

		// Check whether there are 4 valves in the zone
		assertEquals(4, valveSet.size());

		HashSet<Valve> setCopy = (HashSet<Valve>) valveSet.clone();

		for (Valve v : valves)
			if (setCopy.contains(v))
				setCopy.remove(v);

		Valve newValve = (Valve) setCopy.toArray()[0];

		// Is the new valve properly configured?
		assertEquals("i4", newValve.getID());
		assertEquals("loc4", newValve.getLocation());
		assertEquals(1, extractField(newValve, "flowRate"));
		assertEquals(v4, extractField(newValve, "device"));
		assertEquals("t4", extractField(newValve, "type"));
		assertTrue(newValve.isRecorded());
		assertFalse(newValve.isFailed());
		assertFalse(newValve.isOpen());
		assertEquals(0, extractField(newValve, "minutesOpen"));
	}

	@Test
	public void testAddValveNewZone() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests whether adding a valve to an empty zone
		// works as expected

		zone = new Zone("id", "location");
		MockValveDevice v4 = new MockValveDevice();

		zone.addValve("i5", v4, null, Integer.MAX_VALUE, null);

		valveSet = extractField(zone, "valves");

		assertEquals(1, valveSet.size());

		Valve newValve = (Valve) valveSet.toArray()[0];

		assertEquals("i5", newValve.getID());
		assertEquals(null, newValve.getLocation());
		assertEquals(Integer.MAX_VALUE, extractField(newValve, "flowRate"));
		assertEquals(v4, extractField(newValve, "device"));
		assertEquals(null, extractField(newValve, "type"));
		assertTrue(newValve.isRecorded());
		assertFalse(newValve.isFailed());
		assertFalse(newValve.isOpen());
		assertEquals(0, extractField(newValve, "minutesOpen"));
	}

	@Test
	public void testAddValveIllegalParameters() {

		// Tests addValve's behavior when confronted
		// with illegal arguments

		MockValveDevice v4 = new MockValveDevice();
		try {
			zone.addValve(null, v4, "t4", 1, "location");
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			zone.addValve("id", null, "t4", 1, "location");
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			zone.addValve("id", v4, "t4", 0, "location");
			fail();
		} catch (IllegalArgumentException e) {

		}

		try {
			zone.addValve("id", v4, "t4", Integer.MIN_VALUE, "location");
			fail();
		} catch (IllegalArgumentException e) {

		}

	}

	@Test
	public void testRepairDevice() {
		// Tests repairDevice().
		
		zone.setIsFailed("s1", true);
		zone.setIsFailed("i1", true);
		zone.setIsFailed("i3", true);

		// Now repair some of the failed devices.
		
		zone.repairDevice("i1");
		assertFalse(valves[0].isFailed());
		assertFalse(valves[1].isFailed());
		assertTrue(valves[2].isFailed());
		assertTrue(sensor.isFailed());

		zone.repairDevice("s1");
		assertFalse(valves[0].isFailed());
		assertFalse(valves[1].isFailed());
		assertTrue(valves[2].isFailed());
		assertFalse(sensor.isFailed());
	}

	@Test
	public void testRepairDeviceIllegalArguments() {
		// Test repairDevice with Illegal Arguments
		
		zone.setIsFailed("s1", true);
		zone.setIsFailed("i1", true);
		zone.setIsFailed("i3", true);

		zone.repairDevice("sensor");
		assertTrue(valves[0].isFailed());
		assertFalse(valves[1].isFailed());
		assertTrue(valves[2].isFailed());
		assertTrue(sensor.isFailed());

		zone.repairDevice(null);
		assertTrue(valves[0].isFailed());
		assertFalse(valves[1].isFailed());
		assertTrue(valves[2].isFailed());
		assertTrue(sensor.isFailed());
	}

	@Test
	public void testReset() throws DeviceFailureException {
		// Tests reset()

		zone.openAllValves();
		zone.tick();
		zone.tick();

		assertEquals(18, zone.getWaterUsed());

		zone.reset();

		assertEquals(0, zone.getWaterUsed());

		// Can we reset twice?

		zone.reset();

		assertEquals(0, zone.getWaterUsed());

		// Can we reset again?

		zone.tick();
		zone.reset();

		assertEquals(0, zone.getWaterUsed());
	}

	@Test
	public void testResetEmptyZone() {
		// Tests reset on an empty zone

		zone = new Zone("id", "location");

		zone.tick();

		assertEquals(0, zone.getWaterUsed());

		zone.reset();

		assertEquals(0, zone.getWaterUsed());

	}

	@Test
	public void testTickClosedValves() throws DeviceFailureException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests tick().
		
		zone.tick();

		// All valves are closed, so no water should be used
		assertEquals(0, zone.getWaterUsed());
	}

	@Test
	public void testTickAllValvesOpen() throws DeviceFailureException {
		// Tests tick().
		
		zone.openAllValves();

		zone.tick();
		zone.tick();

		// 2 * 9 liters = 18 liters
		assertEquals(18, zone.getWaterUsed());
	}

	@Test
	public void testTickOpenedAndClosedValves() throws DeviceFailureException {
		// Tests whether tick respects all valves that
		// are closed again

		zone.openAllValves();

		zone.tick();

		zone.closeAllValves();

		zone.tick();

		assertEquals(9, zone.getWaterUsed());
	}

	@Test
	public void testTickSomeValvesOpen() throws DeviceFailureException {
		// Tests tick with some valves open, some closed

		zone.toggleValve("i1");
		zone.toggleValve("i3");

		zone.tick();
		zone.tick();

		assertEquals(12, zone.getWaterUsed());
	}

	@Test
	public void testTickFailedValves() throws DeviceFailureException {
		// Tests how tick reacts to failed valves

		zone.openAllValves();
		zone.tick();
		zone.closeAllValves();

		v2.isFailed = true;

		try {
			zone.openAllValves();
			fail();
		} catch (DeviceFailureException e) {
		}

		zone.toggleValve("i1");

		zone.tick();
		zone.tick();

		// 9 liters + 2*4 liters = 17 liters
		assertEquals(17, zone.getWaterUsed());

	}
	
	@Test
	public void testTickFailedOpenValves() throws DeviceFailureException {
		// Tests how tick reacts to failed valves that cannot be closed

		zone.openAllValves();
		zone.tick();

		v2.isFailed = true;

		try {
			zone.toggleValve("i2");
			fail();
		} catch (DeviceFailureException e) {
		}
		
		zone.toggleValve("i1");

		zone.tick();
		zone.tick();

		// 9 liters + 2*7 liters = 23 liters since
		// the failed valve should still be open ...
		assertEquals(23, zone.getWaterUsed());

	}

	@Test
	public void testTickAddedValve() throws DeviceFailureException {
		// Tests tick(). Between two ticks, a new valve is added
		// to the zone.
		
		zone.openAllValves();
		zone.tick();
		zone.tick();

		zone.addValve("i4", new MockValveDevice(), "t4", 7, zone.getLocation());

		zone.openAllValves();

		zone.tick();

		assertEquals(34, zone.getWaterUsed());
	}

	@Test
	public void testTickWaterUsedOverflow() throws DeviceFailureException {
		// Tests what happens when a very huge
		// amount of water has been used

		zone.addValve("i4", new MockValveDevice(), "t4",
				Integer.MAX_VALUE - 15, zone.getLocation());

		zone.openAllValves();
		zone.tick(); // water used = Integer.MAX_VALUE - 6

		zone.toggleValve("i4");

		zone.tick(); // water used > Integer.MAX_VALUE

		// The amount of water used cannot be negative
		assertTrue(zone.getWaterUsed() >= 0);

	}

	@Test
	public void testToggleClosedValve() throws DeviceFailureException {
		// Toggle closed valve

		zone.toggleValve("i1");
		assertTrue(valves[0].isOpen());
	}

	@Test
	public void testToggleOpenValve() throws DeviceFailureException {
		// toggle open valve

		valves[2].open();
		zone.toggleValve("i3");
		assertFalse(valves[2].isOpen());
	}

	@Test
	public void testToggleFailedValve() throws DeviceFailureException {
		// try to toggle a failed valve.
		// DeviceFailureException expected.

		valves[1].open();
		
		v2.isFailed = true;

		try {
			zone.toggleValve("i2");
			fail();
		} catch (DeviceFailureException e) {
			assertEquals(1, e.getDevices().size());
			assertTrue(e.getDevices().contains("i2"));
		}
		
		// The valve failed while toggling,
		// so it should still be open like before
		assertTrue(valves[1].isOpen());
	}

	@Test
	public void testToogleValveWrongName() throws DeviceFailureException {

		// try to toggle the valve "dummy" which does not exist
		// IllegalArgumentException expected

		try {
			zone.toggleValve("dummy");
			fail();
		} catch (IllegalArgumentException e) {

		}
	}

	@Test
	public void testToggleValveNullArgument() throws DeviceFailureException {
		// try to toggle the valve null
		// IllegalArgumentException expected

		try {
			zone.toggleValve(null);
			fail();
		} catch (IllegalArgumentException e) {

		}

	}

	@Test
	public void testOpenAllValves() throws DeviceFailureException {
		// Initially, all valves are closed

		zone.openAllValves();

		for (Valve v : valves)
			assertTrue(v.isOpen());
	}

	@Test
	public void testOpenAllValvesSingleClosedValve()
			throws DeviceFailureException {
		// Test openAllValves when a single valve is closed,
		// all others are open

		zone.openAllValves();

		valves[1].close();

		zone.openAllValves();
		for (Valve v : valves)
			assertTrue(v.isOpen());
	}

	@Test
	public void testOpenAllValvesFailedValves() throws DeviceFailureException {
		// Tests how openAllValves reacts to failed
		// valves

		valves[0].close();
		valves[1].close();
		v1.isFailed = true;
		v2.isFailed = true;
		try {
			zone.openAllValves();
			fail();
		} catch (DeviceFailureException e) {
			// Expected
			assertEquals(2, e.getDevices().size());
			assertTrue(e.getDevices().contains("i1"));
			assertTrue(e.getDevices().contains("i2"));
		}
		
		assertTrue(valves[2].isOpen());
		assertFalse(valves[0].isOpen());
		assertFalse(valves[1].isOpen());
	}

	@Test
	public void testOpenAllValvesEmptyZone() throws DeviceFailureException {
		// Tests openAllValves() on a newly created zone.
		zone = new Zone("id", "loc");

		zone.openAllValves();
	}

	@Test
	public void testCloseAllValves() throws DeviceFailureException {
		// Tests closeAllValves().
		
		zone.openAllValves();
		zone.closeAllValves();

		for (Valve v : valves)
			assertFalse(v.isOpen());
	}

	@Test
	public void testCloseAllValvesSingleValveOpen()
			throws DeviceFailureException {
		// Tests closeAllValves(). Only one valve is open.

		valves[1].open();

		zone.closeAllValves();
		for (Valve v : valves)
			assertFalse(v.isOpen());
	}

	@Test
	public void testCloseAllValvesFailedValves() throws DeviceFailureException {
		// Tests how closeAllValves reacts
		// to failed ValveDevices

		valves[0].open();
		valves[1].open();
		v1.isFailed = true;
		v2.isFailed = true;
		try {
			zone.closeAllValves();
			fail();
		} catch (DeviceFailureException e) {
			// Expected
			assertEquals(2, e.getDevices().size());
			assertTrue(e.getDevices().contains("i1"));
			assertTrue(e.getDevices().contains("i2"));
		}
		
		// These valves failed while closing, so they
		// should still be open ...
		assertTrue(valves[0].isOpen());
		assertTrue(valves[1].isOpen());
	}

	@Test
	public void testCloseAllValvesEmptyZone() throws DeviceFailureException {
		// Tests closeAllValves() on a newly created zone.
		zone = new Zone("id", "loc");

		zone.closeAllValves();

		for (Valve v : valves)
			assertFalse(v.isOpen());
	}

	@Test
	public void testIsIrrigatedWaterUsed() throws DeviceFailureException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {

		// Tests isIrrigated when water used >= water allocation

		zone.setAllocation(15);
		zone.setCriticalMoistureLevel(20);
		mockSensor.value = 19;

		zone.openAllValves();

		// water used < water allocation
		assertFalse(zone.isIrrigated());

		zone.tick();

		// water used < water allocation
		assertFalse(zone.isIrrigated());

		zone.tick();

		// water used >= water allocation
		assertTrue(zone.isIrrigated());
	}

	@Test
	public void testIsIrrigatedMoistureLevel() throws DeviceFailureException {
		// Test isIrrigated when moisture level >= critical moisture level

		zone.setAllocation(15);
		zone.setCriticalMoistureLevel(20);
		mockSensor.value = 19;

		// moisture level < critical moisture level
		assertFalse(zone.isIrrigated());

		mockSensor.value = 20;

		// moisture level >= critical moisture level
		assertTrue(zone.isIrrigated());
	}

	@Test
	public void testIsIrrigatedBothConditions() throws DeviceFailureException {
		// Tests isIrrigated when both conditions are satisfied
		zone.setAllocation(15);
		zone.setCriticalMoistureLevel(20);
		mockSensor.value = 19;

		// moisture level < critical moisture level
		// && water used < water allocation
		assertFalse(zone.isIrrigated());

		zone.openAllValves();
		zone.tick();
		zone.tick();
		mockSensor.value = 20;

		// moisture level >= critical moisture level
		// && water used >= water allocation
		assertTrue(zone.isIrrigated());
	}

	@Test
	public void testIsIrrigatedSensorFailure() throws DeviceFailureException,
			IllegalArgumentException, IllegalAccessException,
			SecurityException, NoSuchFieldException {
		// Tests isIrrigated when the underlying sensor fails
		zone.setAllocation(15);
		zone.setCriticalMoistureLevel(20);
		mockSensor.value = 19;

		assertFalse(zone.isIrrigated());

		// Substitute the sensor with one that fails
		Field sensorField = extractFieldAsField(sensor, "device");
		sensorField.setAccessible(true);
		sensorField.set(sensor, new MockSensorDevice(3));

		try {
			zone.isIrrigated();
			fail();
		} catch (DeviceFailureException e) {
			// Expected
			assertEquals(1, e.getDevices().size());
			assertTrue(e.getDevices().contains("s1"));
		}
	}

	@Test
	public void testIsIrrigatedValveFailure() throws DeviceFailureException {
		// Tests isIrrigated. Some of the valves fail one after another
		// No influence on zone.isIrrigated expected

		zone.setAllocation(100);
		zone.setCriticalMoistureLevel(20);
		mockSensor.value = 19;

		zone.openAllValves();

		for (Iterator<MockValveDevice> i = mocks.iterator(); i.hasNext();) {
			i.next().isFailed = true;

			assertFalse(zone.isIrrigated());
		}

		// Propagate the failures to zone's Valve's
		// if not already done
		try {
			zone.closeAllValves();
		} catch (DeviceFailureException e) {
		}

		assertFalse(zone.isIrrigated());

	}
}
