package simulation;

import static tools.HelperFunctions.*;
import static org.junit.Assert.*;

import java.awt.event.MouseEvent;

import javax.swing.JButton;

import org.junit.Test;

import util.Command;

public class ButtonRepeaterTest {

	// Created for testing purposes.
	private class MockCommand implements Command {

		private int invocations = 0;
		private Object lastArgument = null;

		public int getInvocations() {
			return invocations;
		}

		public Object getLastArgument() {
			return lastArgument;
		}

		@Override
		public void execute(Object arg) {
			invocations++;
			lastArgument = arg;
		}

	}

	private void testButtonRepeaterInit(int expectedInterval,
			int actualInterval, Command expectedCommand, Command actualCommand)
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests a ButtonRepeater's initial state.

		ButtonRepeater repeater = new ButtonRepeater(actualInterval,
				actualCommand);

		Pulse pulse = (Pulse) extractField(repeater, "repeater");
		assertEquals(expectedInterval, extractField(pulse, "interval"));
		assertEquals(expectedCommand, extractField(repeater, "cmd"));

		// Check whether repeater is observing repeater's pulse
		assertObserved(pulse, repeater);
	}

	@Test
	public void testButtonRepeater() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		MockCommand cmd = null;
		// Tests ButtonRepeater's constructor.

		cmd = new MockCommand();
		testButtonRepeaterInit(1, 1, cmd, cmd);

		cmd = null;
		testButtonRepeaterInit(Integer.MAX_VALUE, Integer.MAX_VALUE, cmd, cmd);
	}

	@Test
	public void testButtonRepeaterIllegalArguments()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests ButtonRepeater's constructor. Some arguments are illegal.

		// The interval should automatically be adjusted to 1000
		// if the parameter is to small.
		MockCommand cmd = new MockCommand();
		testButtonRepeaterInit(1000, Integer.MIN_VALUE, cmd, cmd);

		cmd = null;
		testButtonRepeaterInit(1000, 0, null, null);

	}

	@Test
	public void testMouseEventsRunningPulse() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			InterruptedException {
		// Tests the repeater with several mouse events. The pulse thread
		// is running.
		MockCommand cmd = new MockCommand();

		ButtonRepeater repeater = new ButtonRepeater(5, cmd);

		Pulse pulse = extractField(repeater, "repeater");

		MouseEvent event = new MouseEvent(new JButton(), 0, 0, 0, 0, 0, 0, 0,
				0, false, 0);

		// mouseReleased and mouseExited both stop the pulse thread
		pulse.start();
		repeater.mouseReleased(event);
		assertFalse((Boolean) extractField(pulse, "keepRunning"));
		assertNull(extractField(pulse, "pulseThread"));

		pulse.start();
		repeater.mouseExited(event);
		assertFalse((Boolean) extractField(pulse, "keepRunning"));
		assertNull(extractField(pulse, "pulseThread"));

		// mousePressed has no effect on the pulse thread
		pulse.start();
		repeater.mousePressed(event);
		assertTrue((Boolean) extractField(pulse, "keepRunning"));
		assertTrue(((Thread) extractField(pulse, "pulseThread")).isAlive());

		// mouseReleased and mouseExited stop the pulse thread
		pulse.start();
		repeater.mouseReleased(null);
		assertFalse((Boolean) extractField(pulse, "keepRunning"));
		assertNull(extractField(pulse, "pulseThread"));

		pulse.start();
		repeater.mouseExited(null);
		assertFalse((Boolean) extractField(pulse, "keepRunning"));
		assertNull(extractField(pulse, "pulseThread"));

		// mouseReleased has no effect
		pulse.start();
		repeater.mousePressed(null);
		assertTrue((Boolean) extractField(pulse, "keepRunning"));
		assertTrue(((Thread) extractField(pulse, "pulseThread")).isAlive());
	}

	@Test
	public void testMouseEventsStoppedPulse() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests the repeater with several mouse events. The pulse thread
		// is stopped.
		MockCommand cmd = new MockCommand();

		ButtonRepeater repeater = new ButtonRepeater(5, cmd);

		Pulse pulse = extractField(repeater, "repeater");

		MouseEvent event = new MouseEvent(new JButton(), 0, 0, 0, 0, 0, 0, 0,
				0, false, 0);

		// mousePressed starts the pulse thread
		repeater.mousePressed(event);
		assertTrue((Boolean) extractField(pulse, "keepRunning"));
		assertTrue(((Thread) extractField(pulse, "pulseThread")).isAlive());

		pulse.stop();
		repeater.mousePressed(null);
		assertTrue((Boolean) extractField(pulse, "keepRunning"));
		assertTrue(((Thread) extractField(pulse, "pulseThread")).isAlive());

		// mouseReleased and mouseExited have no effects
		pulse.stop();
		repeater.mouseReleased(event);
		assertFalse((Boolean) extractField(pulse, "keepRunning"));
		assertNull(extractField(pulse, "pulseThread"));

		repeater.mouseExited(event);
		assertFalse((Boolean) extractField(pulse, "keepRunning"));
		assertNull(extractField(pulse, "pulseThread"));

		repeater.mouseReleased(null);
		assertFalse((Boolean) extractField(pulse, "keepRunning"));
		assertNull(extractField(pulse, "pulseThread"));

		repeater.mouseExited(null);
		assertFalse((Boolean) extractField(pulse, "keepRunning"));
		assertNull(extractField(pulse, "pulseThread"));

	}

	@Test
	public void testUpdate() throws InterruptedException {
		MockCommand theCmd = new MockCommand();

		ButtonRepeater repeater = new ButtonRepeater(5, theCmd);

		repeater.update(null, null);

		assertEquals(1, theCmd.getInvocations());
		assertEquals(repeater, theCmd.getLastArgument());
	}

}
