package simulation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static tools.HelperFunctions.extractField;

import org.junit.Before;
import org.junit.Test;

import tools.HelperFunctions;
import tools.TestObserver;

public class PulseTest {

	private TestObserver observer;

	@Before
	public void init() {
		observer = new TestObserver();
	}

	private void testConstructorHelperMethod(int argument, int expected)
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests a Pulse's initial state.

		Pulse pulse = new Pulse(argument);
		assertEquals(expected, extractField(pulse, "interval"));
		// Initially, the pulse should not be running nor be
		// runnable by an external thread...
		assertFalse(HelperFunctions
				.<Boolean> extractField(pulse, "keepRunning"));
		assertNull(extractField(pulse, "pulseThread"));
	}

	@Test
	public void testConstructorInt() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {

		// Tests Pulse's constructor.

		testConstructorHelperMethod(1, 1);

		testConstructorHelperMethod(Integer.MAX_VALUE, Integer.MAX_VALUE);
	}

	@Test
	public void testConstructorIllegalCases() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests Pulse's constructor. The arguments are to low.

		testConstructorHelperMethod(0, 1000);

		testConstructorHelperMethod(Integer.MIN_VALUE, 1000);
	}

	@Test
	public void testDefaultConstructor() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests the default constructor

		Pulse pulse = new Pulse();
		assertEquals(1000, extractField(pulse, "interval"));
		// Initially, the pulse should not be running nor be
		// runnable by an external thread...
		assertFalse(HelperFunctions
				.<Boolean> extractField(pulse, "keepRunning"));
		assertNull(extractField(pulse, "pulseThread"));
	}

	@Test
	public void testStartThreadNotRunning() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			InterruptedException {
		// Tests start(). The pulse thread is not running.

		Pulse pulse = new Pulse(5);
		pulse.addObserver(observer);

		pulse.start();

		Thread thread = extractField(pulse, "pulseThread");
		assertTrue(thread.isAlive());
		assertTrue((Boolean) extractField(pulse, "keepRunning"));
	}

	@Test
	public void testStartThreadRunning() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests start(). The pulse thread is already running.

		Pulse pulse = new Pulse(5);
		pulse.addObserver(observer);

		pulse.start();

		Thread thread = extractField(pulse, "pulseThread");

		pulse.start();

		Thread thread2 = extractField(pulse, "pulseThread");
		assertEquals(thread, thread2);
		assertTrue(thread2.isAlive());
		assertTrue((Boolean) extractField(pulse, "keepRunning"));
	}

	@Test
	public void testStopRunningThread() throws InterruptedException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {

		// Tests stop(). The pulse thread is running

		Pulse pulse = new Pulse(8);
		pulse.addObserver(observer);

		pulse.start();

		Thread.sleep(20);

		pulse.stop();

		Thread.sleep(20);

		assertNull(extractField(pulse, "pulseThread"));
	}

	@Test
	public void testStopStoppedThread() throws InterruptedException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests stop(). The pulse thread has already stopped

		Pulse pulse = new Pulse(8);
		pulse.addObserver(observer);

		pulse.start();

		Thread.sleep(20);

		pulse.stop();

		Thread.sleep(20);

		pulse.stop();

		assertNull(extractField(pulse, "pulseThread"));

	}

	@Test
	public void testSetInterval() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests setInterval().

		Pulse pulse = new Pulse();

		pulse.setInterval(1);
		assertEquals(1, extractField(pulse, "interval"));

		pulse.setInterval(Integer.MAX_VALUE);
		assertEquals(Integer.MAX_VALUE, extractField(pulse, "interval"));
	}

	@Test
	public void testSetIntervalIllegalArguments()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests setInterval(). The argument is not valid.

		Pulse pulse = new Pulse();
		
		pulse.setInterval(Integer.MAX_VALUE);
		
		// Since the argument is to low, it should be ignored.
		pulse.setInterval(0);
		assertEquals(Integer.MAX_VALUE, extractField(pulse, "interval"));

		pulse.setInterval(Integer.MIN_VALUE);
		assertEquals(Integer.MAX_VALUE, extractField(pulse, "interval"));

	}

	@Test
	public void testRunConstantIntervals() throws InterruptedException {

		// Tests run().
		
		TestObserver observer = new TestObserver();

		Pulse pulse = new Pulse(25);
		pulse.addObserver(observer);

		pulse.start();

		Thread.sleep(500);

		// number of notifications should be
		// in the following range
		assertTrue(18 <= observer.notifications());
		assertTrue(23 >= observer.notifications());
	}

	@Test
	public void testRunStoppedDuringInterval() throws InterruptedException {

		// Tests run(). The pulse is stopped during its execution.

		Pulse pulse = new Pulse(25);
		pulse.addObserver(observer);
		pulse.start();

		Thread.sleep(320);

		pulse.stop();
		int nots = observer.notifications();

		Thread.sleep(40);

		assertEquals(nots, observer.notifications());
	}

	@Test
	public void testRunChangeIntervalWhileRunning() throws InterruptedException {

		// Tests run(). Time intervals change.

		Pulse pulse = new Pulse(25);
		pulse.addObserver(observer);
		pulse.start();

		Thread.sleep(160);

		assertTrue(6 <= observer.notifications());
		pulse.setInterval(12);

		Thread.sleep(160);

		assertTrue(19 <= observer.notifications());

	}

}
