package simulation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static tools.HelperFunctions.extractField;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.EventListener;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;

import org.junit.Test;

public class SimDisplayTest {

	@Test
	public void testSimDisplay() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {

		// Tests SimDisplay's initial state.
		
		SimDisplay display = new SimDisplay();

		// Arrays with Buttons and Labels
		JButton[] buttons = extractField(display, "scrnBtn");
		JLabel[] labels = extractField(display, "lineLbl");

		// A SimDisplay contains 5 JPanels
		Component[] comps = display.getComponents();
		int panels = 0;
		int components = 0;
		int buttonCount = 0;
		int labelCount = 0;

		// For every component in display
		for (Component c : comps) {
			// test whether it is a JPanel
			if (c instanceof JPanel) {
				panels++;
				Component[] compsInPanel = ((JPanel) c).getComponents();
				int innerPanels = 0;
				int innerComponents = 0;
				// If so, test whether there are 3 inner panels for
				// display's panels
				for (Component inner : compsInPanel) {
					innerComponents++;
					// If an inner component is a JPanel again
					if (inner instanceof JPanel) {
						innerPanels++;
						Component[] iinnerComps = ((JPanel) inner)
								.getComponents();
						// Check whether it is a west or an east panel, or ...
						if (iinnerComps.length == 1
								&& (iinnerComps[0] instanceof JButton)) {
							JButton button = (JButton) iinnerComps[0];
							if (panels == 1) {
								// Buttons in the first panel are invisible
								assertFalse(button.isVisible());
							} else {
								// the others have a specified action command
								assertEquals("ScrnBtn" + (buttonCount++),
										button.getActionCommand());
								// Check whether all visible buttons are in the
								// buttons array
								assertEquals(button, buttons[buttonCount - 1]);
							}
						}
						// ... if it is a line panel
						else if ((iinnerComps.length >= 3 && panels > 1)
								|| (panels == 1 && iinnerComps.length <= 4)) {
							for (Component c2 : iinnerComps) {
								// Check whether there are only panels in the
								// line panel
								assertTrue(c2 instanceof JLabel);
								// Check whether all labels are in the labels
								// array
								assertEquals(c2, labels[labelCount++]);
							}

						} else
							fail();
					}
				}
				assertEquals(3, innerPanels);
				assertEquals(3, innerComponents);
			}
			components++;
		}

		assertEquals(5, panels);
		assertEquals(5, components);

		assertEquals(8, buttons.length);
		assertEquals(16, labels.length);

		// buttons have no listeners registered

		for (JButton b : buttons)
			assertEquals(0, b.getActionListeners().length);

	}

	@Test
	public void testWriteLine() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {

		// Tests writeLine().
		
		SimDisplay display = new SimDisplay();
		JLabel[] labels = extractField(display, "lineLbl");

		display.writeLine(0, "AAA");
		assertEquals("AAA", labels[0].getText());
		
		display.writeLine(15, "BBB");
		assertEquals("BBB", labels[15].getText());
	}
	
	@Test
	public void testWriteLineIllegalArguments() {
		
		// Tests writeLine(). Some arguments are invalid.
		
		SimDisplay display = new SimDisplay();
		
		try {
			display.writeLine(-1, "AAA");
			fail();
		} catch (IllegalArgumentException e){
		}
		
		try {
			display.writeLine(Integer.MIN_VALUE, "AAA");
			fail();
		} catch (IllegalArgumentException e){
		}
		
		try {
			display.writeLine(Integer.MAX_VALUE, "AAA");
			fail();
		} catch (IllegalArgumentException e){
		}

		try {
			display.writeLine(16, "AAA");
			fail();
		} catch (IllegalArgumentException e){
		}
		
		// String is null
		
		try {
			display.writeLine(1, null);
			fail();
		} catch (IllegalArgumentException e){
		}
	}

	@Test
	public void testSetListener() throws IllegalArgumentException, SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests setListener().
		
		ActionListener l = new ActionListener() {
			
			@Override
			public void actionPerformed(ActionEvent e) {				
			}
		};
		
		SimDisplay display = new SimDisplay();
		JButton[] buttons = extractField(display, "scrnBtn");
		
		// All buttons should have the listener registered.
		
		display.setListener(l);
		for(JButton b: buttons){
			EventListener[] listeners = b.getActionListeners();
			int listenersFound = 0;
			for (EventListener list: listeners){
				if (list.equals(l))
					listenersFound++;
			}
			assertEquals(1, listenersFound);
		}
		
		// Null is a valid argument.
		
		display.setListener(null);
		for(JButton b: buttons){
			EventListener[] listeners = b.getActionListeners();
			assertEquals(1, listeners.length);
			int listenersFound = 0;
			for (EventListener list: listeners){
				if (list.equals(l))
					listenersFound++;
			}
			assertEquals(1, listenersFound);
		}
		
	}

}
