package simulation;

import static org.junit.Assert.*;

import irrigation.Irrigator;

import java.util.Iterator;

import javax.swing.Box;

import org.junit.Before;
import org.junit.Test;

import device.Clock;
import device.StorageDevice;
import device.StoreFailureException;
import device.sim.MockSimClockDevice;
import static tools.HelperFunctions.*;

public class SimEnvironmentTest {

	private SimEnvironment simEnvironment;
	private SimSite simSite;
	private SimSiteView simSiteView;
	private SimFailureControl failControl;
	private SimEvapControl evapControl;
	private SimTimeControl timeControl;
	private Box box;

	@Before
	public void init() throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Extract some interesting objects.
		
		simEnvironment = new SimEnvironment(null);
		simSite = extractField(simEnvironment, "simSite");
		simSiteView = extractField(simEnvironment, "simSiteView");
		failControl = extractField(simEnvironment, "failControl");
		evapControl = extractField(simEnvironment, "evapControl");
		timeControl = extractField(simEnvironment, "timeControl");

		box = (Box) simEnvironment.getComponent(1);

	}

	@Test
	public void testSimEnvironment() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests SimEnvironment's constructor.
		
		assertNotNull(simSite);
		assertNotNull(simSiteView);
		assertNotNull(failControl);
		assertNotNull(evapControl);
		assertNotNull(timeControl);

		assertEquals(failControl, box.getComponent(0));
		assertEquals(evapControl, box.getComponent(1));
		assertEquals(timeControl, box.getComponent(2));

		assertEquals(simSiteView, simEnvironment.getComponent(0));

	}

	@SuppressWarnings("unchecked")
	@Test
	public void testGetSensor() {
		// Tests getSensor(). The requested sensors already exist.
		Iterator<SimZone> i = simSite.zoneIterator();
		i.next();
		SimSensor s1 = i.next().getSensor();
		SimSensor s2 = i.next().getSensor();

		assertEquals(s1, simEnvironment.getSensor(s1.getName()));
		assertEquals(s2, simEnvironment.getSensor(s2.getName()));
	}
	
	@Test
	public void testGetSensorNewEnvironment() {
		// Tests getSensor(). The requested sensor does not exist.

		try {
			simEnvironment.getSensor("abcdefg");
			fail();
		} catch (IllegalArgumentException e) {
		}
		
		try {
			simEnvironment.getSensor(null);
			fail();
		} catch (IllegalArgumentException e) {

		}
	}

	@SuppressWarnings("unchecked")
	@Test
	public void testGetValve() {
		// Tests getValve(). The requested valves already exist.
		
		Iterator<SimZone> zi = simSite.zoneIterator();
		Iterator<SimValve> vi = zi.next().valveIterator();
		vi.next();
		vi.next();
		SimValve valve1 = vi.next();
		SimValve valve2 = vi.next();
		zi.next();
		vi = zi.next().valveIterator();
		SimValve valve3 = vi.next();

		assertEquals(valve1, simEnvironment.getValve(valve1.getName()));
		assertEquals(valve2, simEnvironment.getValve(valve2.getName()));
		assertEquals(valve3, simEnvironment.getValve(valve3.getName()));
	}

	@Test
	public void testGetValveNotExistingValve() {
		// Tests getValve(). The requested valve does not exist.
		
		try {
			simEnvironment.getValve("ABCDEF");
			fail();
		} catch (IllegalArgumentException e) {
		}

		// Check null
		try {
			simEnvironment.getValve(null);
			fail();
		} catch (IllegalArgumentException e) {

		}
	}

	@Test
	public void testSetIrrigator() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests setIrrigator.
		
		Clock.setClockDevice(new MockSimClockDevice());
		
		// Create a mock irrigator.
		Irrigator i = new Irrigator(new StorageDevice() {

			@Override
			public void setData(String name, String value)
					throws StoreFailureException {

			}

			@Override
			public void removeData(String name) throws StoreFailureException {

			}

			@Override
			public String getData(String name, String defaultValue)
					throws StoreFailureException {
				return null;
			}
		});

		simEnvironment.setIrrigator(i);

		assertEquals(i, extractField(timeControl, "irrigator"));
		
		simEnvironment.setIrrigator(null);
		
		assertNull(extractField(timeControl, "irrigator"));

	}

}
