package simulation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static tools.HelperFunctions.extractField;

import java.awt.event.MouseListener;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;

import org.junit.Test;

import util.Command;

public class SimEvapControlTest {

	@Test
	public void testSimEvapControl() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests SimEvapControl's initial state.

		SimSite site = new SimSite();
		SimEvapControl control = new SimEvapControl(site);
		assertEquals(site, extractField(control, "site"));

		// There should be two panels ...
		assertEquals(2, control.getComponents().length);
		JPanel labelPanelLevel1 = (JPanel) control.getComponents()[0];
		JPanel rightPanelLevel1 = (JPanel) control.getComponents()[1];

		// ... each of them containing some additional components.
		// The left panel contains a label.
		assertEquals(1, labelPanelLevel1.getComponents().length);
		assertTrue(labelPanelLevel1.getComponents()[0] instanceof JLabel);

		// The right panel contains 3 buttons.
		assertEquals(3, rightPanelLevel1.getComponents().length);

		JButton slowerButton = (JButton) rightPanelLevel1.getComponents()[1];
		JButton defaultButton = (JButton) rightPanelLevel1.getComponents()[2];
		JButton fasterButton = (JButton) rightPanelLevel1.getComponents()[0];

		// The buttons should have a mouse listener
		// registered.
		assertTrue(slowerButton.getMouseListeners().length > 0);
		assertTrue(fasterButton.getMouseListeners().length > 0);
		assertTrue(defaultButton.getActionListeners().length > 0);

		// Since the evaporation rate is 10, all buttons
		// should be enabled.
		assertTrue(fasterButton.isEnabled());
		assertTrue(slowerButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
	}

	@Test
	public void testSimEvapControlEvapRateZero()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {

		// Testing whether the ui works correctly if the site's evaporation
		// rate is 0.

		SimSite site = new SimSite();
		site.setEvaporationRate(0);
		SimEvapControl control = new SimEvapControl(site);

		assertEquals(site, extractField(control, "site"));
		assertEquals(2, control.getComponents().length);
		JPanel labelPanelLevel1 = (JPanel) control.getComponents()[0];
		JPanel rightPanelLevel1 = (JPanel) control.getComponents()[1];

		assertEquals(1, labelPanelLevel1.getComponents().length);
		assertTrue(labelPanelLevel1.getComponents()[0] instanceof JLabel);

		assertEquals(3, rightPanelLevel1.getComponents().length);

		JButton slowerButton = (JButton) rightPanelLevel1.getComponents()[1];
		JButton defaultButton = (JButton) rightPanelLevel1.getComponents()[2];
		JButton fasterButton = (JButton) rightPanelLevel1.getComponents()[0];

		// The buttons should have a mouse listener
		// registered.
		assertTrue(slowerButton.getMouseListeners().length > 0);
		assertTrue(fasterButton.getMouseListeners().length > 0);
		assertTrue(defaultButton.getActionListeners().length > 0);

		// The slower button should be disabled since
		// the evaporation rate is already on its minimum.
		assertTrue(fasterButton.isEnabled());
		assertFalse(slowerButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
	}

	@Test
	public void testSimEvapControlEvapRateMaximum()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Testing whether the ui works correctly if the site's evaporation
		// rate is 100

		SimSite site = new SimSite();
		site.setEvaporationRate(100);
		SimEvapControl control = new SimEvapControl(site);

		assertEquals(site, extractField(control, "site"));
		assertEquals(2, control.getComponents().length);
		JPanel labelPanelLevel1 = (JPanel) control.getComponents()[0];
		JPanel rightPanelLevel1 = (JPanel) control.getComponents()[1];

		assertEquals(1, labelPanelLevel1.getComponents().length);
		assertTrue(labelPanelLevel1.getComponents()[0] instanceof JLabel);

		assertEquals(3, rightPanelLevel1.getComponents().length);

		JButton slowerButton = (JButton) rightPanelLevel1.getComponents()[1];
		JButton defaultButton = (JButton) rightPanelLevel1.getComponents()[2];
		JButton fasterButton = (JButton) rightPanelLevel1.getComponents()[0];

		// The buttons should have a mouse listener
		// registered.
		assertTrue(slowerButton.getMouseListeners().length > 0);
		assertTrue(fasterButton.getMouseListeners().length > 0);
		assertTrue(defaultButton.getActionListeners().length > 0);

		// The faster button should be disabled since the
		// evaporation rate is on its maximum.
		assertFalse(fasterButton.isEnabled());
		assertTrue(slowerButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
	}

	@Test
	public void testFasterButton() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			InterruptedException {

		// Tests the faster button. The current evaporation rate is 25.

		SimSite site = new SimSite();
		site.setEvaporationRate(25);
		SimEvapControl control = new SimEvapControl(site);

		JButton fasterButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[0];
		JButton slowerButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[1];
		JButton defaultButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[2];

		ButtonRepeater repeater = null;

		// Extract faster button's repeater.
		for (MouseListener l : fasterButton.getMouseListeners())
			if (l instanceof ButtonRepeater)
				repeater = (ButtonRepeater) l;

		// Test if repeater's interval is 50 and execute the command
		// associated with the repeater.
		assertEquals(50,
				extractField(extractField(repeater, "repeater"), "interval"));

		Command cmd = extractField(repeater, "cmd");

		cmd.execute(null);
		// Since the evaporation rate is 24, all buttons
		// should be enabled.
		assertEquals(26, site.getEvaporationRate());
		assertTrue(fasterButton.isEnabled());
		assertTrue(slowerButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
	}

	@Test
	public void testFasterButtonEvapMaximum() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests the faster button. The current evaporation rate is 99.

		SimSite site = new SimSite();
		site.setEvaporationRate(99);
		SimEvapControl control = new SimEvapControl(site);

		JButton fasterButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[0];
		JButton slowerButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[1];
		JButton defaultButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[2];
		ButtonRepeater repeater = null;
		
		// Extract faster button's repeater.
		for (MouseListener l : fasterButton.getMouseListeners())
			if (l instanceof ButtonRepeater)
				repeater = (ButtonRepeater) l;
		
		// Test if repeater's interval is 50 and execute the command
		// associated with the repeater.
		assertEquals(50,
				extractField(extractField(repeater, "repeater"), "interval"));

		Command cmd = extractField(repeater, "cmd");
		cmd.execute(null);
		
		// Since the evaporation rate is 100, the
		// faster button should be disabled.
		assertEquals(100, site.getEvaporationRate());
		assertFalse(fasterButton.isEnabled());
		assertTrue(slowerButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
	}

	@Test
	public void testSlowerButton() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests the slower button. SimSite has an evaporation rate of 25.

		SimSite site = new SimSite();
		site.setEvaporationRate(25);
		SimEvapControl control = new SimEvapControl(site);

		JButton fasterButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[0];
		JButton slowerButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[1];
		JButton defaultButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[2];

		// Extract slower button's repeater.
		ButtonRepeater repeater = null;

		for (MouseListener l : slowerButton.getMouseListeners())
			if (l instanceof ButtonRepeater)
				repeater = (ButtonRepeater) l;

		// Test if repeater's interval is 50 and execute the command
		// associated with the repeater.
		assertEquals(50,
				extractField(extractField(repeater, "repeater"), "interval"));

		Command cmd = extractField(repeater, "cmd");

		cmd.execute(null);
		// Since the evaporation rate is 24, all buttons
		// should be enabled.
		assertEquals(24, site.getEvaporationRate());
		assertTrue(fasterButton.isEnabled());
		assertTrue(slowerButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
	}

	@Test
	public void testSlowerButtonEvapMinimum() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {

		// Tests the slower button. SimSite has an evaporation rate of 25.

		SimSite site = new SimSite();
		site.setEvaporationRate(1);
		SimEvapControl control = new SimEvapControl(site);

		JButton fasterButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[0];
		JButton slowerButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[1];
		JButton defaultButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[2];
		ButtonRepeater repeater = null;

		// Extract slower button's repeater.
		for (MouseListener l : slowerButton.getMouseListeners())
			if (l instanceof ButtonRepeater)
				repeater = (ButtonRepeater) l;
		
		// Test if repeater's interval is 50 and execute the command
		// associated with the repeater.
		assertEquals(50,
				extractField(extractField(repeater, "repeater"), "interval"));

		Command cmd = extractField(repeater, "cmd");

		cmd.execute(null);
		// Since the evaporation rate is 0, the
		// slower button should be disabled.
		assertEquals(0, site.getEvaporationRate());
		assertTrue(fasterButton.isEnabled());
		assertFalse(slowerButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
	}

	@Test
	public void testDefaultButtonEvapLevel25() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests the default button. SimSite has an evaporation rate of 25.

		SimSite site = new SimSite();
		site.setEvaporationRate(25);
		SimEvapControl control = new SimEvapControl(site);

		JButton fasterButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[0];
		JButton slowerButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[1];
		JButton defaultButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[2];

		defaultButton.doClick();
		// The evaporation rate should be set to 10 and
		// all buttons should be enabled.
		assertEquals(10, site.getEvaporationRate());
		assertTrue(fasterButton.isEnabled());
		assertTrue(slowerButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
	}

	@Test
	public void testDefaultButtonEvapLevel0() {
		// Tests the default button. SimSite has an evaporation rate of 0

		SimSite site = new SimSite();
		site.setEvaporationRate(0);
		SimEvapControl control = new SimEvapControl(site);

		JButton fasterButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[0];
		JButton slowerButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[1];
		JButton defaultButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[2];

		defaultButton.doClick();
		// The evaporation rate should be set to 10 and
		// all buttons should be enabled.
		assertEquals(10, site.getEvaporationRate());
		assertTrue(fasterButton.isEnabled());
		assertTrue(slowerButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
	}

	@Test
	public void testDefaultButtonEvapLevel100() {
		// Tests the default button. SimSite has an evaporation rate of 100

		SimSite site = new SimSite();
		site.setEvaporationRate(100);
		SimEvapControl control = new SimEvapControl(site);

		JButton fasterButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[0];
		JButton slowerButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[1];
		JButton defaultButton = (JButton) ((JPanel) control.getComponents()[1])
				.getComponents()[2];

		defaultButton.doClick();
		// The evaporation rate should be set to 10 and
		// all buttons should be enabled.
		assertEquals(10, site.getEvaporationRate());
		assertTrue(fasterButton.isEnabled());
		assertTrue(slowerButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
	}
}
