package simulation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static tools.HelperFunctions.extractField;

import java.awt.Component;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.HashSet;
import java.util.Iterator;

import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JPanel;

import org.junit.Test;

public class SimFailureControlTest {

	@SuppressWarnings("unchecked")
	@Test
	public void testSimFailureControl() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		
		// Tests the initial state of SimFailureControl.

		SimSite site = new SimSite();
		SimFailureControl control = new SimFailureControl(site);

		assertEquals(site, extractField(control, "site"));
		Component[] comps = control.getComponents();

		// Add the names of all devices to a collection.
		int numberOfZones = 0;
		int numberOfValves = 0;
		HashSet<String> names = new HashSet<String>();
		for (Iterator<SimZone> i1 = site.zoneIterator(); i1.hasNext();) {
			numberOfZones++;
			numberOfValves++;
			SimZone currentZone = i1.next();
			names.add(currentZone.getSensor().getName());
			for (Iterator<SimValve> i2 = currentZone.valveIterator(); i2
					.hasNext();) {
				numberOfValves++;
				names.add(i2.next().getName());
			}
		}

		int foundLabels = 0;
		int foundPanels = 0;
		int innerLabels = 0;
		int innerCheckbox = 0;

		// Check whether the components in SimFailureControl are
		// properly combined.
		for (Component c : comps) {
			if (c instanceof JLabel)
				foundLabels++;
			else if (c instanceof JPanel) {
				foundPanels++;
				Component[] innerComps = ((JPanel) c).getComponents();
				for (Component ci : innerComps) {
					if (ci instanceof JCheckBox) {
						JCheckBox currentCB = (JCheckBox) ci;
						innerCheckbox++;
						ItemListener[] l = currentCB.getItemListeners();
						assertEquals(1, l.length);
						assertEquals(control, l[0]);

						// Check if each check box displays the name of a
						// valve or a sensor
						assertTrue(names.contains(currentCB.getText()));
						names.remove(currentCB.getText());
					} else if (ci instanceof JLabel)
						innerLabels++;
				}
			}
		}

		assertEquals(numberOfZones, innerLabels);
		assertEquals(numberOfValves, innerCheckbox);
		assertTrue(names.isEmpty());
	}

	@Test
	public void testSimFailureControlNullArgument() {
		// Since no preconditions for SimFailureControl's constructor
		// exist, the following line of code should work
		new SimFailureControl(null);

	}

	// Mock object created for testing purposes.
	class MockSite extends SimSite {

		public String getLastFailedDevice() {
			return lastFailedDevice;
		}

		public String getLastRepairedDevice() {
			return lastRepairedDevice;
		}

		private String lastFailedDevice = null;
		private String lastRepairedDevice = null;

		@Override
		void setFailed(String deviceName) {
			lastFailedDevice = deviceName;
		}

		@Override
		void setRepaired(String deviceName) {
			lastRepairedDevice = deviceName;
		}

	}

	@Test
	public void testItemStateChanged() {
		// Simulates the arrival of an itemStateChanged-event
		// from a Checkbox
		
		MockSite site = new MockSite();
		SimFailureControl control = new SimFailureControl(site);
		
		control.itemStateChanged(new ItemEvent(new JCheckBox("Test1"), 0, null,
				1));
		assertEquals(site.getLastFailedDevice(), "Test1");

		control.itemStateChanged(new ItemEvent(new JCheckBox("Test2"), 0, null,
				0));
		assertEquals(site.getLastRepairedDevice(), "Test2");
	}

	@Test
	public void testItemStateChangedNullArgument() {
		// Should work
		MockSite site = new MockSite();
		SimFailureControl control = new SimFailureControl(site);
		control.itemStateChanged(null);
	}

}
