package simulation;

import static org.junit.Assert.*;

import java.util.Calendar;
import java.util.Iterator;
import java.util.Observable;

import org.junit.Before;
import org.junit.Test;

import tools.TestObserver;
import static tools.HelperFunctions.*;

public class SimSiteTest {

	// Mock SimSite created for testing purposes.
	private class TestSimSite extends SimSite {

		private int notifications = 0;

		int getNotifications() {
			return notifications;
		}

		@Override
		public void update(Observable subject, Object arg) {
			notifications++;
			super.update(subject, arg);
		}

	}

	private TestObserver observer;

	@Before
	public void init() {
		observer = new TestObserver();
	}

	@SuppressWarnings("unchecked")
	@Test
	public void testSimSiteConstructor() {
		// Tests the initial state of a SimSite.

		TestSimSite site = new TestSimSite();
		int notifications = site.getNotifications();

		// The default evaporation rate should be 10.
		assertEquals(10, site.getEvaporationRate());

		int numberOfZones = 0;
		// For each zone in the site ...
		for (Iterator<SimZone> i = site.zoneIterator(); i.hasNext();) {
			// ... increase the number of zones, ...
			numberOfZones++;

			SimZone currentZone = i.next();

			// ... check the zone's name ...
			assertTrue(currentZone.getName().matches("Z\\d{1,3}"));
			SimSensor currentSensor = currentZone.getSensor();

			// ... as well as its sensor's properties ...
			assertTrue(currentSensor.getName().matches("S\\d{1,3}"));
			assertEquals(40, currentSensor.getLevel());
			assertFalse(currentSensor.getIsFailed());
			currentZone.getSensor().setIsFailed();

			// ... and check whether a change in the sensor's state
			// causes a notification.
			assertEquals(++notifications, site.getNotifications());

			int numberOfValves = 0;

			// Then, inspect the zone's valves, which means ...
			for (Iterator<SimValve> vi = currentZone.valveIterator(); vi
					.hasNext();) {

				// ... count them ...
				numberOfValves++;

				SimValve currentValve = vi.next();

				// ... and check their properties.
				assertTrue(currentValve.getName().matches("V\\d{1,3}"));
				assertEquals(12, currentValve.getFlowRate());
				assertFalse(currentValve.getIsFailed());
				assertFalse(currentValve.isOpen());

				// Finally, check whether the observer is
				// notified when a valve gets changed.
				currentValve.open();
				assertEquals(++notifications, site.getNotifications());
			}
			assertEquals(4, numberOfValves);
		}
		assertEquals(3, numberOfZones);

		SimTime.instance().setSecond(
				Calendar.getInstance().get(Calendar.SECOND) + 10);
		assertEquals(++notifications, site.getNotifications());

	}

	@Test
	public void testSetEvaporationRate() {
		// Tests setEvaporationRate().

		SimSite site = new SimSite();
		site.addObserver(observer);

		// rate = 0
		site.setEvaporationRate(0);
		assertEquals(0, site.getEvaporationRate());

		// rate = 100
		site.setEvaporationRate(100);
		assertEquals(100, site.getEvaporationRate());
	}

	@Test
	public void testSetEvaporationRateIllegalArguments() {
		// Tests how setEvaporationRate() checks its precondition.

		SimSite site = new SimSite();
		site.addObserver(observer);

		try {
			site.setEvaporationRate(101);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			site.setEvaporationRate(Integer.MAX_VALUE);
			fail();
		} catch (IllegalArgumentException e) {
			
		}
		try {
			site.setEvaporationRate(Integer.MIN_VALUE);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			site.setEvaporationRate(-1);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		assertEquals(100, site.getEvaporationRate());
		assertEquals(0, observer.notifications());

	}

	@Test
	public void testSetEvaporationRateToDefault() {
		// Tests setEvaporationRateToDefault().

		SimSite site = new SimSite();
		site.setEvaporationRate(50);
		site.setEvaporationRateToDefault();
		assertEquals(10, site.getEvaporationRate());
	}

	@SuppressWarnings("unchecked")
	@Test
	public void testSetFailed() {
		// Tests setFailed().

		TestSimSite site = new TestSimSite();
		SimValve valve = null;
		SimSensor sensor = null;

		// Extract some of the devices contained in the site.
		Iterator<SimZone> i = site.zoneIterator();
		i.next();
		sensor = i.next().getSensor();
		Iterator<SimValve> iv = i.next().valveIterator();
		iv.next();
		iv.next();
		iv.next();
		valve = iv.next();

		// Change the state of these devices and see whether the observer
		// gets notifications.
		site.setFailed(sensor.getName());
		assertTrue(sensor.getIsFailed());
		assertEquals(1, site.getNotifications());

		site.setFailed(valve.getName());
		assertTrue(valve.getIsFailed());
		assertEquals(2, site.getNotifications());

		// Call setIsFailed twice.
		site.setFailed(valve.getName());
		assertTrue(valve.getIsFailed());

		// No additional notification expected,
		// since sensor's state did not change
		assertEquals(2, site.getNotifications());
	}

	@SuppressWarnings("unchecked")
	@Test
	public void testSetRepaired() {
		TestSimSite site = new TestSimSite();
		SimValve valve = null;
		SimSensor sensor = null;

		// Extract some of the devices contained in the site.
		Iterator<SimZone> i = site.zoneIterator();
		i.next();
		sensor = i.next().getSensor();
		Iterator<SimValve> iv = i.next().valveIterator();
		iv.next();
		iv.next();
		iv.next();
		valve = iv.next();

		site.setFailed(valve.getName());
		site.setFailed(sensor.getName());

		// Repair the devices.
		site.setRepaired(valve.getName());
		assertFalse(valve.getIsFailed());
		assertEquals(1, site.getNotifications());

		site.setRepaired(sensor.getName());
		assertFalse(sensor.getIsFailed());
		assertEquals(2, site.getNotifications());

		// Repair a device again.
		site.setRepaired(sensor.getName());
		assertFalse(sensor.getIsFailed());
		// No additional notification expected,
		// since sensor's state did not change
		assertEquals(2, site.getNotifications());
	}

	@Test
	public void testSetRepairedSetFailedIllegalArguments() {

		// Checks whether setFailed and setRepaird check
		// their preconditions.

		TestSimSite site = new TestSimSite();

		try {
			site.setFailed(null);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			site.setRepaired(null);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		// a device with the specified name does not exist

		try {
			site.setFailed("ABC");
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			site.setRepaired("ABC");
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		assertEquals(4, site.getNotifications());

	}

	@SuppressWarnings("unchecked")
	@Test
	public void testUpdateNotCalledBySimTime() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests update(). The SimTime was not the reason why update()
		// was called.

		SimTime time = SimTime.instance();
		time.setSecond(0);

		SimSite site = new SimSite();
		site.addObserver(observer);

		site.update(new SimValve(null, 0), null);
		assertEquals(1, observer.notifications());

		// The moisture level should not change.
		for (Iterator<SimZone> i = site.zoneIterator(); i.hasNext();) {
			SimZone zone = i.next();
			assertEquals(40.0, extractField(zone, "soilMoisture"));
			assertEquals(40, zone.getSensor().getLevel());
		}

		site.update(new SimSensor(""), "A");
		assertEquals(2, observer.notifications());

		// The moisture level should not change.
		for (Iterator<SimZone> i = site.zoneIterator(); i.hasNext();) {
			SimZone zone = i.next();
			assertEquals(40.0, extractField(zone, "soilMoisture"));
			assertEquals(40, zone.getSensor().getLevel());
		}
	}

	@SuppressWarnings("unchecked")
	@Test
	public void testUpdateCalledBySimTime() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests update(). SimTime caused the call to update().

		SimTime time = SimTime.instance();
		time.setSecond(0);

		SimSite site = new SimSite();
		site.addObserver(observer);

		site.update(time, null);
		// If update is called when a full minute has passed,
		// the moisture level has to be adjusted.
		double newMoisture = 40.0 - (40.0 * (10 / 100.0 / 24 / 60));
		for (Iterator<SimZone> i = site.zoneIterator(); i.hasNext();) {
			SimZone zone = i.next();
			assertEquals(newMoisture, extractField(zone, "soilMoisture"));
			assertEquals(39, zone.getSensor().getLevel());
		}

		// If no minute passed, the moisture level should stay the same.
		time.setSecond(15);

		site.update(time, new Integer(5));
		for (Iterator<SimZone> i = site.zoneIterator(); i.hasNext();) {
			SimZone zone = i.next();
			assertEquals(newMoisture, extractField(zone, "soilMoisture"));
			assertEquals(39, zone.getSensor().getLevel());
		}
	}

	@SuppressWarnings("unchecked")
	@Test
	public void testUpdateIllegalArguments() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests update() with arguments which are not the
		// SimTime nor a device.

		SimSite site = new SimSite();
		site.addObserver(observer);

		site.update(null, "A");

		// No notifications are expected.
		assertEquals(0, observer.notifications());

		// The moisture level should not change.
		for (Iterator<SimZone> i = site.zoneIterator(); i.hasNext();) {
			SimZone zone = i.next();
			assertEquals(40.0, extractField(zone, "soilMoisture"));
			assertEquals(40, zone.getSensor().getLevel());
		}

		site.update(new Observable(), "B");

		// No notifications are expected.
		assertEquals(0, observer.notifications());

		// The moisture level should not change.
		for (Iterator<SimZone> i = site.zoneIterator(); i.hasNext();) {
			SimZone zone = i.next();
			assertEquals(40.0, extractField(zone, "soilMoisture"));
			assertEquals(40, zone.getSensor().getLevel());
		}
	}

}
