package simulation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static tools.HelperFunctions.assertObserved;
import static tools.HelperFunctions.extractField;
import irrigation.Irrigator;

import java.util.HashSet;
import java.util.Hashtable;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;

import org.junit.Before;
import org.junit.Test;

import startup.StartupException;
import util.Command;
import util.Day;
import device.Clock;
import device.StorageDevice;
import device.StoreFailureException;
import device.sim.MockSimClockDevice;

public class SimTimeControlTest {

	// Subclass of storage, created for testing purposes.
	private class MockStorage implements StorageDevice {
		private Hashtable<String, String> storage = new Hashtable<String, String>();

		@Override
		public void setData(String name, String value)
				throws StoreFailureException {
			storage.put(name, value);
		}

		@Override
		public void removeData(String name) throws StoreFailureException {
			storage.remove(name);
		}

		@Override
		public String getData(String name, String defaultValue)
				throws StoreFailureException {
			return storage.get(name);
		}
	}

	private SimTimeControl control;
	private JPanel labelPanel;
	private JLabel timeLabel;
	private JLabel secondLabel;
	private JPanel buttonPanel;
	private JButton pauseButton;
	private JButton resumeButton;
	private JButton jumpButton;
	private JButton fasterButton;
	private JButton slowerButton;
	private JButton defaultButton;

	@Before
	public void init() {
		// Make some interesting objects better accessible.
		control = new SimTimeControl();
		labelPanel = (JPanel) control.getComponent(0);
		timeLabel = (JLabel) labelPanel.getComponent(0);
		secondLabel = (JLabel) labelPanel.getComponent(1);
		buttonPanel = (JPanel) control.getComponent(1);
		pauseButton = (JButton) buttonPanel.getComponent(0);
		resumeButton = (JButton) buttonPanel.getComponent(1);
		jumpButton = (JButton) buttonPanel.getComponent(2);
		fasterButton = (JButton) buttonPanel.getComponent(3);
		slowerButton = (JButton) buttonPanel.getComponent(4);
		defaultButton = (JButton) buttonPanel.getComponent(5);
	}

	@Test
	public void testSimTimeControl() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests SimTimeControl's initial state.
		SimTime time = SimTime.instance();
		assertEquals(time, extractField(control, "time"));
		
		// SimTimeControl should observe the SimTime.
		assertObserved(time, control);

		assertNull(extractField(control, "irrigator"));

		assertNotNull(timeLabel);
		assertNotNull(secondLabel);

		// Check whether all buttons are created and properly initialized,
		// that is whether they have some event listeners attached.
		assertNotNull(pauseButton);
		assertTrue(pauseButton.getActionListeners().length > 0);
		assertNotNull(resumeButton);
		assertTrue(resumeButton.getActionListeners().length > 0);
		assertNotNull(jumpButton);
		assertTrue(jumpButton.getActionListeners().length > 0);
		// No irrigator is set, so no jump is possible since no
		// future irrigation time is known ...
		assertFalse(jumpButton.isEnabled());
		assertNotNull(fasterButton);
		assertTrue(fasterButton.getMouseListeners().length > 0);
		assertNotNull(slowerButton);
		assertTrue(slowerButton.getMouseListeners().length > 0);
		assertNotNull(defaultButton);
		assertTrue(defaultButton.getActionListeners().length > 0);
	}

	@Test
	public void testSimTimeControlAdaptionToProgramState() {
		// Test whether the time control adapts to the
		// clock's state

		SimTime time = SimTime.instance();

		// Tests the "Pause" and "Resume" buttons
		time.stop();
		init();

		assertFalse(pauseButton.isEnabled());
		assertTrue(resumeButton.isEnabled());

		time.start();
		init();

		assertTrue(pauseButton.isEnabled());
		assertFalse(resumeButton.isEnabled());
	}

	@Test
	public void testSimTimeControlAdaptionToProgramState2() {
		// Tests the buttons controlling the clock's speed.
		SimTime time = SimTime.instance();
		
		time.setSpeed(1);
		init();

		assertFalse(slowerButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
		assertTrue(fasterButton.isEnabled());

		time.setSpeed(100);
		init();

		assertTrue(slowerButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
		assertTrue(fasterButton.isEnabled());

		time.setSpeed(1000);
		init();

		assertTrue(slowerButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
		assertFalse(fasterButton.isEnabled());
	}

	@Test
	public void testRunningTimerOnOff() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			InterruptedException {
		// Tests basic interaction with the time control,
		// namely switching the clock on or off.
		SimTime time = SimTime.instance();
		time.setSpeed(900);
		time.start();

		init();

		// Pause the clock.
		pauseButton.doClick();

		Thread.sleep(10);

		assertFalse((Boolean) extractField(time, "isRunning"));
		assertFalse(pauseButton.isEnabled());
		assertTrue(resumeButton.isEnabled());

		// Resume the clock.
		resumeButton.doClick();

		assertTrue((Boolean) extractField(time, "isRunning"));
		assertNotNull(extractField(time, "pulseThread"));
		assertTrue(pauseButton.isEnabled());
		assertFalse(resumeButton.isEnabled());

		// Pause the clock again.
		pauseButton.doClick();

		Thread.sleep(10);

		assertFalse((Boolean) extractField(time, "isRunning"));
		assertFalse(pauseButton.isEnabled());
		assertTrue(resumeButton.isEnabled());

	}

	@Test
	public void testStoppedTimerOnOff() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			InterruptedException {
		// Tests basic interaction with the time control,
		// namely switching the clock on or off.
		SimTime time = SimTime.instance();
		time.setSpeed(900);
		time.stop();

		init();

		// Resume the clock.
		resumeButton.doClick();

		assertTrue((Boolean) extractField(time, "isRunning"));
		assertNotNull(extractField(time, "pulseThread"));
		assertTrue(pauseButton.isEnabled());
		assertFalse(resumeButton.isEnabled());

		// Pause it.
		pauseButton.doClick();

		Thread.sleep(10);

		assertFalse((Boolean) extractField(time, "isRunning"));
		assertFalse(pauseButton.isEnabled());
		assertTrue(resumeButton.isEnabled());

		// Resume it again.
		resumeButton.doClick();

		assertTrue((Boolean) extractField(time, "isRunning"));
		assertNotNull(extractField(time, "pulseThread"));
		assertTrue(pauseButton.isEnabled());
		assertFalse(resumeButton.isEnabled());

	}

	@Test
	public void testJumpButtonActionListenerMoreThanOneHour()
			throws StartupException, IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests the "Jump" button
		SimTime time = SimTime.instance();

		Clock.setClockDevice(new MockSimClockDevice());

		// Configure an Irrigator.
		Irrigator i = new Irrigator(new MockStorage());
		HashSet<Day> daySet = new HashSet<Day>();
		daySet.add(Day.MONDAY);
		daySet.add(Day.FRIDAY);
		daySet.add(Day.SUNDAY);
		i.setIrrigationDays(daySet);
		i.setIrrigationTime(1400);

		control.setIrrigator(i);

		// More than one hour left until next irrigation ...

		time.setDay(Day.MONDAY);
		time.setHour(12);
		time.setMinute(59);
		time.setSecond(59);

		// ... so lets jump to the next irrigation time.
		jumpButton.doClick();

		assertEquals(Day.MONDAY, time.getDay());
		assertEquals(13, time.getHour());
		assertEquals(0, time.getMinute());
		assertEquals(0, time.getSecond());
	}

	@Test
	public void testJumpButtonActionListenerLessThanOneHour() {
		// Tests the "Jump" button
		SimTime time = SimTime.instance();

		Clock.setClockDevice(new MockSimClockDevice());

		// Configure an Irrigator.
		Irrigator i = new Irrigator(new MockStorage());
		HashSet<Day> daySet = new HashSet<Day>();
		daySet.add(Day.MONDAY);
		daySet.add(Day.FRIDAY);
		daySet.add(Day.SUNDAY);
		i.setIrrigationDays(daySet);
		i.setIrrigationTime(1400);

		control.setIrrigator(i);

		// Less than or exactly one hour left until next irrigation ...

		time.setDay(Day.MONDAY);
		time.setHour(13);
		time.setMinute(00);

		// ... but we still want to jump to the next irrigation time ...
		jumpButton.doClick();

		// ... so we jump to the next irrigation time which is at least
		// one our away.
		assertEquals(Day.FRIDAY, time.getDay());
		assertEquals(13, time.getHour());
		assertEquals(0, time.getMinute());
		assertEquals(0, time.getSecond());

	}

	@Test
	public void testSetIrrigator() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests whether the control adapts its state to the presence
		// of an irrigator.
		Irrigator i = new Irrigator(new MockStorage());

		control.setIrrigator(i);

		assertEquals(i, extractField(control, "irrigator"));
		assertTrue(jumpButton.isEnabled());

		// Setting irrigator back to null.

		control.setIrrigator(null);

		assertNull(extractField(control, "irrigator"));
		// no irrigator available, so no future irrigation
		// time can be computed ...
		assertFalse(jumpButton.isEnabled());

	}

	@Test
	public void testFasterButtonMouseListenerLowSpeed()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests the "Faster" button increasing the speed
		// of the clock.
		// 1 <= time.getSpeed() <= 30
		SimTime time = SimTime.instance();

		ButtonRepeater repeater = (ButtonRepeater) fasterButton
				.getMouseListeners()[1];
		Command cmd = extractField(repeater, "cmd");

		time.setSpeed(1);
		cmd.execute(null);

		assertEquals(2, time.getSpeed());
		assertTrue(fasterButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
		assertTrue(slowerButton.isEnabled());

		time.setSpeed(30);
		cmd.execute(null);

		assertEquals(31, time.getSpeed());
		assertTrue(fasterButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
		assertTrue(slowerButton.isEnabled());
	}

	@Test
	public void testFasterButtonMouseListenerHighSpeed()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests the "Faster" button increasing the speed
		// of the clock.
		// time.getSpeed() >= 31
		SimTime time = SimTime.instance();

		ButtonRepeater repeater = (ButtonRepeater) fasterButton
				.getMouseListeners()[1];
		Command cmd = extractField(repeater, "cmd");

		// Repeatedly click the faster button to increase the
		// speed to 1000.
		int[] speeds = { 62, 125, 250, 500, 1000, 1000 };

		for (int i = 0; i < speeds.length; i++) {
			cmd.execute(null);
			assertEquals(speeds[i], time.getSpeed());
			if (i < 4)
				assertTrue(fasterButton.isEnabled());
			else
				assertFalse(fasterButton.isEnabled());
			assertTrue(defaultButton.isEnabled());
			assertTrue(slowerButton.isEnabled());
		}
	}

	@Test
	public void testSlowerButtonMouseListenerSlowSpeed()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests the "Slower" button lowering the speed
		// of the clock.
		// 2 <= time.getSpeed() <= 31
		SimTime time = SimTime.instance();

		ButtonRepeater repeater = (ButtonRepeater) slowerButton
				.getMouseListeners()[1];
		Command cmd = extractField(repeater, "cmd");

		// Lowering the speed from 2 to 1
		time.setSpeed(2);
		cmd.execute(null);

		assertEquals(1, time.getSpeed());
		assertTrue(fasterButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
		assertFalse(slowerButton.isEnabled());

		cmd.execute(null);

		assertEquals(1, time.getSpeed());
		assertTrue(fasterButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
		assertFalse(slowerButton.isEnabled());

		// Lowering the speed from 31 to 30
		time.setSpeed(31);
		cmd.execute(null);

		assertEquals(30, time.getSpeed());
		assertTrue(fasterButton.isEnabled());
		assertTrue(defaultButton.isEnabled());
		assertTrue(slowerButton.isEnabled());
	}

	@Test
	public void testSlowerButtonMouseListenerHighSpeed()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests the "Slower" button lowering the speed
		// of the clock.
		// time.getSpeed() >= 31
		SimTime time = SimTime.instance();

		ButtonRepeater repeater = (ButtonRepeater) slowerButton
				.getMouseListeners()[1];
		Command cmd = extractField(repeater, "cmd");

		// Repeatedly click the "Slower" button to
		// decrease the speed to 30.
		time.setSpeed(1000);

		int[] speeds = { 31, 62, 125, 250, 500 };

		for (int i = speeds.length - 1; i >= 0; i--) {
			cmd.execute(null);
			assertEquals(speeds[i], time.getSpeed());
			assertTrue(slowerButton.isEnabled());
			assertTrue(defaultButton.isEnabled());
			assertTrue(fasterButton.isEnabled());
		}
	}

	@Test
	public void testDefaultButtonActionListener() {
		// Tests the default button reseting the clock's
		// speed to 1.
		SimTime time = SimTime.instance();

		defaultButton.doClick();
		assertEquals(1, time.getSpeed());
		assertFalse(slowerButton.isEnabled());
		assertTrue(fasterButton.isEnabled());
		assertTrue(defaultButton.isEnabled());

		// Current speed is 500.
		time.setSpeed(500);
		defaultButton.doClick();
		assertEquals(1, time.getSpeed());
		assertFalse(slowerButton.isEnabled());
		assertTrue(fasterButton.isEnabled());
		assertTrue(defaultButton.isEnabled());

		// Current speed is 1000.
		time.setSpeed(1000);
		defaultButton.doClick();
		assertEquals(1, time.getSpeed());
		assertFalse(slowerButton.isEnabled());
		assertTrue(fasterButton.isEnabled());
		assertTrue(defaultButton.isEnabled());

	}

	@Test
	public void testUpdateObservableObject() {
		// Tests whether the proper time is
		// displayed in the control.

		SimTime time = SimTime.instance();

		time.setDay(Day.MONDAY);
		time.setHour(6);
		time.setMinute(24);
		time.setSecond(31);

		timeLabel.setText("");

		control.update(null, null);

		assertEquals("Monday 06:24:31", timeLabel.getText());

		time.setDay(Day.FRIDAY);
		time.setHour(4);
		time.setMinute(8);
		time.setSecond(5);

		timeLabel.setText("");

		control.update(null, null);

		assertEquals("Friday 04:08:05", timeLabel.getText());

	}

}
