package simulation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static tools.HelperFunctions.extractField;
import static tools.HelperFunctions.extractFieldAsField;
import static tools.HelperFunctions.isSynchronized;

import java.lang.reflect.Field;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Map;

import org.junit.Before;
import org.junit.Test;

import tools.TestObserver;
import util.Day;

public class SimTimeTest {

	private Map<Integer, Day> dayMap = new HashMap<Integer, Day>();
	private TestObserver observer;

	public SimTimeTest() {
		dayMap.put(Calendar.MONDAY, Day.MONDAY);
		dayMap.put(Calendar.TUESDAY, Day.TUESDAY);
		dayMap.put(Calendar.WEDNESDAY, Day.WEDNESDAY);
		dayMap.put(Calendar.THURSDAY, Day.THURSDAY);
		dayMap.put(Calendar.FRIDAY, Day.FRIDAY);
		dayMap.put(Calendar.SATURDAY, Day.SATURDAY);
		dayMap.put(Calendar.SUNDAY, Day.SUNDAY);
	}

	@Before
	public void init() {
		observer = new TestObserver();
	}

	@Test
	public void testInstance() {
		SimTime time = SimTime.instance();

		assertEquals(time, SimTime.instance());
	}

	@Test
	public void testSimTime() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Check whether time stores the current time

		// Reset the SimTime
		Field instance = extractFieldAsField(SimTime.instance(), "instance");
		instance.setAccessible(true);
		instance.set(null, null);
		
		SimTime time = SimTime.instance();

		Calendar now = Calendar.getInstance();

		assertEquals(now.get(Calendar.SECOND), time.getSecond());
		assertEquals(now.get(Calendar.MINUTE), time.getMinute());
		assertEquals(now.get(Calendar.HOUR_OF_DAY), time.getHour());
		assertEquals(dayMap.get((Integer) now.get(Calendar.DAY_OF_WEEK)),
				time.getDay());

		assertEquals(1000L, extractField(time, "tickInterval"));
		assertEquals(false, extractField(time, "isRunning"));
		assertNull(extractField(time, "pulseThread"));
	}

	@Test
	public void testSetDaySynchronized() throws SecurityException,
			NoSuchMethodException {
		// Ensure that setDay is synchronized.

		assertTrue(isSynchronized(SimTime.instance().getClass(), "setDay",
				new Class[] { Day.class }));
	}

	@Test
	public void testSetDay() {
		// Check whether setDay is working properly.

		SimTime time = SimTime.instance();
		time.addObserver(observer);

		time.setDay(null);
		assertNull(time.getDay());
		assertEquals(1, observer.notifications());

		time.setDay(Day.SUNDAY);
		assertEquals(Day.SUNDAY, time.getDay());
		assertEquals(2, observer.notifications());
	}

	@Test
	public void testSetDayNotifications() {
		// Tests setDay() and the notifications it causes.
		SimTime time = SimTime.instance();
		time.addObserver(observer);
		time.setDay(Day.SUNDAY);
		time.setDay(Day.SUNDAY);
		assertEquals(1, observer.notifications()); // time's state did not
													// change, so no
													// notification
													// is expected

	}

	@Test
	public void testGetDay() throws SecurityException, NoSuchMethodException {

		// Test whether getDay() is synchronized
		assertTrue(isSynchronized(SimTime.instance().getClass(), "getDay",
				new Class[] {}));
	}

	@Test
	public void testSetHourSynchronized() throws SecurityException,
			NoSuchMethodException {

		// Test whether setHour() is synchronized

		assertTrue(isSynchronized(SimTime.instance().getClass(), "setHour",
				new Class[] { int.class }));
	}

	@Test
	public void testSetHour() {
		// Tests setHour().

		SimTime time = SimTime.instance();
		time.addObserver(observer);
		time.setHour(0);

		assertEquals(0, time.getHour());
		assertEquals(1, observer.notifications());

		time.setHour(23);
		assertEquals(23, time.getHour());
		assertEquals(2, observer.notifications());

	}

	@Test
	public void testSetHourNotifications() {
		// Tests setHour() and the notifications it causes.

		SimTime time = SimTime.instance();
		time.addObserver(observer);
		time.setHour(23);
		time.setHour(23);
		assertEquals(1, observer.notifications()); // time's state did not
													// change, therefore
													// no notification is
													// expected

	}

	@Test
	public void testSetHourIllegalArguments() {

		// The following test cases cause setHour
		// to throw an exception and leave time unchanged

		SimTime time = SimTime.instance();
		time.addObserver(observer);
		time.setHour(23);

		try {
			time.setHour(24);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			time.setHour(Integer.MAX_VALUE);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			time.setHour(Integer.MIN_VALUE);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			time.setHour(-1);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		assertEquals(23, time.getHour());
		assertEquals(1, observer.notifications());

	}

	@Test
	public void testGetHourSynchronized() throws SecurityException,
			NoSuchMethodException {
		// Test whether getHour() is synchronized
		assertTrue(isSynchronized(SimTime.instance().getClass(), "getHour",
				new Class[] {}));
	}

	@Test
	public void testSetMinuteSynchronized() throws SecurityException,
			NoSuchMethodException {

		// Test whether setMinute() is synchronized

		assertTrue(isSynchronized(SimTime.instance().getClass(), "setMinute",
				new Class[] { int.class }));
	}

	@Test
	public void testSetMinute() {
		// Tests setMinute().

		SimTime time = SimTime.instance();
		time.addObserver(observer);

		time.setMinute(0);
		assertEquals(0, time.getMinute());
		assertEquals(1, observer.notifications());

		time.setMinute(59);
		assertEquals(59, time.getMinute());
		assertEquals(2, observer.notifications());

	}

	@Test
	public void testSetMinuteNotifications() {
		// Tests setMinute() and the notifications it causes.

		SimTime time = SimTime.instance();
		time.addObserver(observer);
		time.setMinute(59);
		time.setMinute(59);
		assertEquals(1, observer.notifications()); // time's state did not
													// change, therefore
													// no notification is
													// expected

	}

	@Test
	public void testSetMinuteInvalidArguments() {

		// The following test cases should cause setMinute
		// to throw an exception an leave time unchanged

		SimTime time = SimTime.instance();
		time.addObserver(observer);
		time.setMinute(59);

		try {
			time.setMinute(60);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			time.setMinute(Integer.MAX_VALUE);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			time.setMinute(Integer.MIN_VALUE);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			time.setMinute(-1);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		assertEquals(59, time.getMinute());
		assertEquals(1, observer.notifications());

	}

	@Test
	public void testGetMinute() throws SecurityException, NoSuchMethodException {
		// Test whether getMinute() is synchronized

		assertTrue(isSynchronized(SimTime.instance().getClass(), "getMinute",
				new Class[] {}));
	}

	@Test
	public void testSetSecondSynchronized() throws SecurityException,
			NoSuchMethodException {
		// Test whether setSecond() is synchronized

		assertTrue(isSynchronized(SimTime.instance().getClass(), "setSecond",
				new Class[] { int.class }));
	}

	@Test
	public void testSetSecond() {
		// Tests setSecond().

		SimTime time = SimTime.instance();
		time.addObserver(observer);

		time.setSecond(0);
		assertEquals(0, time.getSecond());
		assertEquals(1, observer.notifications());

		time.setSecond(59);
		assertEquals(59, time.getSecond());
		assertEquals(2, observer.notifications());
	}

	@Test
	public void testSetSecondNotifications() {
		// Tests setSecond() and the notifications it causes.

		SimTime time = SimTime.instance();
		time.addObserver(observer);
		time.setSecond(59);
		time.setSecond(59);
		assertEquals(1, observer.notifications()); // time's state did not
													// change, therefore
													// no notification is
													// expected

	}

	@Test
	public void testSetSecondIllegalArguments() {

		// The following test cases should cause setSecond
		// to throw an exception an leave time unchanged

		SimTime time = SimTime.instance();
		time.addObserver(observer);
		time.setSecond(59);

		try {
			time.setSecond(60);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			time.setSecond(Integer.MAX_VALUE);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			time.setSecond(Integer.MIN_VALUE);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			time.setSecond(-1);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		assertEquals(59, time.getSecond());
		assertEquals(1, observer.notifications());

	}

	@Test
	public void testGetSecond() throws SecurityException, NoSuchMethodException {
		// Test whether getSecond() is synchronized

		assertTrue(isSynchronized(SimTime.instance().getClass(), "getSecond",
				new Class[] {}));
	}

	@Test
	public void testSetSpeed() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests setSpeed().

		SimTime time = SimTime.instance();
		time.setSpeed(1);
		assertEquals(1, time.getSpeed());
		assertEquals(1000L, extractField(time, "tickInterval"));

		time.setSpeed(1000);
		assertEquals(1000, time.getSpeed());
		assertEquals(1L, extractField(time, "tickInterval"));

		time.setSpeed(50);
		assertEquals(50, time.getSpeed());
		assertEquals(20L, extractField(time, "tickInterval"));

	}

	@Test
	public void testSetSpeedIllegalCases() {

		// Tests setSpeed().

		SimTime time = SimTime.instance();

		try {
			time.setSpeed(0);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			time.setSpeed(Integer.MIN_VALUE);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			time.setSpeed(Integer.MAX_VALUE);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			time.setSpeed(1001);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

	}

	@Test
	public void testSetSpeedToDefault() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Test whether setSpeedToDefault sets time's speed
		// back to 1000.

		SimTime time = SimTime.instance();
		time.setSpeed(50);

		time.setSpeedToDefault();
		assertEquals(1, time.getSpeed());
		assertEquals(1000L, extractField(time, "tickInterval"));

	}

	@Test
	public void testStart() throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, InterruptedException {
		// Test start().

		SimTime time = SimTime.instance();
		time.setSpeed(100);

		time.start();

		Thread thread = extractField(time, "pulseThread");
		assertTrue(thread.isAlive());
		assertTrue((Boolean) extractField(time, "isRunning"));

		// Call start() again.
		time.start();

		assertEquals(thread, extractField(time, "pulseThread"));
		assertTrue(thread.isAlive());
		assertTrue((Boolean) extractField(time, "isRunning"));

	}

	@Test
	public void testStop() throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, InterruptedException {
		// Tests stop().

		// Start the time.
		SimTime time = SimTime.instance();
		time.setSpeed(900);
		time.start();

		time.stop();

		Thread.sleep(30);

		assertFalse(((Thread) extractField(time, "pulseThread")).isAlive());
		assertFalse((Boolean) extractField(time, "isRunning"));

		// Call stop() again.
		time.stop();

		assertFalse(((Thread) extractField(time, "pulseThread")).isAlive());
		assertFalse((Boolean) extractField(time, "isRunning"));

	}

	@Test
	public void testRunStartedInExternalThread() throws InterruptedException,
			IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests run(). A stopped thread cannot run without
		// calling start() first

		SimTime time = SimTime.instance();
		time.setSecond(0);
		time.setSpeed(500);

		Thread thread = new Thread(time);
		thread.start();

		Thread.sleep(50);

		assertFalse(thread.isAlive());
		assertFalse((Boolean) extractField(time, "isRunning"));
	}

	@Test
	public void testRunDependsOnIsRunning() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException,
			InterruptedException {
		// Tests run(). run() is only executed
		// as long as isRunnig is true.

		SimTime time = SimTime.instance();
		time.setSpeed(500);

		time.start();

		Thread thread = extractField(time, "pulseThread");
		assertTrue(thread.isAlive());

		Field keepRunningField = extractFieldAsField(time, "isRunning");
		keepRunningField.setAccessible(true);
		keepRunningField.setBoolean(time, false);

		Thread.sleep(10);

		assertFalse(thread.isAlive());
	}

	@Test
	public void testRun() throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException, InterruptedException {
		SimTime time = SimTime.instance();

		observer = new TestObserver();
		time.addObserver(observer);
		time.setSecond(10);
		time.setMinute(59);
		time.setHour(0);
		time.setSpeed(200);

		time.start();
		Thread.sleep(50);
		time.stop();

		Thread.sleep(10);

		assertEquals(0, time.getHour());
		assertEquals(59, time.getMinute());

		// Every time a second is added, observers observing
		// time should be informed
		assertEquals(time.getSecond() - 10 + 3, observer.notifications());
		// Additionally, at least 10 seconds should pass.
		assertTrue(Long.toString(time.getSecond()), time.getSecond() - 10 >= 10);

		assertFalse(((Thread) extractField(time, "pulseThread")).isAlive());

	}

}
