package simulation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static tools.HelperFunctions.extractFieldAsField;

import java.lang.reflect.Field;

import org.junit.Before;
import org.junit.Test;

import tools.TestObserver;

public class SimValveTest {

	private TestObserver observer;

	@Before
	public void init() {
		observer = new TestObserver();
	}

	@Test
	public void testSimValve() {
		// Test SimValve's constructor.

		SimValve valve = new SimValve("sim1", Integer.MIN_VALUE);

		assertEquals("sim1", valve.getName());
		assertEquals(Integer.MIN_VALUE, valve.getFlowRate());
		assertFalse(valve.isOpen());
		assertFalse(valve.getIsFailed());
	}

	@Test
	public void testSimValveDifferentArguments() {
		// Test SimValve's constructor.

		SimValve valve = new SimValve("sim2", 0);

		assertEquals("sim2", valve.getName());
		assertEquals(0, valve.getFlowRate());
		assertFalse(valve.isOpen());
		assertFalse(valve.getIsFailed());
	}

	@Test
	public void testSimValveDifferentArguments2() {
		// Test SimValve's constructor.

		SimValve valve = new SimValve(null, Integer.MAX_VALUE);

		assertEquals(null, valve.getName());
		assertEquals(Integer.MAX_VALUE, valve.getFlowRate());
		assertFalse(valve.isOpen());
		assertFalse(valve.getIsFailed());

	}

	@Test
	public void testOpen() {
		// Tests open().

		SimValve valve = new SimValve("valve", 0);
		valve.addObserver(observer);
		int result = valve.open();

		assertTrue(valve.isOpen());
		assertEquals(1, result);
		assertTrue(observer.notifications() > 0);
	}

	@Test
	public void testOpenValveFailure() {
		// Tests open(). The device fails.

		SimValve valve = new SimValve("valve", 0);
		valve.setIsFailed();
		observer = new TestObserver();
		valve.addObserver(observer);

		int result = valve.open();

		assertFalse(valve.isOpen());
		assertEquals(-1, result);
		assertFalse(observer.notifications() > 0);
	}

	@Test
	public void testOpenNotificationsCheck() {
		// Tests open(). The registered observer should
		// only be called when the valve's state is
		// changed.

		SimValve valve = new SimValve("valve", 5);
		valve.open();

		observer = new TestObserver();
		valve.addObserver(observer);

		valve.open();

		assertEquals(0, observer.notifications()); // Since the valve was
													// already open,
													// it didn't change.

	}

	@Test
	public void testClose() throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests close().

		SimValve valve = new SimValve("valve", 0);
		valve.addObserver(observer);

		valve.open();
		int result = valve.close();

		assertFalse(valve.isOpen());
		assertEquals(1, result);
		assertTrue(observer.notifications() == 2);
	}

	@Test
	public void testCloseValveFailure() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests close(). The device fails.

		SimValve valve = new SimValve("valve", 0);
		observer = new TestObserver();
		valve.open();
		valve.addObserver(observer);

		Field failedField = extractFieldAsField(valve, "isFailed");
		failedField.setAccessible(true);
		failedField.set(valve, true);

		int result = valve.close();

		assertTrue(valve.isOpen());
		assertEquals(-1, result);
		assertFalse(observer.notifications() == 1);
	}

	@Test
	public void testCloseNotificationsCheck() {
		// Tests close(). The registered observer should
		// only be called when the valve's state is
		// changed.

		SimValve valve = new SimValve("name", 0);
		observer = new TestObserver();

		valve.open();
		valve.close();

		valve.addObserver(observer);
		valve.close();

		assertTrue(observer.notifications() == 0); // Since the valve was
													// already closed,
													// it didn't change.

	}

	@Test
	public void testSetIsFailed() {
		// Tests setIsFailed(). The registered observer should
		// only be called when the valve's state is
		// changed.
		
		SimValve valve = new SimValve("valve", 12);
		valve.open();

		TestObserver observer = new TestObserver();
		valve.addObserver(observer);

		valve.setIsFailed();

		assertFalse(valve.isOpen());
		assertTrue(valve.getIsFailed());
		assertEquals(1, observer.notifications());

		valve.setIsFailed();

		assertEquals(1, observer.notifications()); // Since valve's state didn't
													// change
													// no notification is
													// necessary
	}
}
