package simulation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static tools.HelperFunctions.extractField;
import static tools.HelperFunctions.extractFieldAsField;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Iterator;

import org.junit.Test;

public class SimZoneTest {

	@Test
	public void testSimZoneConstructor() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests SimZone's constructor.

		SimSensor sensor = new SimSensor("A");
		SimZone zone = new SimZone("zone", sensor);

		assertEquals("zone", extractField(zone, "name"));
		assertEquals(sensor, extractField(zone, "sensor"));
		assertNotNull(extractField(zone, "valves"));
		assertEquals(40.0, extractField(zone, "soilMoisture"));
		assertEquals(40, sensor.getLevel());
	}

	@Test
	public void testSimZoneConstructorDifferentArguments()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests SimZone's constructor with different arguments.
		SimSensor sensor = new SimSensor("A");
		SimZone zone = new SimZone(null, sensor);

		assertEquals(null, extractField(zone, "name"));
		assertEquals(sensor, extractField(zone, "sensor"));
		assertNotNull(extractField(zone, "valves"));
		assertEquals(40.0, extractField(zone, "soilMoisture"));
		assertEquals(40, sensor.getLevel());
	}

	@Test
	public void testSimZoneConstructorDifferentArguments2()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests SimZone's constructor with different arguments.
		SimSensor sensor = null;
		SimZone zone = new SimZone("zone2", sensor);

		assertEquals("zone", extractField(zone, "name"));
		assertEquals(sensor, extractField(zone, "sensor"));
		assertNotNull(extractField(zone, "valves"));
		assertEquals(40.0, extractField(zone, "soilMoisture"));

	}

	@Test
	public void testAdjustMoistureNoMoistureLoss()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests adjustMoisture().
		// Both valves are open, no moisture is lost.

		SimZone zone = new SimZone("A", new SimSensor("sensor"));

		zone.addValve(new SimValve("one", 2));
		zone.addValve(new SimValve("two", 3));

		for (Iterator<?> i = zone.valveIterator(); i.hasNext();)
			((SimValve) i.next()).open();

		zone.adjustMoisture(0);

		double newMoisture = 40.0 + (5.0 / 497.0);

		assertEquals(newMoisture, extractField(zone, "soilMoisture"));
		SimSensor sensor = (SimSensor) extractField(zone, "sensor");
		assertEquals(40, sensor.getLevel());
	}

	@Test
	public void testAdjustMoistureCompleteMoistureLoss()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests adjustMoisture().
		// One valves are open, all moisture is lost.

		SimZone zone = new SimZone("A", new SimSensor("sensor"));

		zone.addValve(new SimValve("one", 2));
		zone.addValve(new SimValve("two", 3));

		SimSensor sensor = (SimSensor) extractField(zone, "sensor");

		((SimValve) zone.valveIterator().next()).open();

		zone.adjustMoisture(100);

		double newMoisture = 40.0;
		newMoisture = newMoisture - (newMoisture / 1440.0) + (2.0 / 497.0);

		assertEquals(newMoisture, extractField(zone, "soilMoisture"));
		sensor = (SimSensor) extractField(zone, "sensor");
		assertEquals((int) newMoisture, sensor.getLevel());
	}

	@Test
	public void testAdjustMoistureFailedValve()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests adjustMoisture().
		// One valves is failing, all moisture is lost.

		SimZone zone = new SimZone("A", new SimSensor("sensor"));

		zone.addValve(new SimValve("one", 2));
		zone.addValve(new SimValve("two", 3));

		SimSensor sensor = (SimSensor) extractField(zone, "sensor");

		SimValve failedValve = new SimValve("A", 6);
		failedValve.open();
		Field failedField = extractFieldAsField(failedValve, "isFailed");
		failedField.setAccessible(true);
		failedField.set(failedValve, true);

		zone.addValve(failedValve);

		zone.adjustMoisture(100);

		double newMoisture = 40.0;
		newMoisture = newMoisture - (newMoisture / 1440.0) + (6.0 / 497.0);

		assertEquals(newMoisture, extractField(zone, "soilMoisture"));
		sensor = (SimSensor) extractField(zone, "sensor");
		assertEquals((int) newMoisture, sensor.getLevel());

	}

	@Test
	public void testAdjustMoistureSoilMoistureOverflow()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests adjustMoisture. No moisture is lost and
		// one of the valve's has a flow rate
		// that causes an integer overflow.

		SimZone zone = new SimZone("A", new SimSensor("sensor"));

		zone.addValve(new SimValve("one", 25));
		zone.addValve(new SimValve("two", Integer.MAX_VALUE - 20));

		for (Iterator<?> i = zone.valveIterator(); i.hasNext();)
			((SimValve) i.next()).open();

		zone.adjustMoisture(0);

		double soilMoisture = extractField(zone, "soilMoisture");
		SimSensor sensor = (SimSensor) extractField(zone, "sensor");

		assertTrue(soilMoisture >= 0); // a soil moisture less than 0 does
										// not make sense

		assertEquals(0, sensor.getLevel());

	}

	@Test
	public void testAdjustMoistureInvalidCases()
			throws IllegalArgumentException, SecurityException,
			IllegalAccessException, NoSuchFieldException {
		// Tests adjustMoisture. The argument is not valid.

		SimZone zone = new SimZone("A", new SimSensor("sensor"));

		try {
			zone.adjustMoisture(-1);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			zone.adjustMoisture(Integer.MIN_VALUE);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			zone.adjustMoisture(101);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		try {
			zone.adjustMoisture(Integer.MAX_VALUE);
			fail();
		} catch (IllegalArgumentException e) {
			
		}

		assertEquals(40.0, extractField(zone, "soilMoisture"));

	}

	@SuppressWarnings("unchecked")
	@Test
	public void testAddValveValidCases() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests addValve().

		SimValve valve = new SimValve("A", 1);
		SimZone zone = new SimZone("zone1", new SimSensor("sensor"));
		zone.addValve(valve);

		assertTrue(((ArrayList<SimValve>) extractField(zone, "valves"))
				.contains(valve));
	}

	@Test
	public void testAddValveInvalidCases() {
		// Tests addValve(), some of the arguments are not valid.

		SimValve valve = null;
		SimZone zone = new SimZone("zone1", new SimSensor("sensor"));
		try {
			zone.addValve(valve);
			fail();
		} catch (NullPointerException e) {
			
		}

		// The new valve is already in the zone.

		zone = new SimZone("zone1", new SimSensor("sensor"));
		for (int i = 0; i < 2; i++) {
			zone.addValve(new SimValve(Integer.toString(i), i));
		}
		valve = (SimValve) zone.valveIterator().next();

		zone.addValve(valve);

		int similarValves = 0;

		for (Iterator<?> iter = zone.valveIterator(); iter.hasNext();) {
			if (valve.equals(iter.next()))
				similarValves++;
		}

		assertEquals(1, similarValves);
	}

}
