package simulation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static tools.HelperFunctions.extractFieldAsField;
import irrigation.Irrigator;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.lang.reflect.Field;

import org.junit.Before;
import org.junit.Test;

import device.Clock;
import device.sim.MockSimClockDevice;

/**
 * Test class <code>Simulation</code>. Corresponds to requirements
 * <ul>
		<li>7.1.1.2: The left-hand side of the user interface must control and display the state of the simulated irrigation site.
 * </ul>
 */

public class SimulationTest {

	@SuppressWarnings("serial")
	// Subclass of SimEnvironment, created for testing purposes.
	private class MockEnvironment extends SimEnvironment {

		public Irrigator getLastIrrigator() {
			return lastIrrigator;
		}

		public String getLastQueriedSensor() {
			return lastQueriedSensor;
		}

		public String getLastQueriedValve() {
			return lastQueriedValve;
		}

		private String lastQueriedSensor;
		private String lastQueriedValve;
		private Irrigator lastIrrigator;

		@Override
		SimSensor getSensor(String name) {
			lastQueriedSensor = name;
			return null;
		}

		@Override
		SimValve getValve(String name) {
			lastQueriedValve = name;
			return null;
		}

		@Override
		void setIrrigator(Irrigator irrigator) {
			lastIrrigator = irrigator;
		}

		MockEnvironment(File stateFile) {
			super(stateFile);
		}
	}

	@SuppressWarnings("serial")
	// Subclass of SimControlPanel, created for Testing purposes.
	private class MockControlPanel extends SimControlPanel {

		public ActionListener getLastKeypadListener() {
			return lastListener;
		}

		public ActionListener getLastScreenButtonListener() {
			return screenButtonListener;
		}

		private ActionListener lastListener = null;
		private ActionListener screenButtonListener = null;

		@Override
		SimDisplay getDisplay() {
			return super.getDisplay();
		}

		@Override
		void setKeypadListener(ActionListener l) {
			lastListener = l;
		}

		@Override
		void setScreenButtonListener(ActionListener l) {
			screenButtonListener = l;
		}

	}

	private SimControlPanel controlPanel;
	private SimEnvironment simEnvironment;
	private Simulation simulation;

	@Before
	public void init() {
		simulation = new Simulation(null);
		controlPanel = (SimControlPanel) simulation.getComponent(0);
		simEnvironment = (SimEnvironment) simulation.getComponent(1);
	}

	@Test
	public void testSimulation() {
		// Tests Simulation's look and feel.
		assertNotNull(controlPanel);
		assertNotNull(simEnvironment);
		BorderLayout layout = (BorderLayout) simulation.getLayout();
		assertEquals(BorderLayout.WEST, layout.getConstraints(simEnvironment));
		assertEquals(BorderLayout.EAST, layout.getConstraints(controlPanel));
	}

	@Test
	public void testGetDisplay() {
		// Tests whether the control panel uses the same display
		// as the simulation.
		assertEquals(controlPanel.getDisplay(), simulation.getDisplay());
	}

	@Test
	public void testGetSensor() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests getSensor().
		
		// Install a MockEnvironment.
		Field simEnvField = extractFieldAsField(simulation, "environment");
		simEnvField.setAccessible(true);
		MockEnvironment mockEnv = new MockEnvironment(null);
		simEnvField.set(simulation, mockEnv);

		// Check whether the query for a device
		// was properly delegated to the underlying
		// environment.
		simulation.getSensor("Sensor1");
		assertEquals("Sensor1", mockEnv.getLastQueriedSensor());

		simulation.getSensor(null);
		assertNull(mockEnv.getLastQueriedSensor());

	}

	@Test
	public void testGetValve() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests getSensor().
		
		// Install a MockEnvironment.
		Field simEnvField = extractFieldAsField(simulation, "environment");
		simEnvField.setAccessible(true);
		MockEnvironment mockEnv = new MockEnvironment(null);
		simEnvField.set(simulation, mockEnv);

		// Check whether the query for a device
		// was properly delegated to the underlying
		// environment.
		simulation.getValve("Valve1");
		assertEquals("Valve1", mockEnv.getLastQueriedValve());

		simulation.getValve(null);
		assertNull(mockEnv.getLastQueriedValve());
	}

	@Test
	public void testSetKeypadListener() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests setKeypadListener().
		
		ActionListener listener = new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {

			}
		};

		MockControlPanel mockPanel = new MockControlPanel();

		// Replace the current controlPanel with a mock panel.
		Field controlPanelField = extractFieldAsField(simulation,
				"controlPanel");
		controlPanelField.setAccessible(true);
		controlPanelField.set(simulation, mockPanel);

		simulation.setKeypadListener(listener);

		assertEquals(listener, mockPanel.getLastKeypadListener());

	}

	@Test
	public void testSetScreenButtonListener() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests setScreenButtonListener().

		ActionListener listener = new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {

			}
		};

		MockControlPanel mockPanel = new MockControlPanel();

		// Replace the current controlPanel with a mock panel.
		Field controlPanelField = extractFieldAsField(simulation,
				"controlPanel");
		controlPanelField.setAccessible(true);
		controlPanelField.set(simulation, mockPanel);

		simulation.setScreenButtonListener(listener);

		assertEquals(listener, mockPanel.getLastScreenButtonListener());

	}

	@Test
	public void testSetIrrigator() throws IllegalArgumentException,
			SecurityException, IllegalAccessException, NoSuchFieldException {
		// Tests setIrrigator().
		
		// Replace the current environment with a mock panel.
		Field simEnvField = extractFieldAsField(simulation, "environment");
		simEnvField.setAccessible(true);
		MockEnvironment mockEnv = new MockEnvironment(null);
		simEnvField.set(simulation, mockEnv);
		
		Clock.setClockDevice(new MockSimClockDevice());
		
		Irrigator i = new Irrigator(null);

		simulation.setIrrigator(i);
		assertEquals(i, mockEnv.getLastIrrigator());

		simulation.setIrrigator(null);
		assertNull(mockEnv.getLastIrrigator());
	}

}
